// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <unistd.h>

#include <qcstring.h>
#include <qfile.h>
#include <qstringlist.h>
#include <klocale.h>
#include <kmessagebox.h>

#include <sqlite3.h>

#include "kldraw.h"
#include "libldr.h"
#include "mainwindow.h"
#include "partrenderer.h"
#include "progresswidget.h"

#include "dbmanager.h"

using namespace std;
using namespace ldraw;

KLDrawDBManager *KLDrawDBManager::instance_ = 0L;

KLDrawDBManager::KLDrawDBManager()
{
	instance_ = this;
	isLoaded_ = false;
	
	bool needRescan = false;
	const QCString path = QFile::encodeName(KLDraw::saveLocation("")+"parts.db");
	
	QFile file(path);
	if (!file.exists() || file.size() == 0L)
		needRescan = true;
	else if (file.open(IO_ReadOnly)) {
		QString format;
		file.readLine(format, 50);
		if (!format.startsWith("SQLite format 3")) {
			file.close();
			QFile::remove(path);
		} else {
			file.close();
		}
	}
	
	if (sqlite3_open(path, &db_) != SQLITE_OK) {
		KMessageBox::error(KLDrawMainWindow::self(), i18n("Could not open the part database."));
		return;
	}
	
	isLoaded_ = true;
	
	if (needRescan || !hasTables()) {
		createTables();
		rescanDB();
	}
	
	query("PRAGMA default_synchronous = OFF;");
}

KLDrawDBManager::~KLDrawDBManager()
{
	if (isLoaded_)
		sqlite3_close(db_);
}

void KLDrawDBManager::rescanDB()
{
	query("DELETE FROM parts;");
	// TODO delete partimgs
	
	KLDrawProgressWidget *progressWidget = new KLDrawProgressWidget();
	progressWidget->show();
	
	QString path = KLDraw::saveLocation("partimgs/");
	string ldrpartpath = g_ldrawpath+"/parts/";
	int totalSize = g_partlist.size();
	
	KLDrawPartRenderer *renderer = new KLDrawPartRenderer(320, 240);
	
	int i = 0;
	for (map<string, string>::iterator it = g_partlist.begin(); it != g_partlist.end(); ++it) {
		progressWidget->updateProgress(++i, totalSize);
		if (progressWidget->wasCancelled()) {
			KMessageBox::error(KLDrawMainWindow::self(), i18n("Interrupted by user."));
			break;
		}
		
		// Omit subparts
		if (((*it).second[0] == 's' || (*it).second[0] == 'S') && (*it).second[1] == DIRECTORY_SEPARATOR)
			continue;
		model_multipart *m = new model_multipart(ldrpartpath + (*it).second, true);
		
		// If current part is unnecessary, skip it.
		if (m->get_model()->desc().find("oved to") != string::npos) { // *oved to #####
			delete m;
			continue;
		}
		
		// Render
		KLDraw::checkQuadValidity(m->get_model());
		renderer->render(m->get_model(), true).save(path+(*it).second.c_str()+".png", "PNG");
		
		// Insert this into db
		if (!m->get_model()->has_dimension())
			m->get_model()->get_dimension();
		vector3d &min = m->get_model()->dimension().first;
		vector3d &max = m->get_model()->dimension().second;
		
		QString qFilename = KLDraw::escape(QString((*it).second.c_str()));
		QString qPartno = qFilename.section('.', 0, 0);
		QString qDesc = KLDraw::escape(QString(m->get_model()->desc()));
		QString qFamily;
		
		if (qDesc[0] == '~')
			qFamily = i18n("~");
		else
			qFamily = qDesc.section(' ', 0, 0);
		
		if (qFamily[0] == '_') // Precolored parts
			qFamily = qFamily.right(qFamily.length()-1);
		
		insert(QString("INSERT INTO parts(partno, desc, filename, family, minx, maxx, miny, maxy, minz, maxz) VALUES('%1', '%2', '%3', '%4', %5, %6, %7, %8, %9, %9);").arg(qPartno).arg(qDesc).arg(qFilename).arg(qFamily).arg(min.x()).arg(max.x()).arg(min.y()).arg(max.y()).arg(min.z()).arg(max.z()));
		
		delete m;
	}
	
	delete renderer;
}

bool KLDrawDBManager::hasTables()
{
	QStringList v1 = query("SELECT COUNT(partno) FROM parts LIMIT 1 OFFSET 0;");
	QStringList v2 = query("SELECT COUNT(partno) FROM favorite LIMIT 1 OFFSET 0;");
	
	return !(v1.isEmpty() && v2.isEmpty());
}

void KLDrawDBManager::createTables()
{
	query(
		"CREATE TABLE parts ("
		"    partno TEXT,"
		"    desc TEXT,"
		"    filename TEXT,"
		"    family TEXT,"
		"    xsize INTEGER,"
		"    ysize INTEGER,"
		"    zsize INTEGER,"
		"    minx REAL,"
		"    maxx REAL,"
		"    miny REAL,"
		"    maxy REAL,"
		"    minz REAL,"
		"    maxz REAL,"
		"    count INTEGER"
		");"
	);
	
	query(
		"CREATE TABLE favorite ("
		"    partno TEXT,"
		"    pgroup TEXT"
		");"
	);
}

QStringList KLDrawDBManager::query(const QString &statement)
{
	if (!isLoaded_) // There is no DB connection
		return QStringList();
	
	QStringList values;
	int error, rc;
	const char *tail;
	sqlite3_stmt *stmt;
	int busyCnt = 0;
	int retryCnt = 0;
	
	do {
		do {
			if (busyCnt) {
				::usleep(100000);
				// BUSY COUNTER
			}
			error = sqlite3_prepare(db_, statement.utf8(), -1, &stmt, &tail);
		} while (error == SQLITE_BUSY && busyCnt++ < 120);
		
		if (error != SQLITE_OK) {
			if (error == SQLITE_BUSY)
				KMessageBox::sorry(KLDrawMainWindow::self(), i18n("Sorry, Database is locked right now. Please try again later."));
			else
				std::cerr << sqlite3_errmsg(db_) << std::endl;
			values = QStringList();
		} else {
			busyCnt = 0;
			int number = sqlite3_column_count(stmt);
			
			while (1) {
				error = sqlite3_step(stmt);
				
				if (error == SQLITE_BUSY) {
					if (busyCnt++ > 120) {
						KMessageBox::sorry(KLDrawMainWindow::self(), i18n("Sorry, Database is locked right now. Please try again later."));
						break;
					}
					::usleep(100000);
					continue;
				}
				if (error == SQLITE_DONE || error == SQLITE_ERROR)
					break;
				
				for (int i = 0; i < number; i++)
					values << QString::fromUtf8(reinterpret_cast<const char *>(sqlite3_column_text(stmt, i)));
			}
			
			rc = sqlite3_finalize(stmt);
			
			if (error != SQLITE_DONE && rc != SQLITE_SCHEMA) {
				values = QStringList();
			}
			if (rc == SQLITE_SCHEMA) {
				retryCnt++;
				if (retryCnt >= 10) {
					KMessageBox::error(KLDrawMainWindow::self(), i18n("Retry count has reached maximum"));
					values = QStringList();
				}
			}
		}
	} while (rc == SQLITE_SCHEMA && retryCnt < 10);
	
	return values;
}

int KLDrawDBManager::insert(const QString &statement)
{
	if (!isLoaded_) // There is no DB connection
		return 0;
	
	QStringList values;
	int error, rc;
	const char *tail;
	sqlite3_stmt *stmt;
	int busyCnt = 0;
	int retryCnt = 0;
	
	do {
		do {
			if (busyCnt) {
				::usleep(100000);
				// BUSY COUNTER
			}
			error = sqlite3_prepare(db_, statement.utf8(), -1, &stmt, &tail);
		} while (error == SQLITE_BUSY && busyCnt++ < 120);
		
		if (error != SQLITE_OK) {
			if (error == SQLITE_BUSY)
				KMessageBox::sorry(KLDrawMainWindow::self(), i18n("Sorry, Database is locked right now. Please try again later."));
			values = QStringList();
		} else {
			busyCnt = 0;
			
			while (1) {
				error = sqlite3_step(stmt);
				
				if (error == SQLITE_BUSY) {
					if (busyCnt++ > 120) {
						KMessageBox::sorry(KLDrawMainWindow::self(), i18n("Sorry, Database is locked right now. Please try again later."));
						break;
					}
						::usleep(100000);
						continue;
				}
				if (error == SQLITE_DONE || error == SQLITE_ERROR)
					break;
			}
			
			rc = sqlite3_finalize(stmt);
			
			if (error != SQLITE_DONE && rc != SQLITE_SCHEMA) {
				values = QStringList();
			}
			if (rc == SQLITE_SCHEMA) {
				retryCnt++;
				if (retryCnt >= 10) {
					KMessageBox::error(KLDrawMainWindow::self(), i18n("Retry count has reached maximum"));
					values = QStringList();
				}
			}
		}
	} while (rc == SQLITE_SCHEMA && retryCnt < 10);
	
	return sqlite3_last_insert_rowid(db_);
}
