// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <math.h>

#include <qimage.h>
#include <qtimer.h>
#include <qthread.h>

#include "mainwindow.h"
#include "libldr.h"
#include "kldraw.h"
#include "partrenderer.h"

#include "document.h"

using namespace std;
using namespace ldraw;

/* Default quarter-view transformation matrix */
const matrix isometricProjectionMatrix(
	 0.707106f, 0.0f     ,  0.707106f,
	 0.408248f, 0.816497f, -0.408248f,
	-0.57735f , 0.57735f ,  0.57735f
);

const KLDrawViewport defaultViewport = { -200.0f, -150.0f, 200.0f, 150.0f, 1.333333f };

// Creates *new* documents from scratch
KLDrawDocument::KLDrawDocument(int number, QObject *parent) : QObject(parent), QTab(QString::null)
{
	canSave_ = false;
	readOnly_ = false;
	firstSave_ = true;
	
	renderer_ = new KLDrawPartRenderer(256, 256);
	
	QString fileName = QString("untitled%1.ldr").arg(number);
	setText(fileName);
	
	modelBase_ = new model_multipart;
	modelBase_->get_model()->name() = string(fileName.local8Bit());
	modelBase_->get_model()->author() = "KLDraw";
	modelBase_->get_model()->desc() = "Untitled model";
	
	activeModel_ = modelBase_->get_model();
	freeViewport_ = isometricProjectionMatrix;
	
	modelPixmap_ = prettyPreview(modelBase_->get_model());
	
	recalibrateScreenDimension();
}

// Load an existing model from a file
KLDrawDocument::KLDrawDocument(QString fileName, QObject *parent) : QObject(parent), QTab(QString::null)
{
	canSave_ = false;
	readOnly_ = false;
	firstSave_ = false;
	
	renderer_ = new KLDrawPartRenderer(256, 256);
	
	modelBase_ = new model_multipart(string(fileName.local8Bit()), true);
	
	model *mainModel = modelBase_->get_model();
	KLDraw::checkQuadValidity(mainModel);
	
	mainModel->get_dimension();
	activeModel_ = mainModel;
	freeViewport_ = isometricProjectionMatrix;
	
	setText(QString::fromLocal8Bit(modelBase_->get_model()->name().c_str()));
	
	updatePixmap();
	recalibrateScreenDimension();
}

KLDrawDocument::~KLDrawDocument()
{
	if (modelBase_)
		delete modelBase_;
	
	delete renderer_;
}

QPixmap* KLDrawDocument::getSubmodelPixmap(string key)
{
	map<string, QPixmap>::iterator it = submodelPixmap_.find(key);
	
	if (it != submodelPixmap_.end())
		return &((*it).second);
	else
		return 0L;
}

void KLDrawDocument::setSaveable(bool s)
{
	canSave_ = s;
	
	// TODO 'disk' icon on tab when the model is modified
}

void KLDrawDocument::recalibrateScreenDimension()
{
	if (activeModel_->size() == 0) {
		for (int i = 0; i < 7; i++)
			viewport_[i] = defaultViewport;
	} else {
		int count = 0;
		
		for (model::iterator it = activeModel_->begin(); it != activeModel_->end(); ++it) {
			type t = (*it)->get_type();
			if (t == type_line || t == type_triangle || t == type_quadrilateral || t == type_condline)
				++count;
			else if (t == type_extref) {
				element_extref *e = CAST_AS_EXTREF(*it);
				if (e->get_model())
					++count;
			}
		}
		
		// If current model has actual content
		if (count) {
			// Update bounding box
			activeModel_->get_dimension();
			
			vector3d &min = activeModel_->dimension().first;
			vector3d &max = activeModel_->dimension().second;
			
			center_ = (min+max) * 0.5f;
			length_ = vector3d(fabs(max.x()-min.x()), fabs(max.y()-min.y()), fabs(max.z()-min.z()));
			
			float xlen = length_.x()*0.5f;
			float ylen = length_.y()*0.5f;
			float zlen = length_.z()*0.5f;
			
			/* Each index stores viewport data of following perspectives:
			 * 0: Top
			 * 1: Bottom
			 * 2: Front
			 * 3: Back
			 * 4: Left
			 * 5: Right
			 * 6: Free
			 */
			
			viewport_[0].left   =  min.x() - (xlen*0.1f);
			viewport_[0].right  =  max.x() + (xlen*0.1f);
			viewport_[0].top    = -max.z() - (zlen*0.1f);
			viewport_[0].bottom = -min.z() + (zlen*0.1f);
			viewport_[0].aspectRatio = length_.x()/length_.z();
			
			viewport_[1].left   = min.x() - (xlen*0.1f);
			viewport_[1].right  = max.x() + (xlen*0.1f);
			viewport_[1].top    = min.z() - (zlen*0.1f);
			viewport_[1].bottom = max.z() + (zlen*0.1f);
			viewport_[1].aspectRatio = length_.x()/length_.z();
			
			viewport_[2].left   = min.x() - (xlen*0.1f);
			viewport_[2].right  = max.x() + (xlen*0.1f);
			viewport_[2].top    = min.y() - (ylen*0.1f);
			viewport_[2].bottom = max.y() + (ylen*0.1f);
			viewport_[2].aspectRatio = length_.x()/length_.y();
			
			viewport_[3].left   = -max.x() - (xlen*0.1f);
			viewport_[3].right  = -min.x() + (xlen*0.1f);
			viewport_[3].top    =  min.y() - (ylen*0.1f);
			viewport_[3].bottom =  max.y() + (ylen*0.1f);
			viewport_[3].aspectRatio = length_.x()/length_.y();
			
			viewport_[4].left   = -max.z() - (zlen*0.1f);
			viewport_[4].right  = -min.z() + (zlen*0.1f);
			viewport_[4].top    =  min.y() - (ylen*0.1f);
			viewport_[4].bottom =  max.y() + (ylen*0.1f);
			viewport_[4].aspectRatio = length_.z()/length_.y();
			
			viewport_[5].left   = min.z() - (zlen*0.1f);
			viewport_[5].right  = max.z() + (zlen*0.1f);
			viewport_[5].top    = min.y() - (ylen*0.1f);
			viewport_[5].bottom = max.y() + (ylen*0.1f);
			viewport_[5].aspectRatio = length_.z()/length_.y();
			
			/* Getting viewport dimension from 3d-view is a bit expensive task.
			 * Following method takes viewport dimension from precalculated bounding box
			 * instead of every meshes and lines from the model.
			 * Since a bounding box can't represent exact dimension of model from arbitrary viewpoint,
			 * this method is reasonably fast but inaccurate.
			 */
			
			// Initialize with extreme value for comparison
			viewport_[6].left   = 1e30;
			viewport_[6].right  = -1e30;
			viewport_[6].top    = 1e30;
			viewport_[6].bottom = -1e30;
			
			vector3d freeVec[8];
			
			// Transform
			freeVec[0] = freeViewport_ * min;
			freeVec[1] = freeViewport_ * vector3d(min.x(), min.y(), max.z());
			freeVec[2] = freeViewport_ * vector3d(min.x(), max.y(), min.z());
			freeVec[3] = freeViewport_ * vector3d(min.x(), max.y(), max.z());
			freeVec[4] = freeViewport_ * vector3d(max.x(), min.y(), min.z());
			freeVec[5] = freeViewport_ * vector3d(max.x(), min.y(), max.z());
			freeVec[6] = freeViewport_ * vector3d(max.x(), max.y(), min.z());
			freeVec[7] = freeViewport_ * max;
			
			// Compare
			for (int i = 0; i < 8; i++) {
				if (freeVec[i].x() < viewport_[6].left)
					viewport_[6].left = freeVec[i].x();
				if (freeVec[i].x() > viewport_[6].right)
					viewport_[6].right = freeVec[i].x();
				if (freeVec[i].y() > viewport_[6].bottom)
					viewport_[6].bottom = freeVec[i].y();
				if (freeVec[i].y() < viewport_[6].top)
					viewport_[6].top = freeVec[i].y();
			}
			
			float fxlen = fabs(viewport_[6].right  - viewport_[6].left)*0.5f; 
			float fylen = fabs(viewport_[6].bottom - viewport_[6].top )*0.5f;
			
			// Update
			viewport_[6].left   -= fxlen*0.1f;
			viewport_[6].right  += fxlen*0.1f;
			viewport_[6].top    -= fylen*0.1f;
			viewport_[6].bottom += fylen*0.1f;
			viewport_[6].aspectRatio = fxlen/fylen;
		} else {
			for (int i = 0; i < 7; i++)
				viewport_[i] = defaultViewport;
		}
	}
}

void KLDrawDocument::resetPerspective()
{
	freeViewport_ = isometricProjectionMatrix;
}

void KLDrawDocument::updatePixmap()
{
	modelPixmap_ = prettyPreview(getModel()->get_model());
	for (std::map<std::string, model *>::iterator it = getModel()->get_submodels().begin(); it != getModel()->get_submodels().end(); ++it)
		submodelPixmap_[(*it).first] = prettyPreview((*it).second);
}

QPixmap KLDrawDocument::prettyPreview(model *m)
{
	QPixmap pm;
	pm.convertFromImage(renderer_->render(m).convertToImage().smoothScale(96, 96));
	
	return pm;
}
