// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <string>
#include <cmath>

#include "document.h"
#include "kldraw.h"
#include "list.h"
#include "mainwindow.h"

#include "editor.h"

#define EXIT_IF_NO_DOCUMENT if (!KLDrawMainWindow::self()->getCurrentCursor()) return;
#define DOCUMENT (*(KLDrawMainWindow::self()->activeDocument())) 
#define ACTIVE_MODEL ((*(KLDrawMainWindow::self()->activeDocument()))->getActiveModel())
#define CURSOR (KLDrawMainWindow::self()->getCurrentCursor())
#define CURSOR_ELEMENT (KLDrawMainWindow::self()->getCurrentCursor()->item())

using namespace std;
using namespace ldraw;

KLDrawEditor::KLDrawEditor(QObject *parent, const char *name) : QObject(parent, name)
{
	grid_ = Standard;
}

KLDrawEditor::~KLDrawEditor()
{
	
}

float KLDrawEditor::gridScale(int axis) const
{
	switch (grid_) {
		case Sparse:
			if (axis == 1)
				return 24.0f;
			else
				return 20.0f;
		case Standard:
			if (axis == 1)
				return 8.0f;
			else
				return 10.0f;
		case Dense:
			if (axis == 1)
				return 4.0f;
			else
				return 5.0f;
		case Fine:
		default:
			return 1.0f;
	}
}

void KLDrawEditor::insertObject(const QString &filename, const matrix &tr, const color &col)
{
	KLDrawListItem *curItem = CURSOR;
	element_extref *newItem;
	KLDrawListItem *newListItem;
	string name(filename.lower().local8Bit());
	bool notIncluded = false;
	
	if (g_modelpool.find(name) == g_modelpool.end())
		notIncluded = true;
	newItem = new element_extref(col, tr, name);
	if (notIncluded)
		KLDraw::checkQuadValidity(newItem->get_model());
	
	if (!curItem || !(curItem->itemBelow()))
		ACTIVE_MODEL->push_back(newItem);
	else {
		int i = 0;
		for (model::iterator it = ACTIVE_MODEL->begin(); it != ACTIVE_MODEL->end(); ++it) {
			if (i == curItem->id() + 1) {
				ACTIVE_MODEL->insert(it, newItem);
				break;
			}
			++i;
		}
	}
	
	if (!curItem) {
		newListItem = new KLDrawListItem(KLDrawList::instance(), 0, newItem);
	} else {
		newListItem = new KLDrawListItem(dynamic_cast<KLDrawList *>(curItem->listView()), curItem, curItem->id() + 1, newItem);
		if (newListItem->itemBelow())
			dynamic_cast<KLDrawListItem *>(newListItem->itemBelow())->increaseFollowing(1);
	}
	
	newListItem->listView()->setSelected(newListItem, true);
	emit cursorChanged(newListItem);
	
	changeModel();
}

void KLDrawEditor::moveCurrentObject(const ldraw::matrix &m)
{
	EXIT_IF_NO_DOCUMENT;
	
	KLDrawListItem *c = CURSOR;
	
	if (c->item()->get_type() != type_extref)
		return;
	
	element_extref *i = CAST_AS_EXTREF(c->item());
	i->set_matrix(m);
	
	// Update information appearing on list
	c->setText(1, i->serialize());
	
	changeModel();
}

void KLDrawEditor::modifyMatrix(KLDrawListItem *item, const ldraw::matrix &m)
{
	if (item->item()->get_type() != type_extref)
		return;
	
	CAST_AS_EXTREF(item->item())->set_matrix(m);
	item->setText(1, item->item()->serialize());
	
	changeModel();
}

void KLDrawEditor::gridSparse()
{
	grid_ = Sparse;
}

void KLDrawEditor::gridStandard()
{
	grid_ = Standard;
}

void KLDrawEditor::gridDense()
{
	grid_ = Dense;
}

void KLDrawEditor::gridFine()
{
	grid_ = Fine;
}

void KLDrawEditor::rotateXCW()
{
	ldraw::matrix tm;
	float r;
	
	switch (grid_) {
		case Sparse:
			r = 90.0f * M_PI / 180.0f;
			break;
		case Standard:
			r = 45.0f * M_PI / 180.0f;
			break;
		case Dense:
			r = 22.5f * M_PI / 180.0f;
			break;
		case Fine:
			r = 10.0f * M_PI / 180.0f;
	}
	
	tm.value(1, 1) = cos(r);
	tm.value(1, 2) = sin(-r);
	tm.value(2, 1) = sin(r);
	tm.value(2, 2) = cos(r);
	
	doRotate(tm);
}

void KLDrawEditor::rotateXCCW()
{
	ldraw::matrix tm;
	float r;
	
	switch (grid_) {
		case Sparse:
			r = -90.0f * M_PI / 180.0f;
			break;
		case Standard:
			r = -45.0f * M_PI / 180.0f;
			break;
		case Dense:
			r = -22.5f * M_PI / 180.0f;
			break;
		case Fine:
			r = -10.0f * M_PI / 180.0f;
	}
	
	tm.value(1, 1) = cos(r);
	tm.value(1, 2) = sin(-r);
	tm.value(2, 1) = sin(r);
	tm.value(2, 2) = cos(r);
	
	doRotate(tm);
}

void KLDrawEditor::rotateYCW()
{
	ldraw::matrix tm;
	float r;
	
	switch (grid_) {
		case Sparse:
			r = 90.0f * M_PI / 180.0f;
			break;
		case Standard:
			r = 45.0f * M_PI / 180.0f;
			break;
		case Dense:
			r = 22.5f * M_PI / 180.0f;
			break;
		case Fine:
			r = 10.0f * M_PI / 180.0f;
	}
	
	tm.value(0, 0) = cos(r);
	tm.value(2, 0) = sin(-r);
	tm.value(0, 2) = sin(r);
	tm.value(2, 2) = cos(r);
	
	doRotate(tm);
}

void KLDrawEditor::rotateYCCW()
{
	ldraw::matrix tm;
	float r;
	
	switch (grid_) {
		case Sparse:
			r = -90.0f * M_PI / 180.0f;
			break;
		case Standard:
			r = -45.0f * M_PI / 180.0f;
			break;
		case Dense:
			r = -22.5f * M_PI / 180.0f;
			break;
		case Fine:
			r = -10.0f * M_PI / 180.0f;
	}
	
	tm.value(0, 0) = cos(r);
	tm.value(2, 0) = sin(-r);
	tm.value(0, 2) = sin(r);
	tm.value(2, 2) = cos(r);
	
	doRotate(tm);
}

void KLDrawEditor::rotateZCW()
{
	ldraw::matrix tm;
	float r;
	
	switch (grid_) {
		case Sparse:
			r = 90.0f * M_PI / 180.0f;
			break;
		case Standard:
			r = 45.0f * M_PI / 180.0f;
			break;
		case Dense:
			r = 22.5f * M_PI / 180.0f;
			break;
		case Fine:
			r = 10.0f * M_PI / 180.0f;
	}
	
	tm.value(0, 0) = cos(r);
	tm.value(0, 1) = sin(-r);
	tm.value(1, 0) = sin(r);
	tm.value(1, 1) = cos(r);
	
	doRotate(tm);
}

void KLDrawEditor::rotateZCCW()
{
	ldraw::matrix tm;
	float r;
	
	switch (grid_) {
		case Sparse:
			r = -90.0f * M_PI / 180.0f;
			break;
		case Standard:
			r = -45.0f * M_PI / 180.0f;
			break;
		case Dense:
			r = -22.5f * M_PI / 180.0f;
			break;
		case Fine:
			r = -10.0f * M_PI / 180.0f;
	}
	
	tm.value(0, 0) = cos(r);
	tm.value(0, 1) = sin(-r);
	tm.value(1, 0) = sin(r);
	tm.value(1, 1) = cos(r);
	
	doRotate(tm);
}

void KLDrawEditor::moveXM()
{
	moveRel(0, -1);
}

void KLDrawEditor::moveXP()
{
	moveRel(0, 1);
}

void KLDrawEditor::moveYM()
{
	moveRel(1, -1);
}

void KLDrawEditor::moveYP()
{
	moveRel(1, 1);
}

void KLDrawEditor::moveZM()
{
	moveRel(2, -1);
}

void KLDrawEditor::moveZP()
{
	moveRel(2, 1);
}

void KLDrawEditor::changeColor(const ldraw::color_entity *e)
{
	EXIT_IF_NO_DOCUMENT;
	
	ldraw::element_extref *r = CAST_AS_EXTREF(CURSOR_ELEMENT);
	r->set_color(color(e->id1));	
	
	// Update information appearing on list
	CURSOR->setText(1, CURSOR_ELEMENT->serialize());
	
	changeModel();
}

void KLDrawEditor::remove()
{
	EXIT_IF_NO_DOCUMENT;
	
	model *m = DOCUMENT->getActiveModel();
	KLDrawListItem *tmp = CURSOR;
	KLDrawListItem *nptr = 0L;
	int id = tmp->id();
	
	if (tmp->itemBelow()) {
		dynamic_cast<KLDrawListItem *>(tmp->itemBelow())->increaseFollowing(-1);
		nptr = dynamic_cast<KLDrawListItem *>(tmp->itemBelow());
	} else if (tmp->itemAbove()) {
		nptr = dynamic_cast<KLDrawListItem *>(tmp->itemAbove());
	}
	
	// Delete an item
	int i = 0;
	for (model::iterator it = m->begin(); it != m->end(); ++it) {
		if (id == i) {
			delete *it;
			m->erase(it);
			break;
		}
		++i;
	}
	
	// Remove from list
	KLDrawList::instance()->takeItem(tmp);
	if (nptr)
		KLDrawList::instance()->setSelected(nptr, true);
	emit cursorChanged(nptr ? nptr : 0L);
	
	changeModel();
}

void KLDrawEditor::changeModel()
{
	KLDrawDocument *d = *(KLDrawMainWindow::self()->activeDocument());
	
	DOCUMENT->getActiveModel()->get_dimension();
	
	if (!d->canSave()) {
		d->setSaveable(true);
		emit modelModified();
	}
	
	emit needUpdate();
}

void KLDrawEditor::doRotate(const ldraw::matrix &m)
{
	EXIT_IF_NO_DOCUMENT;
	
	ldraw::element_base *e = CURSOR_ELEMENT;
	
	if (e->get_type() == ldraw::type_extref) {
		ldraw::element_extref *r = CAST_AS_EXTREF(e);
		r->set_matrix(KLDraw::estimate(r->get_matrix() * m));
		
		// Update information appearing on list
		CURSOR->setText(1, CURSOR_ELEMENT->serialize());
		
		changeModel();
	}
}

void KLDrawEditor::moveRel(int axis, int dir)
{
	EXIT_IF_NO_DOCUMENT;
	
	ldraw::element_base *e = CURSOR_ELEMENT;
	float factor;
	
	if (e->get_type() != type_extref)
		return;
	
	if (axis != 1) {
		switch (grid_) {
			case Sparse:
				factor = 20.0f;
				break;
			case Standard:
				factor = 10.0f;
				break;
			case Dense:
				factor = 5.0f;
				break;
			case Fine:
				factor = 1.0f;
		}
	} else {
		switch (grid_) {
			case Sparse:
				factor = 24.0f;
				break;
			case Standard:
				factor = 8.0f;
				break;
			case Dense:
				factor = 4.0f;
				break;
			case Fine:
				factor = 1.0f;
		}
	}
	
	ldraw::element_extref *r = CAST_AS_EXTREF(e);
	ldraw::matrix m = r->get_matrix();
	m.value(axis, 3) += factor * dir;
	r->set_matrix(m);
	
	// Update information appearing on list
	CURSOR->setText(1, CURSOR_ELEMENT->serialize());
	
	changeModel();
}
