// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <cmath>

#include <kglobal.h>
#include <kstandarddirs.h>

#include "kldraw.h"

using namespace std;
using namespace ldraw;

namespace KLDraw
{

QMutex globalDirsMutex;

// Replace ' by \'. Used for SQL query.
QString escape(const QString &str)
{
	QString r = str;
	r.replace('\'', "''");
	return r;
}

QString unescape(const QString &str)
{
	QString r = str;
	r.replace("''", "'");
	return r;
}

QString saveLocation(const QString &directory)
{
	globalDirsMutex.lock();
	QString result = KGlobal::dirs()->saveLocation("data", QString("kldraw/") + directory, true);
	globalDirsMutex.unlock();
	
	return result;
}

/* ------------------------------------------------
 *         0     1                  0     1
 *
 *         3     2                  2     3
 * <normal drawing order>   <twisted drawing order>
 * ------------------------------------------------
 *
 * Drawing a quad by drawing lines through 0->1->2->3,
 * In the first case, it draws correct quad but the other makes a bow tie.
 * Following function recurses down the model, finds bowtie quads and fix them.
 */
void checkQuadValidity(model *m)
{
	for (model::iterator it = m->begin(); it != m->end(); ++it) {
		type elemtype = (*it)->get_type();
		if (elemtype == type_quadrilateral) {
			element_quadrilateral *l = CAST_AS_QUADRILATERAL(*it);
			if (l->bowtie_test()) {
				vector3d tmp = l->pos3();
				l->pos3() = l->pos4();
				l->pos4() = tmp;
			}
		} else if (elemtype == type_extref) {
			element_extref *l = CAST_AS_EXTREF(*it);
			if (l->get_model())
				checkQuadValidity(l->get_model());
		}
	}
}

matrix estimate(const matrix &m)
{
	matrix n = m;
	
	for (int i = 0; i < 3; ++i) {
		for (int j = 0; j < 3; ++j) {
			float e = m.value(i, j);
			if (fabs(fabs(e) - 1.0f) < EPSILON)
				n.value(i, j) = e > 0.0f ? 1.0f:-1.0f;
			else if (fabs(e) < EPSILON)
				n.value(i, j) = 0.0f;
			else if (fabs(fabs(e) - 0.707106f) < EPSILON)
				n.value(i, j) = e > 0.0f ? 0.707106f:-0.707106f;
			else if (fabs(fabs(e) - 0.5f) < EPSILON)
				n.value(i, j) = e > 0.0f ? 0.5f:-0.5f;
			else if (fabs(fabs(e) - 0.866025f) < EPSILON)
				n.value(i, j) = e > 0.0f ? 0.866025f:-0.866025f;
			else
				n.value(i, j) = e;
		}
	}
	
	return n;
}

// Simple-and-dirty modulo calculation for floating points
float modulo(float dividend, float divisor)
{
	if (dividend > 0.0f) {
		while (dividend >= divisor)
			dividend -= divisor;
		return dividend;
	} else if (dividend < 0.0f) {
		while (dividend <= -divisor)
			dividend += divisor;
		return divisor;
	} else {
		return 0.0f;
	}
}

string serializePOV(const vector3d &v)
{
	ostringstream s;
	s << "<" << v.x() << ", " << v.y() << ", " << v.z() << ">";
	return s.str();
}

string serializePOV(const matrix &m)
{
	ostringstream s;
	s << "<" << m.value(0, 0) << ", " << m.value(0, 1) << ", " << m.value(0, 2) << ", " << m.value(1, 0) << ", " << m.value(1, 1) << ", " << m.value(1, 2) << ", " << m.value(2, 0) << ", " << m.value(2, 1) << ", " << m.value(2, 2) << ", " << m.value(3, 0) << ", " << m.value(3, 1) << ", " << m.value(3, 2) << ">";
	return s.str();
}

}
