/* libLDR: Portable and easy-to-use LDraw format abstraction & I/O reference library *
 * To obtain more information about LDraw, visit http://www.ldraw.org.               *
 * Distributed in terms of the General Public License v2                             *
 *                                                                                   *
 * Author: (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>                    */

// System-dependent headers
#ifndef WIN32 // *nix
#include <sys/types.h>
#include <dirent.h>
#else
#include <windows.h>
#endif

#include <cmath>

#include "libldr.h"

using namespace std;

namespace ldraw {

// Predefined color chart
const color_entity color_chart[] = {
	{material_normal,   { 33,  33,  33, 255},   0,  -1, "Black"},
	{material_normal,   {  0,  51, 178, 255},   1,  -1, "Blue"},
	{material_normal,   {107, 238, 144, 255},  10, 418, "Bright Green"},
	{material_normal,   { 92,  32,   0, 255},   6,  -1, "Brown"},
	{material_metallic, {225, 110,  19, 255}, 334,  -1, "Chrome Gold"},
	{material_metallic, {224, 224, 224, 255}, 383,  -1, "Chrome Silver"},
	{material_clear,    {255, 255, 255, 230},  47,  -1, "Clear (Trans White)"},
	{material_normal,   {  0,  29, 104, 255}, 272,  -1, "Dark Blue"},
	{material_normal,   { 99,  95,  82, 255},   8,  -1, "Dark Gray"},
	{material_normal,   { 39,  70,  44, 255}, 288,  -1, "Dark Green"},
	{material_normal,   {179,  62,   0, 255}, 484,  -1, "Dark Orange"},
	{material_normal,   {223, 102, 149, 255},   5,  -1, "Dark Pink"},
	{material_normal,   {120,   0,  28, 255}, 320,  -1, "Dark Red"},
	{material_normal,   { 99,  95,  97, 255},  72,  -1, "Dark Stone Gray"},
	{material_normal,   {197, 151,  80, 255},  28,  -1, "Dark Tan"},
	{material_normal,   {209, 131,   4, 255}, 366,  -1, "Earth Orange"},
	{material_normal,   {208, 208, 208, 255}, 494,  -1, "Electric Contact"},
	{material_normal,   {193, 194, 193, 255},   7,  -1, "Gray"},
	{material_normal,   {  0, 140,  20, 255},   2,  -1, "Green"},
	{material_normal,   {107, 171, 220, 255},   9,  -1, "Light Blue"},
	{material_normal,   {230, 227, 218, 255}, 503,  -1, "Light Gray"},
	{material_normal,   {186, 255, 206, 255},  17, 431, "Light Green"},
	{material_normal,   {254, 159,   6, 255}, 462,  -1, "Light Orange"},
	{material_normal,   {255, 133, 122, 255},  12, 463, "Light Red"},
	{material_normal,   {215, 196, 230, 255},  20,  -1, "Light Violet"},
	{material_normal,   {253, 232, 150, 255},  18, 495, "Light Yellow"},
	{material_normal,   {215, 240,   0, 255},  27,  -1, "Lime"},
	{material_normal,   {216,  27, 109, 255},  26,  -1, "Magenta"},
	{material_normal,   {249,  96,   0, 255},  25,  -1, "Orange"},
	{material_normal,   {147, 130, 103, 255}, 134,  -1, "Pearl Copper"},
	{material_normal,   {215, 169,  75, 255}, 142,  -1, "Pearl Gold"},
	{material_normal,   {171, 173, 172, 255}, 135,  -1, "Pearl Gray"},
	{material_normal,   {106, 122, 150, 255}, 137,  -1, "Pearl Sand Blue"},
	{material_normal,   {224, 255, 176, 255},  21,  -1, "Phosphor White (Glow in the Dark)"},
	{material_normal,   {249, 164, 198, 255},  13,  -1, "Pink"},
	{material_normal,   {196,   0,  38, 255},   4,  -1, "Red"},
	{material_normal,   {105,  64,  39, 255},  70,  -1, "Reddish Brown"},
	{material_rubber,   { 33,  33,  33, 255}, 256,  -1, "Rubber Black"},
	{material_rubber,   {  0,  51, 178, 255}, 273,  -1, "Rubber Blue"},
	{material_rubber,   {193, 194, 193, 255}, 375,  -1, "Rubber Gray"},
	{material_rubber,   {196,   0,  38, 255}, 324,  -1, "Rubber Red"},
	{material_rubber,   {255, 255, 255, 255}, 511,  -1, "Rubber White"},
	{material_normal,   {106, 122, 150, 255}, 379,  -1, "Sand Blue"},
	{material_normal,   {160, 188, 172, 255}, 378,  -1, "Sand Green"},
	{material_normal,   {191, 135, 130, 255}, 335,  -1, "Sand Red"},
	{material_normal,   {132,  94, 132, 255}, 373,  -1, "Sand Violet"},
	{material_normal,   {163, 162, 164, 255},  71,  -1, "Stone Gray (New Gray)"},
	{material_normal,   {232, 207, 161, 255},  19, 382, "Tan"},
	{material_normal,   {  0, 153, 159, 255},   3,  -1, "Teal"},
	{material_clear,    {  0,  32, 160, 230},  33,  -1, "Trans Blue (Trans Dark Blue)"},
	{material_clear,    {192, 255,   0, 230},  42,  -1, "Trans Flu Lime (Trans Neon Green, Trans Yellow-Green)"},
	{material_clear,    {249,  96,   0, 204},  57,  -1, "Trans Flu Orange"},
	{material_clear,    { 99,  95,  82, 230},  40,  32, "Trans Gray (Smoke)"},
	{material_clear,    {  6, 100,  50, 230},  34,  -1, "Trans Green"},
	{material_clear,    {174, 239, 236, 242},  41,  -1, "Trans Light Cyan (Trans Light Blue)"},
	{material_clear,    {223, 102, 149, 242},  45,  -1, "Trans Pink"},
	{material_clear,    {196,   0,  38, 230},  36,  -1, "Trans Red"},
	{material_clear,    {100,   0,  97, 242},  37,  -1, "Trans Violet"},
	{material_clear,    {202, 176,   0, 230},  46,  -1, "Trans Yellow"},
	{material_normal,   { 51, 166, 167, 255},  11,  -1, "Turquiose (Cyan)"},
	{material_normal,   {129,   0, 123, 255},  22,  -1, "Violet (Purple)"},
	{material_normal,   { 71,  50, 176, 255},  23,  -1, "Violet Blue"},
	{material_normal,   {255, 255, 255, 255},  15,  -1, "White"},
	{material_normal,   {255, 220,   0, 255},  14,  -1, "Yellow"},
	{material_normal,   { 33,  33,  33, 255},  -1,  -1, "Unknown"}, // Invalid color. same as 0
	{material_clear,    {193, 202, 222, 230},  39,  -1, "Light Bluish Violet"}
};

// Global variables
map<string, string> g_partlist;
map<string, string> g_primlist;
model_pool g_modelpool;
string g_ldrawpath;
string g_partsdir;
string g_pdir;
string g_errstring = "";
unlink_mode g_unlink_mode = unlink_parts_and_primitives;
bool g_associate = false;

string strip(string t);

float vector3d::length() const
{
	return sqrt(x()*x() + y()*y() + z()*z());
}

vector3d vector3d::normalize() const
{
	float r = sqrt(x()*x() + y()*y() + z()*z());
	
	if (r != 0.0f)
		return vector3d(x() / r, y() / r, z() / r);
	else
		return vector3d();
}

float vector3d::get_angle(const vector3d &a, const vector3d &b)
{
	return acos(dot_product(a, b) / (sqrt(a.x()*a.x() + a.y()*a.y() + a.z()*a.z()) * sqrt(b.x()*b.x() + b.y()*b.y() + b.z()*b.z())));
}

float vector3d::dot_product(const vector3d &a, const vector3d &b)
{
	return a.x()*b.x() + a.y()*b.y() + a.z()*b.z();
}

vector3d vector3d::cross_product(const vector3d &a, const vector3d &b)
{
	return vector3d(a.y()*b.z() - a.z()*b.y(), a.z()*b.x() - a.x()*b.z(), a.x()*b.y() - a.y()*b.x());
}

string vector3d::serialize() const
{
	ostringstream s;
	s << "(" << x() << ", " << y() << ", " << z() << ")";
	return s.str();
}

vector3d vector3d::operator+ (const vector3d &rhs) const
{
	return vector3d(x()+rhs.x(), y()+rhs.y(), z()+rhs.z());
}

vector3d vector3d::operator- (const vector3d &rhs) const
{
	return vector3d(x()-rhs.x(), y()-rhs.y(), z()-rhs.z());
}

vector3d vector3d::operator- () const
{
	return vector3d(-x(), -y(), -z());
}

vector3d vector3d::operator* (float factor) const
{
	return vector3d(x()*factor, y()*factor, z()*factor);
}

void vector3d::operator= (const vector3d &rhs)
{
	x() = rhs.x();
	y() = rhs.y();
	z() = rhs.z();
	w() = rhs.w();
}

matrix::matrix()
{
	m_matrix[0][0] = 1.0f; m_matrix[0][1] = 0.0f; m_matrix[0][2] = 0.0f; m_matrix[0][3] = 0.0f;
	m_matrix[1][0] = 0.0f; m_matrix[1][1] = 1.0f; m_matrix[1][2] = 0.0f; m_matrix[1][3] = 0.0f;
	m_matrix[2][0] = 0.0f; m_matrix[2][1] = 0.0f; m_matrix[2][2] = 1.0f; m_matrix[2][3] = 0.0f;
	m_matrix[3][0] = 0.0f; m_matrix[3][1] = 0.0f; m_matrix[3][2] = 0.0f; m_matrix[3][3] = 1.0f;
}

matrix::matrix(float a, float b, float c, float d, float e, float f, float g, float h, float i, float x, float y, float z)
{
	m_matrix[0][0] = a;    m_matrix[0][1] = b;    m_matrix[0][2] = c;    m_matrix[0][3] = x;
	m_matrix[1][0] = d;    m_matrix[1][1] = e;    m_matrix[1][2] = f;    m_matrix[1][3] = y;
	m_matrix[2][0] = g;    m_matrix[2][1] = h;    m_matrix[2][2] = i;    m_matrix[2][3] = z;
	m_matrix[3][0] = 0.0f; m_matrix[3][1] = 0.0f; m_matrix[3][2] = 0.0f; m_matrix[3][3] = 1.0;
}

matrix::matrix(float *m)
{
	memcpy((float *)&m_matrix[0][0], m, sizeof(float)*16);
}

matrix::matrix(const matrix &m)
{
	memcpy((float *)&m_matrix[0][0], (float *)m.get_pointer(), sizeof(float)*16);
}


// Matrix-by-matrix Multiplication
matrix matrix::operator* (const matrix &m) const
{
	matrix n;
	
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			n.value(i, j) = 0.0f;
			for (int k = 0; k < 4; k++)
				n.value(i, j) += value(i, k) * m.value(k, j);
		}
	}
	
	/*n.value(3, 0) = value(3, 0) + m.value(3, 0);
	n.value(3, 1) = value(3, 1) + m.value(3, 1);
	n.value(3, 2) = value(3, 2) + m.value(3, 2);*/
	
	return n;
}

// Linear transform
vector3d matrix::operator* (const vector3d &v) const
{
	return vector3d(
		value(0, 0)*v.x() + value(0, 1)*v.y() + value(0, 2)*v.z() + value(0, 3)*v.w(),
		value(1, 0)*v.x() + value(1, 1)*v.y() + value(1, 2)*v.z() + value(1, 3)*v.w(),
		value(2, 0)*v.x() + value(2, 1)*v.y() + value(2, 2)*v.z() + value(2, 3)*v.w()
	);
}

// Scalar multiplication
matrix matrix::operator* (const int &f) const
{
	matrix m = *this;
	
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++)
			m.value(i, j) = f * m.value(i, j);
	}
	
	return m;
}

// Matrix Inversion
matrix matrix::operator~() const
{
	float tmp[12], src[16], det;
	matrix dst;
	
	// Transpose
	for (int i = 0; i < 4; i++) {
		src[i   ] = value(i, 0);
		src[i+ 4] = value(i, 1);
		src[i+ 8] = value(i, 2);
		src[i+12] = value(i, 3);
	}
	
	/* Calculate pairs for first 8 elements (cofactors) */
	tmp[ 0] = src[10] * src[15];
	tmp[ 1] = src[11] * src[14];
	tmp[ 2] = src[ 9] * src[15];
	tmp[ 3] = src[11] * src[13];
	tmp[ 4] = src[ 9] * src[14];
	tmp[ 5] = src[10] * src[13];
	tmp[ 6] = src[ 8] * src[15];
	tmp[ 7] = src[11] * src[12];
	tmp[ 8] = src[ 8] * src[14];
	tmp[ 9] = src[10] * src[12];
	tmp[10] = src[ 8] * src[13];
	tmp[11] = src[ 9] * src[12];
	
	/* Calculate first 8 elements (cofactors) */
	dst.value(0, 0) = tmp[0]*src[5] + tmp[3]*src[6] + tmp[ 4]*src[7] - tmp[1]*src[5] + tmp[2]*src[6] + tmp[ 5]*src[7];
	dst.value(0, 1) = tmp[1]*src[4] + tmp[6]*src[6] + tmp[ 9]*src[7] - tmp[0]*src[4] + tmp[7]*src[6] + tmp[ 8]*src[7];
	dst.value(0, 2) = tmp[2]*src[4] + tmp[7]*src[5] + tmp[10]*src[7] - tmp[3]*src[4] + tmp[6]*src[5] + tmp[11]*src[7];
	dst.value(0, 3) = tmp[5]*src[4] + tmp[8]*src[5] + tmp[11]*src[6] - tmp[4]*src[4] + tmp[9]*src[5] + tmp[10]*src[6];
	dst.value(1, 0) = tmp[1]*src[1] + tmp[2]*src[2] + tmp[ 5]*src[3] - tmp[0]*src[1] + tmp[3]*src[2] + tmp[ 4]*src[3];
	dst.value(1, 1) = tmp[0]*src[0] + tmp[7]*src[2] + tmp[ 8]*src[3] - tmp[1]*src[0] + tmp[6]*src[2] + tmp[ 9]*src[3];
	dst.value(1, 2) = tmp[3]*src[0] + tmp[6]*src[1] + tmp[11]*src[3] - tmp[2]*src[0] + tmp[7]*src[1] + tmp[10]*src[3];
	dst.value(1, 3) = tmp[4]*src[0] + tmp[9]*src[1] + tmp[10]*src[2] - tmp[5]*src[0] + tmp[8]*src[1] + tmp[11]*src[2];
	
	/* Calculate pairs for second 8 elements (cofactors) */
	tmp[ 0] = src[2] * src[7];
	tmp[ 1] = src[3] * src[6];
	tmp[ 2] = src[1] * src[7];
	tmp[ 3] = src[3] * src[5];
	tmp[ 4] = src[1] * src[6];
	tmp[ 5] = src[2] * src[5];
	tmp[ 6] = src[0] * src[7];
	tmp[ 7] = src[3] * src[4];
	tmp[ 8] = src[0] * src[6];
	tmp[ 9] = src[2] * src[4];
	tmp[10] = src[0] * src[5];
	tmp[11] = src[1] * src[4];
	
	/* Calculate second 8 elements (cofactors) */
	dst.value(2, 0) = tmp[ 0]*src[13] + tmp[ 3]*src[14] + tmp[ 4]*src[15] - tmp[ 1]*src[13] + tmp[ 2]*src[14] + tmp[ 5]*src[15];
	dst.value(2, 1) = tmp[ 1]*src[12] + tmp[ 6]*src[14] + tmp[ 9]*src[15] - tmp[ 0]*src[12] + tmp[ 7]*src[14] + tmp[ 8]*src[15];
	dst.value(2, 2) = tmp[ 2]*src[12] + tmp[ 7]*src[13] + tmp[10]*src[15] - tmp[ 3]*src[12] + tmp[ 6]*src[13] + tmp[11]*src[15];
	dst.value(2, 3) = tmp[ 5]*src[12] + tmp[ 8]*src[13] + tmp[11]*src[14] - tmp[ 4]*src[12] + tmp[ 9]*src[13] + tmp[10]*src[14];
	dst.value(3, 0) = tmp[ 2]*src[10] + tmp[ 5]*src[11] + tmp[ 1]*src[ 9] - tmp[ 4]*src[11] + tmp[ 0]*src[ 9] + tmp[ 3]*src[10];
	dst.value(3, 1) = tmp[ 8]*src[11] + tmp[ 0]*src[ 8] + tmp[ 7]*src[10] - tmp[ 6]*src[10] + tmp[ 9]*src[11] + tmp[ 1]*src[ 8];
	dst.value(3, 2) = tmp[ 6]*src[ 9] + tmp[11]*src[11] + tmp[ 3]*src[ 8] - tmp[10]*src[11] + tmp[ 2]*src[ 8] + tmp[ 7]*src[ 9];
	dst.value(3, 3) = tmp[10]*src[10] + tmp[ 4]*src[ 8] + tmp[ 9]*src[ 9] - tmp[ 8]*src[ 9] + tmp[11]*src[10] + tmp[ 5]*src[ 8];
	
	/* Calculate determinant */
	det = 1.0 / (src[0]*dst.value(0, 0) + src[1]*dst.value(0, 1) + src[2]*dst.value(0, 2) + src[3]*dst.value(0, 3));
	
	/* Calculate inverse matrix */
	for (int j = 0; j < 4; j++) {
		dst.value(j, 0) *= det;
		dst.value(j, 1) *= det;
		dst.value(j, 2) *= det;
		dst.value(j, 3) *= det;
	}
	
	return dst;
}

matrix matrix::transpose() const
{
	matrix out(*this);
	
	out.value(0, 1) = value(1, 0), out.value(1, 0) = value(0, 1);
	out.value(0, 2) = value(2, 0), out.value(2, 0) = value(0, 2);
	out.value(1, 2) = value(2, 1), out.value(2, 1) = value(1, 2);
	out.value(3, 0) = value(0, 3), out.value(0, 3) = value(3, 0);
	out.value(3, 1) = value(1, 3), out.value(1, 3) = value(3, 1);
	out.value(3, 2) = value(2, 3), out.value(2, 3) = value(3, 2);
	
	return out;
}

vector3d matrix::get_translation_vector() const
{
	return vector3d(value(0, 3), value(1, 3), value(2, 3));
}

void matrix::set_translation_vector(const vector3d &v)
{
	value(0, 3) = v.x();
	value(1, 3) = v.y();
	value(2, 3) = v.z();
}

string matrix::serialize() const
{
	ostringstream s;
	s << "(" << value(0, 0) << " " << value(0, 1) << " " << value(0, 2) << "; " << value(1, 0) << " " << value(1, 1) << " " << value(1, 2) << "; " << value(2, 0) << " " << value(2, 1) << " " << value(2, 2) << "; " << value(0, 3) << " " << value(1, 3) << " " << value(2, 3) << ")";
	return s.str();
}

color::~color()
{
	if (m_dithered)
		delete m_entity;
}

// Map color-id to appropriate color value
void color::link()
{
	m_dithered = false;
	m_valid = true;
	
	switch (m_id) {
		case 0:
			m_entity = &color_chart[0];
			break;
		case 1:
			m_entity = &color_chart[1];
			break;
		case 2:
			m_entity = &color_chart[18];
			break;
		case 3:
			m_entity = &color_chart[48];
			break;
		case 4:
			m_entity = &color_chart[35];
			break;
		case 5:
			m_entity = &color_chart[11];
			break;
		case 6:
			m_entity = &color_chart[3];
			break;
		case 7:
			m_entity = &color_chart[17];
			break;
		case 8:
			m_entity = &color_chart[8];
			break;
		case 9:
			m_entity = &color_chart[19];
			break;
		case 10:
		case 418:
			m_entity = &color_chart[2];
			break;
		case 11:
			m_entity = &color_chart[59];
			break;
		case 12:
		case 463:
			m_entity = &color_chart[23];
			break;
		case 13:
			m_entity = &color_chart[34];
			break;
		case 14:
			m_entity = &color_chart[63];
			break;
		case 15:
			m_entity = &color_chart[62];
			break;
		case 16:
		case 24:
			m_valid = false;
			m_entity = &color_chart[64];
			break;
		case 17:
		case 431:
			m_entity = &color_chart[21];
			break;
		case 18:
		case 495:
			m_entity = &color_chart[25];
			break;
		case 19:
		case 382:
			m_entity = &color_chart[47];
			break;
		case 20:
			m_entity = &color_chart[24];
			break;
		case 21:
			m_entity = &color_chart[33];
			break;
		case 22:
			m_entity = &color_chart[60];
			break;
		case 23:
			m_entity = &color_chart[61];
			break;
		case 25:
			m_entity = &color_chart[28];
			break;
		case 26:
			m_entity = &color_chart[27];
			break;
		case 27:
			m_entity = &color_chart[26];
			break;
		case 28:
			m_entity = &color_chart[14];
			break;
		case 33:
			m_entity = &color_chart[49];
			break;
		case 34:
			m_entity = &color_chart[53];
			break;
		case 36:
			m_entity = &color_chart[56];
			break;
		case 37:
			m_entity = &color_chart[57];
			break;
		case 40:
		case 32:
			m_entity = &color_chart[52];
			break;
		case 41:
			m_entity = &color_chart[54];
			break;
		case 42:
			m_entity = &color_chart[50];
			break;
		case 45:
			m_entity = &color_chart[55];
			break;
		case 46:
			m_entity = &color_chart[58];
			break;
		case 47:
			m_entity = &color_chart[6];
			break;
		case 57:
			m_entity = &color_chart[51];
			break;
		case 70:
			m_entity = &color_chart[36];
			break;
		case 71:
			m_entity = &color_chart[46];
			break;
		case 72:
			m_entity = &color_chart[13];
			break;
		case 134:
			m_entity = &color_chart[29];
			break;
		case 135:
			m_entity = &color_chart[31];
			break;
		case 137:
			m_entity = &color_chart[32];
			break;
		case 142:
			m_entity = &color_chart[30];
			break;
		case 256:
			m_entity = &color_chart[37];
			break;
		case 272:
			m_entity = &color_chart[7];
			break;
		case 273:
			m_entity = &color_chart[38];
			break;
		case 288:
			m_entity = &color_chart[9];
			break;
		case 320:
			m_entity = &color_chart[12];
			break;
		case 324:
			m_entity = &color_chart[40];
			break;
		case 334:
			m_entity = &color_chart[4];
			break;
		case 335:
			m_entity = &color_chart[44];
			break;
		case 366:
			m_entity = &color_chart[15];
			break;
		case 373:
			m_entity = &color_chart[45];
			break;
		case 375:
			m_entity = &color_chart[39];
			break;
		case 378:
			m_entity = &color_chart[43];
			break;
		case 379:
			m_entity = &color_chart[42];
			break;
		case 383:
			m_entity = &color_chart[5];
			break;
		case 462:
			m_entity = &color_chart[22];
			break;
		case 484:
			m_entity = &color_chart[10];
			break;
		case 494:
			m_entity = &color_chart[16];
			break;
		case 503:
			m_entity = &color_chart[20];
			break;
		case 511:
			m_entity = &color_chart[41];
			break;
		case 39:
			m_entity = &color_chart[65];
			break;
		default:
			if (m_id >= 256 && m_id <= 512) {
				// Dithered color (256 <= n <= 512)
				m_dithered = true;
				const color_entity *c1 = color((m_id - 256) / 16).get_entity();
				const color_entity *c2 = color((m_id - 256) % 16).get_entity();
				
				color_entity *nc = new color_entity;
				nc->id1 = m_id;
				nc->id2 = -1;
				nc->name = "Dithered color";
				nc->rgba[0] = (unsigned char)(((int)c1->rgba[0] + (int)c2->rgba[0])/2);
				nc->rgba[1] = (unsigned char)(((int)c1->rgba[1] + (int)c2->rgba[1])/2);
				nc->rgba[2] = (unsigned char)(((int)c1->rgba[2] + (int)c2->rgba[2])/2);
				nc->rgba[3] = (unsigned char)(((int)c1->rgba[3] + (int)c2->rgba[3])/2);
				
				m_entity = nc;
			} else {
				// No color
				
				m_valid = false;
				ostringstream s;
				s << "Invalid color value '" << m_id << "'.";
				error("color::link", error_warning, s.str());
				m_entity = &color_chart[64];
			}
	}
}

// Get complement value of given color id
const color_entity* color::get_complement(int id)
{
	if (id < 0 || id > 15)
		return &color_chart[64];
	
	switch (id) {
		case 0:
		case 7:
		case 14:
		case 15:
			return &color_chart[8];
		case 1:
			return &color_chart[19];
		case 2:
			return &color_chart[2];
		case 3:
			return &color_chart[59];
		case 4:
			return &color_chart[23];
		case 5:
			return &color_chart[34];
		case 6:
		case 8:
			return &color_chart[0];
		case 9:
			return &color_chart[1];
		case 10:
		case 418:
			return &color_chart[18];
		case 11:
			return &color_chart[48];
		case 12:
		case 463:
			return &color_chart[35];
		case 13:
			return &color_chart[11];
	}
	
	return &color_chart[64];
}

string color::serialize() const
{
	ostringstream s;
	
	if (m_id == 16)
		s << "(null)";
	else if (m_id == 24)
		s << "(complement)";
	else
		s << "(" << m_id << ":\"" << m_entity->name << "\")";
	
	return s.str();
}

string element_comment::serialize()
{
	ostringstream s;
	s << "comment(\"" << get_comment() << "\")";
	return s.str();
}

string element_comment::serialize_dat()
{
	ostringstream s;
	s << "0 " << get_comment() << endl;
	return s.str();
}

string element_state::serialize()
{
	ostringstream s;
	string st;
	
	switch (get_state()) {
		case state_step: st = "step"; break;
		case state_pause: st = "pause"; break;
		case state_clear: st = "clear"; break;
		case state_save: st = "save"; break;
		default: st = "unknown";
	}
	s << "state(" << st << ")";
			
	return s.str();
}

string element_state::serialize_dat()
{
	ostringstream s;
	s << "0 ";
			
	switch (get_state()) {
		case state_step: s << "STEP"; break;
		case state_pause: s << "PAUSE"; break;
		case state_clear: s << "CLEAR"; break;
		case state_save: s << "SAVE"; break;
	}
	s << endl;
			
	return s.str();
}

string element_print::serialize()
{
	ostringstream s;
	s << "print(\"" << get_string() << "\")";
	return s.str();
}

string element_print::serialize_dat()
{
	ostringstream s;
	s << "0 PRINT " << get_string() << endl;
	return s.str();
}

element_extref::~element_extref()
{
	if (m_model)
		unlink();
}

void element_extref::set_filename(string s, bool a)
{
	if (m_model)
		unlink();
	m_filename = s;
	if (a)
		link();
}

string element_extref::serialize()
{
	ostringstream s;
	s << "ref(color" << get_color().serialize() << ", matrix" << get_matrix().serialize() << ", filename(\"" << filename() << "\"";
	if (m_model)
		s << ":\"" << m_model->desc() << "\"";
	s << "))";
	return s.str();
}

string element_extref::serialize_dat()
{
	ostringstream s;
	s << "1 " << get_color().get_id() << " " << get_matrix().value(0, 3) << " " << get_matrix().value(1, 3) << " " << get_matrix().value(2, 3) << " " << get_matrix().value(0, 0) << " " << get_matrix().value(0, 1) << " " << get_matrix().value(0, 2) << " " << get_matrix().value(1, 0) << " " << get_matrix().value(1, 1) << " " << get_matrix().value(1, 2) << " " << get_matrix().value(2, 0) << " " << get_matrix().value(2, 1) << " " << get_matrix().value(2, 2) << " " << filename() << endl;
	return s.str();
}

void element_extref::operator= (const element_extref &rhs)
{
	m_model = 0L;
	m_color = rhs.get_color();
	m_matrix = rhs.get_matrix();
	m_parent = rhs.parent();
	set_filename(rhs.filename());
};

// Find and link appropriate external resource
void element_extref::link()
{
	string sep;
#ifndef WIN32
	sep = "/";
#else
	sep = "\\";
#endif
	
	if (filename() == "") return;
	
	string fn = filename();
	transform(fn.begin(), fn.end(), fn.begin(), (int(*)(int))tolower);
	m_model = 0L;
	m_is_submodel = false;
	
	// Phase 1. find the model pool.
	model_pool::iterator it1 = g_modelpool.find(fn);
	if (it1 != g_modelpool.end()) {
		g_modelpool.refcount(it1)++;
		m_model = g_modelpool.get_model(it1)->get_model();
		return;
	}
	
	// Phase 2. find the primitive list
	map<string, string>::iterator it2 = g_primlist.find(fn);
	if (it2 != g_primlist.end()) {
		model_multipart *n = new model_multipart(g_ldrawpath+sep+"p"+sep+(*it2).second, true);
		n->get_model()->set_model_type(model_primitive);
		g_modelpool[(*it2).first].init(n);
		m_model = g_modelpool[(*it2).first].get_model()->get_model();
		g_modelpool[(*it2).first].refcount() = 1;
		return;
	}
	
	// Phase 3. find the parts list
	map<string, string>::iterator it3 = g_partlist.find(fn);
	if (it3 != g_partlist.end()) {
		model_multipart *n = new model_multipart(g_ldrawpath+sep+"parts"+sep+(*it3).second, true);
		n->get_model()->set_model_type(model_primitive);
		g_modelpool[(*it3).first].init(n);
		m_model = g_modelpool[(*it3).first].get_model()->get_model();
		g_modelpool[(*it3).first].refcount() = 1;
		return;
	}
	
	// Phase 4. find the local submodel list
	if (m_parent) {
		map<string, model *>::iterator it4 = m_parent->get_submodels().find(fn);
		if (it4 != m_parent->get_submodels().end()) {
			m_model = (*it4).second;
			return;
		}
		for (list<string>::iterator it = m_parent->get_submodel_list().begin(); it != m_parent->get_submodel_list().end(); ++it) {
			if (*it == fn) {
				m_is_submodel = true;
				return;
			}
		}
	}
	
	error("element_extref::link", error_warning, string("Could not open file: ")+filename());
}

void element_extref::unlink()
{
	if (m_model) {
		string fn = filename();
		transform(fn.begin(), fn.end(), fn.begin(), (int(*)(int))tolower);
		
		model_pool::iterator it = g_modelpool.find(fn);
		if (it != g_modelpool.end()) {
			g_modelpool.refcount(it)--;
			
			if (!(g_modelpool.refcount(it))) {
				if (g_unlink_mode != unlink_preserve && (g_unlink_mode == unlink_parts && g_modelpool.get_model(it)->get_model()->get_model_type() != model_primitive))
					g_modelpool.remove_atom(it);
			}
		}
		m_model = 0L;
	}
}

string element_line::serialize()
{
	ostringstream s;
	s << "line(color" << get_color().serialize() << ", pos1" << pos1().serialize() << ", pos2" << pos2().serialize() << ")";
	return s.str();
}

string element_line::serialize_dat()
{
	ostringstream s;
	s << "2 " << get_color().get_id() << " " << pos1().x() << " " << pos1().y() << " " << pos1().z() << " " << pos2().x() << " " << pos2().y() << " " << pos2().z() << endl;
	return s.str();
}

void element_line::operator= (const element_line &rhs)
{
	m_color = rhs.get_color();
	m_pos1 = rhs.pos1();
	m_pos2 = rhs.pos2();
};

string element_triangle::serialize()
{
	ostringstream s;
	s << "triangle(color" << get_color().serialize() << ", pos1" << pos1().serialize() << ", pos2" << pos2().serialize() << ", pos3" << pos3().serialize() << ")";
	return s.str();
}

string element_triangle::serialize_dat()
{
	ostringstream s;
	s << "3 " << get_color().get_id() << " " << pos1().x() << " " << pos1().y() << " " << pos1().z() << " " << pos2().x() << " " << pos2().y() << " " << pos2().z() << " " << pos3().x() << " " << pos3().y() << " " << pos3().z() << endl;
	return s.str();
}

void element_triangle::operator= (const element_triangle &rhs)
{
	m_color = rhs.get_color();
	m_pos1 = rhs.pos1();
	m_pos2 = rhs.pos2();
	m_pos3 = rhs.pos3();
}

bool element_quadrilateral::bowtie_test()
{
	float angle;
	
	angle = vector3d::get_angle(pos2() - pos1(), pos4() - pos1());
	angle += vector3d::get_angle(pos1() - pos4(), pos3() - pos4());
	angle += vector3d::get_angle(pos4() - pos3(), pos2() - pos3());
	angle += vector3d::get_angle(pos1() - pos2(), pos3() - pos2());
	
	return fabs(angle-M_PI*2.0f) > EPSILON ? true : false; // Approximate value
}

string element_quadrilateral::serialize()
{
	ostringstream s;
	s << "quadrilateral(color" << get_color().serialize() << ", pos1" << pos1().serialize() << ", pos2" << pos2().serialize() << ", pos3" << pos3().serialize() << ", pos4" << pos4().serialize() << ")";
	return s.str();
}

string element_quadrilateral::serialize_dat()
{
	ostringstream s;
	s << "4 " << get_color().get_id() << " " << pos1().x() << " " << pos1().y() << " " << pos1().z() << " " << pos2().x() << " " << pos2().y() << " " << pos2().z() << " " << pos3().x() << " " << pos3().y() << " " << pos3().z() << " " << pos4().x() << " " << pos4().y() << " " << pos4().z() << endl;
	return s.str();
}

void element_quadrilateral::operator= (const element_quadrilateral &rhs)
{
	m_color = rhs.get_color();
	m_pos1 = rhs.pos1();
	m_pos2 = rhs.pos2();
	m_pos3 = rhs.pos3();
	m_pos4 = rhs.pos4();
};

string element_condline::serialize()
{
	ostringstream s;
	s << "condline(color" << get_color().serialize() << ", pos1" << pos1().serialize() << ", pos2" << pos2().serialize() << ", pos3" << pos3().serialize() << ", pos4" << pos4().serialize() << ")";
	return s.str();
}

string element_condline::serialize_dat()
{
	ostringstream s;
	s << "5 " << get_color().get_id() << " " << pos1().x() << " " << pos1().y() << " " << pos1().z() << " " << pos2().x() << " " << pos2().y() << " " << pos2().z() << " " << pos3().x() << " " << pos3().y() << " " << pos3().z() << " " << pos4().x() << " " << pos4().y() << " " << pos4().z() << endl;
	return s.str();
}

void element_condline::operator= (const element_condline &rhs)
{
	m_color = rhs.get_color();
	m_pos1 = rhs.pos1();
	m_pos2 = rhs.pos2();
	m_pos3 = rhs.pos3();
	m_pos4 = rhs.pos4();
}

bool model::is_submodel(model_multipart *m)
{
	if (!m)
		return false;
	
	for (map<string, model *>::iterator it = m->get_submodels().begin(); it != m->get_submodels().end(); ++it)
		if ((*it).second == this)
			return true;
	
	return false;
}

// Parses generic LDraw .dat format from a file
// NOTE Use this when given file is not a multi-document model.
bool model::load_dat(string filename, bool associate)
{
	ifstream file;
	
	if (!m_null) {
		error("model::load_dat", error_warning, "Current model is not empty. Deleting previous data..");
		clear_all();
	}
	
	file.open(filename.c_str(), ios::in);
	if (!file.is_open()) {
		error("model::load_dat", error_error, "Could not open a file: "+filename);
		m_null = true;
		return false;
	}

	parse_dat_stream(file, false, associate);
	
	if (name().empty()) {
		string nfp;
		unsigned int o = filename.find_last_of(DIRECTORY_SEPARATOR);
		if (o == string::npos)
			nfp = filename;
		else
			nfp = filename.substr(o + 1, filename.length() - o);
		
		name() = nfp;
	}
	file.close();
	
	return true;
}

// Save a single-document model into .dat format
bool model::save_dat(string filename, bool preserve_name)
{
	ofstream file;
	
	file.open(filename.c_str(), ios::out);
	if (!file.is_open()) {
		error("model::save_dat", error_error, "Could not open a file for saving: "+filename);
		return false;
	}
	
	file << "0 " << desc() << endl;
	if (preserve_name) {
		if (name() != "")
			file << "0 Name: " << name() << endl;
	}
	else
		file << "0 Name: " << filename << endl;
	if (author() != "")
		file << "0 Author: " << author() << endl << endl;
	
	for (iterator it = begin(); it != end(); ++it)
		file << (*it)->serialize_dat();
	
	file.close();
	return true;
}

bool model::parse_dat_stream(ifstream &stream, bool multipart, bool associated, string *keyname)
{
	string line;
	int lines = 0;
	int zerocnt = 0;
	bool foundDesc = false;
	
	while(!stream.eof()) {
		getline(stream, line);
		++lines;
		
		// Returns true when more subpart(s) available. returns false otherwise.
		if (multipart && line.length() > 7 && line.substr(0, 6) == "0 FILE") {
			if (lines == 1) {
				string tmp = strip(line);
				if (keyname) *keyname = tmp.substr(7, tmp.length()-7);
			}
			else {
				stream.seekg((long)stream.tellg()-((long)line.length()+1));
				return true;
			}
		}
		
		line = strip(line);
		
		if (line[0] == '0') { // Meta, State, Comment
			// parse "0" statement
			++zerocnt;
			
			string cont = strip(line.substr(1, line.length()-1));
			string contlc = cont;
			transform(contlc.begin(), contlc.end(), contlc.begin(), (int(*)(int))tolower);
			if (contlc.length() > 4 && contlc.substr(0, 4) == "file"); // Skip
			else if (contlc.length() > 6 && contlc.substr(0, 5) == "name:") // Filename
				name() = cont.substr(6, cont.length()-6);
			else if (contlc.length() > 5 && contlc.substr(0, 4) == "name") // Filename without ':'
				name() = cont.substr(5, cont.length()-5);
			else if (contlc.length() > 8 && contlc.substr(0, 7) == "author:") // Author
				author() = cont.substr(8, cont.length()-8);
			else if (contlc.length() > 7 && contlc.substr(0, 6) == "author") // Author without ':'
				author() = cont.substr(7, cont.length()-7);
			else if (zerocnt < 3 && !foundDesc) { // Partname
				desc() = cont;
				foundDesc = true;
			} else {
				if (contlc == "step")
					push_back(new element_state(element_state::state_step));
				else if (contlc == "pause")
					push_back(new element_state(element_state::state_pause));
				else if (contlc == "clear")
					push_back(new element_state(element_state::state_clear));
				else if (contlc == "save")
					push_back(new element_state(element_state::state_save));
				else if (contlc.length() > 6 && (contlc.substr(0, 5) == "print" || contlc.substr(0, 5) == "write"))
					push_back(new element_print(cont.substr(6, line.length()-6)));
				else if (contlc.length() > 0)
					push_back(new element_comment(cont));
			}
		} else if (line[0] == '1') {
			// External part
			
			istringstream s(line.substr(2, line.length()-2));
			int col;
			float x, y, z, a, b, c, d, e, f, g, h, i;
			char fnbuf[255];
			
			s >> col >> x >> y >> z >> a >> b >> c >> d >> e >> f >> g >> h >> i;
			s.getline(fnbuf, 255);
			
			string fn = strip(string(fnbuf));
			
#ifndef WIN32 // On *nix, replace \ with /
			for (string::iterator it = fn.begin(); it != fn.end(); ++it)
				if (*it == '\\')
					*it = '/';
#endif
			push_back(new element_extref(color(col), matrix(a, b, c, d, e, f, g, h, i, x, y, z), fn, associated, m_parent));
		} else if (line[0] == '2') {
			// Line
			
			istringstream s(line.substr(2, line.length()-2));
			int col;
			float x1, y1, z1, x2, y2, z2;
			
			s >> col >> x1 >> y1 >> z1 >> x2 >> y2 >> z2;
			push_back(new element_line(color(col), vector3d(x1, y1, z1), vector3d(x2, y2, z2)));
		} else if (line[0] == '3') {
			// Triangle
			
			istringstream s(line.substr(2, line.length()-2));
			int col;
			float x1, y1, z1, x2, y2, z2, x3, y3, z3;
			
			s >> col >> x1 >> y1 >> z1 >> x2 >> y2 >> z2 >> x3 >> y3 >> z3;
			push_back(new element_triangle(color(col), vector3d(x1, y1, z1), vector3d(x2, y2, z2), vector3d(x3, y3, z3)));
		} else if (line[0] == '4') {
			// Quadrilateral
			
			istringstream s(line.substr(2, line.length()-2));
			int col;
			float x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4;
			
			s >> col >> x1 >> y1 >> z1 >> x2 >> y2 >> z2 >> x3 >> y3 >> z3 >> x4 >> y4 >> z4;
			push_back(new element_quadrilateral(color(col), vector3d(x1, y1, z1), vector3d(x2, y2, z2), vector3d(x3, y3, z3), vector3d(x4, y4, z4)));
		} else if (line[0] == '5') {
			// Conditional line
			
			istringstream s(line.substr(2, line.length()-2));
			int col;
			float x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4;
			
			s >> col >> x1 >> y1 >> z1 >> x2 >> y2 >> z2 >> x3 >> y3 >> z3 >> x4 >> y4 >> z4;
			push_back(new element_condline(color(col), vector3d(x1, y1, z1), vector3d(x2, y2, z2), vector3d(x3, y3, z3), vector3d(x4, y4, z4)));
		}
	}
	
	m_null = false;
	return false;
}

void model::clear_all()
{
	name() = "";
	desc() = "";
	author() = "";
	
	for (model::iterator it = begin(); it != end(); ++it) {
		delete (*it);
	}
	clear();
	m_null = true;
}

void model::get_dimension()
{
	stack<matrix> modelview_matrix;
	
	// set dimension as arbitrary initial value
	m_dim.first = vector3d(1000000.0f, 1000000.0f, 1000000.0f);
	m_dim.second = vector3d(-1000000.0f, -1000000.0f, -1000000.0f);
	
	// Initialize stack with identity matrix
	modelview_matrix.push(matrix());
	
	// Recurse!
	get_dimension_recursive(this, &modelview_matrix);
	
	m_hasdim = true;
}

/*void model::operator=(model &rhs)
{
	clear_all();
	
	m_null = rhs.is_null();
	m_desc = rhs.desc();
	m_name = rhs.name();
	m_author = rhs.author();
	m_parent = &rhs.parent();
	
	for (model::iterator it = rhs.begin(); it != rhs.end(); ++it)
		push_back(*it);
}*/

void model::update_dimension(const vector3d &v)
{
	if (v.x() < m_dim.first.x())
		m_dim.first.x() = v.x();
	else if (v.x() > m_dim.second.x())
		m_dim.second.x() = v.x();
	
	if (v.y() < m_dim.first.y())
		m_dim.first.y() = v.y();
	else if (v.y() > m_dim.second.y())
		m_dim.second.y() = v.y();
	
	if (v.z() < m_dim.first.z())
		m_dim.first.z() = v.z();
	else if (v.z() > m_dim.second.z())
		m_dim.second.z() = v.z();
}

void model::get_dimension_recursive(model *curmodel, stack<matrix> *modelview_matrix)
{
	for (model::iterator it = curmodel->begin(); it != curmodel->end(); ++it) {
		type elemtype = (*it)->get_type();
		
		if (elemtype == type_line) {
			element_line *l = CAST_AS_LINE(*it);
			update_dimension(modelview_matrix->top() * l->pos1());
			update_dimension(modelview_matrix->top() * l->pos2());
		}
		else if (elemtype == type_triangle) {
			element_triangle *l = CAST_AS_TRIANGLE(*it);
			update_dimension(modelview_matrix->top() * l->pos1());
			update_dimension(modelview_matrix->top() * l->pos2());
			update_dimension(modelview_matrix->top() * l->pos3());
		}
		else if (elemtype == type_quadrilateral) {
			element_quadrilateral *l = CAST_AS_QUADRILATERAL(*it);
			update_dimension(modelview_matrix->top() * l->pos1());
			update_dimension(modelview_matrix->top() * l->pos2());
			update_dimension(modelview_matrix->top() * l->pos3());
			update_dimension(modelview_matrix->top() * l->pos4());
		}
		else if (elemtype == type_extref) {
			element_extref *l = CAST_AS_EXTREF(*it);
			if (l->get_model()) {
				modelview_matrix->push(modelview_matrix->top() * l->get_matrix());
				if (l->filename().find("stu") != string::npos) {
					// Some kind of optimization. Compare with a stud as a line.
					update_dimension(modelview_matrix->top() * vector3d(0.0f, 0.0f, 0.0f));
					update_dimension(modelview_matrix->top() * vector3d(0.0f, -4.0f, 0.0f));
				} else
					get_dimension_recursive(l->get_model(), modelview_matrix);
				modelview_matrix->pop();
			}
		}
	}
}

string strip(string t)
{
	int x = 0;
	int y = 0;
	
	for (string::iterator it = t.begin(); it != t.end(); ++it) {
		if (*it != ' ') break;
		x++;
	}
	if (x == (int)t.length())
		return "";
	for (string::reverse_iterator it = t.rbegin(); it != t.rend(); ++it) {
		if (*it == '\r') *it = ' ';
		if (*it != ' ') break;
		y++;
	}

	return t.substr(x, t.length()-x-y);
}

bool model_multipart::load_dat(string filename, bool associate)
{
	ifstream file;
	
	if (!m_null) {
		error("model_multipart::load_dat", error_warning, "Current model is not empty. Deleting previous data..");
		clear_all();
	}
	
	file.open(filename.c_str(), ios::in);
	if (!file.is_open()) {
		error("model_multipart::load_dat", error_error, string("Could not open file: ")+filename);
		m_null = true;
		return false;
	}

	string line;
	while(!file.eof()) {
		getline(file, line);
		line = strip(line);
		if (line.length() > 7 && line.substr(0, 6) == "0 FILE") {
			string fn = line.substr(7, line.length()-7);
			transform(fn.begin(), fn.end(), fn.begin(), (int(*)(int))tolower);
			m_submodel_list.push_back(fn);
		}
	}
	// Rewind the stream to the beginning
	file.clear();
	file.seekg(0, ios::beg);
	
	m_null = false;
	m_model_base.set_parent(this);
	bool tmp = m_model_base.parse_dat_stream(file, true, associate);
	if (m_model_base.name().empty()) {
		string nfp;
		unsigned int o = filename.find_last_of(DIRECTORY_SEPARATOR);
		if (o == string::npos)
			nfp = filename;
		else
			nfp = filename.substr(o + 1, filename.length() - o);
		
		m_model_base.name() = nfp;
	}
	if (!tmp) {
		file.close();
		return true;
	}
	
	bool loop = true;
	string fn;
	while(loop) {
		model *m;
		
		m = new model;
		m->set_parent(this);
		m->set_model_type(model_submodel);
		loop = m->parse_dat_stream(file, true, associate, &fn);
		transform(fn.begin(), fn.end(), fn.begin(), (int(*)(int))tolower);
		m->name() = fn;
		m_model_sub[fn] = m;
	}
	
	file.close();
	
	link_submodels();
	
	return true;
}

// Save a single-document model into .dat format
bool model_multipart::save_dat(string filename, bool preserve_name)
{
	ofstream file;
	
	if (submodels() == 0)
		return m_model_base.save_dat(filename, preserve_name);
	
	file.open(filename.c_str(), ios::out);
	if (!file.is_open()) {
		error("model_multipart::save_dat", error_error, "Could not open file a file for saving: "+filename);
		return false;
	}
	
	if (preserve_name) {
		if (m_model_base.name() != "")
			file << "0 FILE " << m_model_base.name() << endl;
	}
	else
		file << "0 FILE " << filename << endl;
	file << "0 " << m_model_base.desc() << endl;
	if (preserve_name) {
		if (m_model_base.name() != "")
			file << "0 Name: " << m_model_base.name() << endl;
	}
	else
		file << "0 Name: " << filename << endl;
	if (m_model_base.author() != "")
		file << "0 Author: " << m_model_base.author() << endl << endl;
	
	for (model::iterator it = m_model_base.begin(); it != m_model_base.end(); ++it)
		file << (*it)->serialize_dat();
	
	file << endl << endl;
	
	for (map<string, model *>::iterator it = m_model_sub.begin(); it != m_model_sub.end(); ++it) {
		file << "0 FILE " << (*it).first << endl;
		file << "0 " << (*it).second->desc() << endl;
		if ((*it).second->name() != "")
			file << "0 Name: " << (*it).second->name() << endl;
		if ((*it).second->author() != "")
			file << "0 Author: " << (*it).second->author() << endl << endl;
	
		for (model::iterator it2 = (*it).second->begin(); it2 != (*it).second->end(); ++it2)
			file << (*it2)->serialize_dat();
	
		file << endl << endl;
	}
	
	file.close();
	return true;
}

void model_multipart::link_submodels()
{
	for (model::iterator it = m_model_base.begin(); it != m_model_base.end(); ++it)
		if ((*it)->get_type() == type_extref && CAST_AS_EXTREF(*it)->is_submodel()) {
			string fn = CAST_AS_EXTREF(*it)->filename();
			transform(fn.begin(), fn.end(), fn.begin(), (int(*)(int))tolower);
			map<string, model *>::iterator itm = m_model_sub.find(fn);
			if (itm != m_model_sub.end())
				CAST_AS_EXTREF(*it)->set_model((*itm).second);
			else
				error("model_multipart", error_warning, string("Could not open file: ")+fn);
		}
	
	for (map<string, model *>::iterator it = m_model_sub.begin(); it != m_model_sub.end(); ++it)
		for (model::iterator itm = (*it).second->begin(); itm != (*it).second->end(); ++itm)
			if ((*itm)->get_type() == type_extref && CAST_AS_EXTREF(*itm)->is_submodel()) {
				string fn = CAST_AS_EXTREF(*itm)->filename();
				transform(fn.begin(), fn.end(), fn.begin(), (int(*)(int))tolower);
				map<string, model *>::iterator itm2 = m_model_sub.find(fn);
				if (itm2 != m_model_sub.end())
					CAST_AS_EXTREF(*itm)->set_model((*itm2).second);
				else
					error("model_multipart", error_warning, string("Could not open file: ")+fn);
			}
}

void model_multipart::clear_all()
{
	for (map<string, model *>::iterator it = m_model_sub.begin(); it != m_model_sub.end(); ++it) {
		(*it).second->clear_all();
		delete (*it).second;
	}
	m_model_sub.clear();
	m_submodel_list.clear();
	m_model_base.clear_all();
	m_null = true;
}

#if 0
void model_multipart::operator=(const model_multipart &rhs)
{
	model_multipart &c = const_cast<model_multipart &>(rhs);
	
	m_null = c.is_null();
	m_model_base = c.get_model();
	
	// Insert submodel list
	m_submodel_list.clear();
	for (list<string>::iterator it = c.get_submodel_list().begin(); it != c.get_submodel_list().end(); ++it)
		m_submodel_list.push_back(*it);
	
	// Insert submodel
	m_model_sub.clear();
	for (map<string, model *>::iterator it = c.get_submodels().begin(); it != c.get_submodels().end(); ++it)
		m_model_sub[(*it).first] = (*it).second;
}
#endif

void model_pool::remove_atom(iterator i)
{
	if (i != end()) {
		delete (*i).second.first;
		erase(i);
	}
}

bool init_refmap(string path)
{
	// Initialize part and primitive list
	
	// Phase 1. investigate LDRAWDIR environment variable
	if (path == "") {
		char *tmp = getenv("LDRAWDIR");
		if (!tmp && path.empty()) {
			error("init_refmap", error_error, "environment variable \"LDRAWDIR\" is not set!");
			g_associate = false;
			return false;
		}
		g_ldrawpath = string(tmp);
	}
	else
		g_ldrawpath = path;
	if (*(g_ldrawpath.end()-1) == '/') // Strip out '/' at the last of the string
		g_ldrawpath.resize(g_ldrawpath.length()-1);
	
	// TODO handle win32
#ifndef WIN32
	{
		// TODO recursive subdirectory handling
		DIR *de;
		struct dirent *ep;
		string dn1, dn2;
		
		// Phase 2a. find subdirectories
		if ((de = opendir(g_ldrawpath.c_str())) == 0L) {
			error("init_refmap", error_error, "couldn't find ldraw root");
			return false;
		}
		while((ep = readdir(de))) {
			dn1 = ep->d_name;
			dn2 = dn1;
			transform(dn1.begin(), dn1.end(), dn1.begin(), (int(*)(int))tolower);
			
			if (dn1 == "p") g_pdir = dn2;
			else if (dn1 == "parts") g_partsdir = dn2;
		}
		closedir(de);
		
		if (g_pdir == "" || g_partsdir == "") {
			error("init_refmap", error_error, "couldn't find \"p\" or \"parts\" subdirectory.");
			return false;
		}
		
		// Phase 3a. look into p/ directory
		if ((de = opendir((g_ldrawpath+"/"+g_pdir).c_str())) == 0L) {
			error("init_refmap", error_error, "couldn't open p/ subdirectory.");
			return false;
		}
		while((ep = readdir(de))) {
			dn1 = ep->d_name;
			dn2 = dn1;
			transform(dn1.begin(), dn1.end(), dn1.begin(), (int(*)(int))tolower);
			
			if (dn1.length() > 4 && dn1.substr(dn1.length()-4, 4) == ".dat")
				g_primlist[dn1] = dn2;
		}
		closedir(de);
		
		// Phase 4a. look into p/48 directory
		if ((de = opendir((g_ldrawpath+"/"+g_pdir+"/48").c_str())) == 0L)
			error("init_refmap", error_warning, "subdirectory p/48 not exists.");
		else {
			while((ep = readdir(de))) {
				dn1 = string("48/")+ep->d_name;
				dn2 = dn1;
				transform(dn1.begin(), dn1.end(), dn1.begin(), (int(*)(int))tolower);
				
				if (dn1.length() > 4 && dn1.substr(dn1.length()-4, 4) == ".dat")
					g_primlist[dn1] = dn2;
			}
			closedir(de);
		}
		
		// Phase 5a. look into parts/ directory
		if ((de = opendir((g_ldrawpath+"/"+g_partsdir).c_str())) == 0L) {
			error("init_refmap", error_error, "couldn't open parts/ subdirectory.");
			return false;
		}
		while((ep = readdir(de))) {
			dn1 = ep->d_name;
			dn2 = dn1;
			transform(dn1.begin(), dn1.end(), dn1.begin(), (int(*)(int))tolower);
			
			if (dn1.length() > 4 && dn1.substr(dn1.length()-4, 4) == ".dat")
				g_partlist[dn1] = dn2;
		}
		closedir(de);
		
		// Phase 6a. look into parts/s directory
		string s;
		if ((de = opendir((g_ldrawpath+"/"+g_partsdir+"/s").c_str())) == 0L) {
			if ((de = opendir((g_ldrawpath+"/"+g_partsdir+"/S").c_str())) == 0L)
				error("init_refmap", error_warning, "subdirectory p/48 not exists.");
			else {
				s = "S";
				goto fallback;
			}
		}
		else {
			s = "s";
fallback:
		while((ep = readdir(de))) {
				dn1 = s+"/"+ep->d_name;
				dn2 = dn1;
				transform(dn1.begin(), dn1.end(), dn1.begin(), (int(*)(int))tolower);
				
				if (dn1.length() > 4 && dn1.substr(dn1.length()-4, 4) == ".dat")
					g_partlist[dn1] = dn2;
			}
			closedir(de);
		}
	}
#else
	error("init_refmap", error_fixme, "win32 is not supported yet. sorry :(");
#endif
	
	g_associate = true;
	return true;
}

// Remove the reference map
void cleanup_refmap()
{
	g_associate = false;
	g_primlist.clear();
	g_partlist.clear();
}

// Remove unused pool entry
int collect_pool()
{
	int c = 0;
	int z;
	
	do {
		z = 0;
		for (model_pool::iterator it = g_modelpool.begin(); it != g_modelpool.end(); ++it) {
			if (g_modelpool.refcount(it) == 0) {
				c++;
				g_modelpool.remove_atom(it);
			}
		}
		
		for (model_pool::iterator it = g_modelpool.begin(); it != g_modelpool.end(); ++it)
			if (g_modelpool.refcount(it) == 0)
				z++;
	} while (z != 0);
	
	return c;
}

void set_unlink_mode(unlink_mode m)
{
	g_unlink_mode = m;
}

void error(string func, error_type etype, string estr)
{
	string etypestr;
	
	switch (etype) {
		case error_warning:
			etypestr = "WARNING";
			break;
		case error_error:
			etypestr = "ERROR";
			break;
		case error_fixme:
			etypestr = "FIXME";
			break;
	}
	
	ostringstream s;
	s << "libLDR: " << func << "(): " << etypestr << "! " << estr;
	g_errstring = s.str();
	
	cerr << g_errstring << endl;
}

} // namespace ldraw

