/* libLDR: Portable and easy-to-use LDraw format abstraction & I/O reference library *
 * To obtain more information about LDraw, visit http://www.ldraw.org.               *
 * Distributed in terms of the General Public License v2                             *
 *                                                                                   *
 * Author: (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>                    */

#ifndef _LIBLDR_H_
#define _LIBLDR_H_

#include <iostream>
#include <cctype>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <locale>

#include <algorithm>
#include <typeinfo>
#include <list>
#include <map>
#include <utility>
#include <stack>

#include <cstdlib>
#include <cstring>

#define CAST_AS_COMMENT(c)         (dynamic_cast<ldraw::element_comment         *>(c))
#define CAST_AS_STATE(c)           (dynamic_cast<ldraw::element_state           *>(c))
#define CAST_AS_PRINT(c)           (dynamic_cast<ldraw::element_print           *>(c))
#define CAST_AS_EXTREF(c)          (dynamic_cast<ldraw::element_extref          *>(c))
#define CAST_AS_LINE(c)            (dynamic_cast<ldraw::element_line            *>(c))
#define CAST_AS_TRIANGLE(c)        (dynamic_cast<ldraw::element_triangle        *>(c))
#define CAST_AS_QUADRILATERAL(c)   (dynamic_cast<ldraw::element_quadrilateral   *>(c))
#define CAST_AS_CONDLINE(c)        (dynamic_cast<ldraw::element_condline        *>(c))

#define EPSILON 1e-4

#ifdef WIN32
#define DIRECTORY_SEPARATOR '\\'
#else
#define DIRECTORY_SEPARATOR '/'
#endif

namespace ldraw
{

class model;
class model_multipart;
//class submodel_context;

// Element types used throughout libLDR
enum type
{
	type_null,          // Null type
	type_comment,       // Comment
	type_state,         // State-related
	type_print,         // State-related (print)
	type_extref,        // External file reference
	type_line,          // Line
	type_triangle,      // Triangle
	type_quadrilateral, // Quadrilateral
	type_condline       // Conditional line
};

// Enumerations
enum error_type
{
	error_warning,
	error_error,
	error_fixme
};

enum unlink_mode
{
	unlink_preserve,
	unlink_parts,
	unlink_parts_and_primitives
};

enum model_type
{
	model_primitive,
	model_part,
	model_submodel,
	model_external_file,
	model_general
};

bool init_refmap(std::string path = "");
void cleanup_refmap();
int collect_pool();

void set_unlink_mode(unlink_mode m);
void error(std::string func, error_type etype, std::string estr);

enum material_type {
	material_normal, material_clear, material_metallic, material_rubber
};

struct color_entity
{
	material_type material;
	unsigned char rgba[4]; // RGBA Array
	int id1; // LDraw Color ID
	int id2; // Alternate Color ID
	std::string name; // Name String
};

// Space vector
class vector3d
{
	public:
		vector3d() { m_array[0] = 0.0f, m_array[1] = 0.0f, m_array[2] = 0.0f; }
		vector3d(float x, float y, float z, float w = 1.0f) { m_array[0] = x, m_array[1] = y, m_array[2] = z; m_array[3] = w; }
		vector3d(const vector3d &p) { m_array[0] = p.x(), m_array[1] = p.y(), m_array[2] = p.z(), m_array[3] = p.w(); }
		~vector3d() {}
		
		float& x() { return m_array[0]; }
		const float& x() const { return m_array[0]; }
		float& y() { return m_array[1]; }
		const float& y() const { return m_array[1]; }
		float& z() { return m_array[2]; }
		const float& z() const { return m_array[2]; }
		// w component of quaternion
		float& w() { return m_array[3]; }
		const float& w() const { return m_array[3]; }
		const float* get_pointer() const { return m_array; }
		
		float length() const;
		vector3d normalize() const;
		static float get_angle(const vector3d &a, const vector3d &b);
		static float dot_product(const vector3d &a, const vector3d &b);
		static vector3d cross_product(const vector3d &a, const vector3d &b);
		
		std::string serialize() const;
		
		vector3d operator+ (const vector3d &rhs) const;
		vector3d operator- (const vector3d &rhs) const;
		vector3d operator- () const;
		vector3d operator* (float factor) const; // Scalar multiplication
		float& operator[] (int n) { return m_array[n]; }
		const float& operator[] (int n) const { return m_array[n]; }
		
		void operator= (const vector3d &rhs);
		
	private:
		float m_array[4];
};

// Generic 3D-space transformation matrix
class matrix
{
	public:
		matrix();
		matrix(float a, float b, float c, float d, float e, float f, float g, float h, float i, float x = 0.0f, float y = 0.0f, float z = 0.0f);
		matrix(float *m);
		matrix(const matrix &m);
		~matrix() {}
		
		matrix operator* (const matrix &m) const; // 4x4 * 4x4
		vector3d operator* (const vector3d &v) const; // 4x4 * 4x1
		matrix operator* (const int &f) const; // Scalar multiplication
		matrix operator~ () const; // Inversion
		matrix transpose() const;

		vector3d get_translation_vector() const;
		void set_translation_vector(const vector3d &v);
		
		float& value(int r, int c) { return m_matrix[r][c]; }
		const float& value(int r, int c) const { return m_matrix[r][c]; }
		const float* get_pointer() const { return &m_matrix[0][0]; }
		
		std::string serialize() const;
		
		void operator= (const matrix &rhs) { memcpy(&m_matrix, rhs.get_pointer(), sizeof(float)*16); }
		
	private:
		float m_matrix[4][4];
};

// Represents a color.
class color
{
	public:
		color() : m_valid(true), m_id(0) { link(); }
		color(int id) : m_id(id) { link();}
		color(const color &c) : m_id(c.get_id()) { link(); }
		~color();
		
		void operator=(int rhs) { m_id = rhs; link(); }
		void operator=(const color &rhs) { m_id = rhs.get_id(); link(); }
		
		int get_id() const { return m_id; }
		void set_id(int i) { m_id = i; link(); }
		
		bool is_valid() { return m_valid; }
		bool is_null() { return m_id == 16 || m_id == 24; }
		const color_entity* get_entity() const { return m_entity; }
		static const color_entity* get_complement(int id);
		const color_entity* get_complement() const { return get_complement(m_id); }
		
		std::string serialize() const;
	
	private:
		void link();
		
		bool m_dithered;
		bool m_valid;
		int m_id;
		const color_entity *m_entity;
};

// Parent class of every element classes.
class element_base
{
	// Actually, do nothing
	public:
		element_base() {}
		virtual ~element_base() {}
		
		virtual type get_type() { return type_null; }
		virtual std::string serialize() { return ""; }
		virtual std::string serialize_dat() { return ""; }
		virtual std::string serialize_ldx() { return ""; }
};

// Comment. does nothing
class element_comment : public element_base
{
	public:
		element_comment(std::string s) : m_str(s) {}
		element_comment(const element_comment &c) : element_base(), m_str(c.get_comment()) {}
		~element_comment() {}
		
		std::string& comment() { return m_str; }
		const std::string& get_comment() const { return m_str; }
		
		type get_type() { return type_comment; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_comment &rhs) { m_str = rhs.get_comment(); }
		
	private:
		std::string m_str;
};

// status definition
class element_state : public element_base
{
	public:
		enum state { state_step, state_pause, state_clear, state_save };
		
		element_state(state s) : element_base(), m_state(s) {}
		element_state(const element_state &s) : element_base(), m_state(s.get_state()) {}
		~element_state() {}
		
		state& get_state() { return m_state; }
		const state& get_state() const { return m_state; }
		
		type get_type() { return type_state; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_state &rhs) { m_state = rhs.get_state(); }
		
	private:
		state m_state;
};

// Can be used for printing a string on the screen.
class element_print : public element_base
{
	public:
		element_print(std::string s) : element_base(), m_str(s) {}
		element_print(const element_print &p) : element_base(), m_str(p.get_string()) {}
		~element_print() {}
		
		std::string& get_string() { return m_str; }
		const std::string& get_string() const { return m_str; }
		
		type get_type() { return type_print; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_print &rhs) { m_str = rhs.get_string(); }
		
	private:
		std::string m_str;
};

// External reference. (parts, primitives, submodels)
class element_extref : public element_base
{
	public:
		element_extref(color c, matrix m, std::string fn, bool a = true, model_multipart *mp = 0L) :
			element_base(), m_color(c), m_matrix(m), m_filename(fn), m_model(0L), m_parent(mp)
		{ if (a) link(); }
		element_extref(const element_extref &e) :
			element_base(), m_color(e.get_color()), m_matrix(e.get_matrix()), m_model(0L), m_parent(e.parent()) { set_filename(e.filename()); }
		~element_extref();
		
		const color& get_color() const { return m_color; }
		const matrix& get_matrix() const { return m_matrix; }
		const std::string& filename() const { return m_filename; }
		model* get_model() const { return m_model; }
		model_multipart* parent() const { return m_parent; }
		const bool& is_submodel() const { return m_is_submodel; }
		
		void set_color(color c) { m_color = c; }
		void set_matrix(matrix m) { m_matrix = m; }
		void set_filename(std::string s, bool a = true);
		void set_model(model *m) { m_model = m; }
		void set_parent(model_multipart *p) { m_parent = p; }
		
		type get_type() { return type_extref; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_extref &rhs);
		
	private:
		void link();
		void unlink();
		
		color m_color;
		matrix m_matrix;
		std::string m_filename;
		model *m_model;
		model_multipart *m_parent;
		bool m_is_submodel;
};

// Line
class element_line : public element_base
{
	public:
		element_line(color c, vector3d p1, vector3d p2) :
			element_base(), m_color(c), m_pos1(p1), m_pos2(p2) {}
		element_line(const element_line &l) :
			element_base(), m_color(l.get_color()), m_pos1(l.pos1()), m_pos2(l.pos2()) {}
		~element_line() {}
		
		color& get_color() { return m_color; }
		const color& get_color() const { return m_color; }
		vector3d& pos1() { return m_pos1; }
		const vector3d& pos1() const { return m_pos1; }
		vector3d& pos2() { return m_pos2; }
		const vector3d& pos2() const { return m_pos2; }
		
		type get_type() { return type_line; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_line &rhs);
		
	private:
		color m_color;
		vector3d m_pos1;
		vector3d m_pos2;
};

// Triangle
class element_triangle : public element_base
{
	public:
		element_triangle(color c, vector3d p1, vector3d p2, vector3d p3) :
			element_base(), m_color(c), m_pos1(p1), m_pos2(p2), m_pos3(p3) {}
		element_triangle(const element_triangle &t) :
			element_base(), m_color(t.get_color()), m_pos1(t.pos1()), m_pos2(t.pos2()), m_pos3(t.pos3()) {}
		~element_triangle() {}
		
		color& get_color() { return m_color; }
		const color& get_color() const { return m_color; }
		vector3d& pos1() { return m_pos1; }
		const vector3d& pos1() const { return m_pos1; }
		vector3d& pos2() { return m_pos2; }
		const vector3d& pos2() const { return m_pos2; }
		vector3d& pos3() { return m_pos3; }
		const vector3d& pos3() const { return m_pos3; }
		
		type get_type() { return type_triangle; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_triangle &rhs);
		
	private:
		color m_color;
		vector3d m_pos1, m_pos2, m_pos3;
};

// Quadrilateral
class element_quadrilateral : public element_base
{
	public:
		element_quadrilateral(color c, vector3d p1, vector3d p2, vector3d p3, vector3d p4) :
			element_base(), m_color(c), m_pos1(p1), m_pos2(p2), m_pos3(p3), m_pos4(p4) {}
		element_quadrilateral(const element_quadrilateral &q) :
			element_base(), m_color(q.get_color()), m_pos1(q.pos1()), m_pos2(q.pos2()), m_pos3(q.pos3()), m_pos4(q.pos4()) {}
		~element_quadrilateral() {}
		
		color& get_color() { return m_color; }
		const color& get_color() const { return m_color; }
		vector3d& pos1() { return m_pos1; }
		const vector3d& pos1() const { return m_pos1; }
		vector3d& pos2() { return m_pos2; }
		const vector3d& pos2() const { return m_pos2; }
		vector3d& pos3() { return m_pos3; }
		const vector3d& pos3() const { return m_pos3; }
		vector3d& pos4() { return m_pos4; }
		const vector3d& pos4() const { return m_pos4; }
		
		bool bowtie_test();
		
		type get_type() { return type_quadrilateral; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_quadrilateral &rhs);
		
	private:
		
		color m_color;
		vector3d m_pos1, m_pos2, m_pos3, m_pos4;
};

// Conditional line.
class element_condline : public element_base
{
	public:
		element_condline(color c, vector3d p1, vector3d p2, vector3d p3, vector3d p4) :
			element_base(), m_color(c), m_pos1(p1), m_pos2(p2), m_pos3(p3), m_pos4(p4) {}
		element_condline(const element_condline &l) :
			element_base(), m_color(l.get_color()), m_pos1(l.pos1()), m_pos2(l.pos2()), m_pos3(l.pos3()), m_pos4(l.pos4()) {}
		~element_condline() {}
		
		color& get_color() { return m_color; }
		const color& get_color() const { return m_color; }
		vector3d& pos1() { return m_pos1; }
		const vector3d& pos1() const { return m_pos1; }
		vector3d& pos2() { return m_pos2; }
		const vector3d& pos2() const { return m_pos2; }
		vector3d& pos3() { return m_pos3; }
		const vector3d& pos3() const { return m_pos3; }
		vector3d& pos4() { return m_pos4; }
		const vector3d& pos4() const { return m_pos4; }
		
		type get_type() { return type_condline; }
		std::string serialize();
		std::string serialize_dat();
		
		void operator= (const element_condline &rhs);
		
	private:
		color m_color;
		vector3d m_pos1, m_pos2, m_pos3, m_pos4;
};

// Represents a single-part model.
class model : public std::list<element_base *>
{
	public:
		model() : m_null(true), m_parent(0L), m_hasdim(false), m_model_type(model_general) {}
		model(std::string filename, bool associate = false, model_multipart *parent = 0L)
			: m_null(true), m_parent(parent), m_hasdim(false), m_model_type(model_general)
			{ load_dat(filename, associate); }
		model(std::string desc, std::string name, std::string author, model_multipart *parent = 0L)
			: m_desc(desc), m_name(name), m_author(author), m_null(false), m_parent(parent), m_hasdim(false), m_model_type(model_general) {}
		~model() { clear_all(); }
		
		bool is_null() { return m_null; }
		
		bool has_dimension() { return m_hasdim; }
		bool is_submodel(model_multipart *m);
		
		model_type get_model_type() { return m_model_type; }
		void set_model_type(model_type t) { m_model_type = t; }
		
		std::string& desc() { return m_desc; }
		const std::string& desc() const { return m_desc; }
		std::string& name() { return m_name; }
		const std::string& name() const { return m_name; }
		std::string& author() { return m_author; }
		const std::string& author() const { return m_author; }
		model_multipart* parent() { return m_parent; }
		std::pair<vector3d, vector3d>& dimension() { return m_dim; }
		const std::pair<vector3d, vector3d>& dimension() const { return m_dim; }
		
		void set_parent(model_multipart *parent) { m_parent = parent; }
		
		bool load_dat(std::string filename, bool associate = false);
		bool save_dat(std::string filename, bool preserve_name = false);
		bool parse_dat_stream(std::ifstream &stream, bool multipart, bool associated, std::string *keyname = 0L);
		void clear_all();
		
		void get_dimension();
		
		//void operator=(const model &rhs) const;
		
	private:
		void update_dimension(const vector3d &v);
		void get_dimension_recursive(model *curmodel, std::stack<matrix> *modelview_matrix);
		
		std::string m_desc;
		std::string m_name;
		std::string m_author;
		
		bool m_null;
		model_multipart *m_parent;
		
		bool m_hasdim;
		std::pair<vector3d, vector3d> m_dim;
		
		model_type m_model_type;
};

// Multi-part model.
class model_multipart
{
	public:
		model_multipart() : m_null(true) { m_model_base.set_parent(this); }
		model_multipart(std::string filename, bool associate = false) : m_null(true)
			{ load_dat(filename, associate); }
		~model_multipart() { clear_all(); }
		
		int submodels() { return m_model_sub.size(); }
		
		model* get_model() { return &m_model_base; }
		std::map<std::string, model *>& get_submodels() { return m_model_sub; }
		std::list<std::string>& get_submodel_list() { return m_submodel_list; }
		
		bool is_null() { return m_null; }
		bool load_dat(std::string filename, bool associate = false);
		bool save_dat(std::string filename, bool preserve_name = false);
		void link_submodels();
		void clear_all();
		
		//void operator=(const model_multipart &rhs);
		
	private:
		bool m_null;
		
		model m_model_base;
		std::map<std::string, model *> m_model_sub;
		std::list<std::string> m_submodel_list;
};

// Element class for model_pool
class model_pool_atom : public std::pair<model_multipart *, int>
{
	public:
		model_pool_atom() {}
		~model_pool_atom() {}
		
		void init(model_multipart *m) { first = m; }
		int& refcount() { return second; }
		model_multipart* get_model() { return first; }
};

// Model pool
class model_pool : public std::map<std::string, model_pool_atom>
{
	public:
		model_pool() {}
		~model_pool() {}
		
		const std::string& key(iterator i) const { return (*i).first; }
		model_pool_atom& atom(iterator i) { return (*i).second; }
		
		int& refcount(iterator i) { return (*i).second.refcount(); }
		model_multipart* get_model(iterator i) { return (*i).second.get_model(); }
		
		void remove_atom(iterator i);
};

extern std::map<std::string, std::string> g_partlist;
extern std::map<std::string, std::string> g_primlist;
extern model_pool g_modelpool;
extern std::string g_ldrawpath;
extern std::string g_partsdir;
extern std::string g_pdir;
extern std::string g_errstring;
extern bool g_associate;
extern unlink_mode g_unlink_mode;
extern const color_entity color_chart[];
const int max_colors = 65;

} // namespace ldraw

#endif
