// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <klocale.h>
#include <kpopupmenu.h>

#include "mainwindow.h"
#include "libldr.h"
#include "document.h"
#include "transformdialog.h"

#include "list.h"

using namespace ldraw;

/******************************************************************************
 * KLDrawList
 ******************************************************************************/

KLDrawList *KLDrawList::instance_ = 0L;

KLDrawList::KLDrawList(QWidget *parent, const char *name) : KListView(parent, name)
{
	instance_ = this;
	
	selectedItem_ = 0L;
	activeDocument_ = KLDrawMainWindow::self()->activeDocument();
	
	modifyDialog_ = new KLDrawTransformDialog();
	
	colNo_ = addColumn(i18n("No."), 50);
	colItem_ = addColumn(i18n("Item"));
	
	setColumnAlignment(colNo_, Qt::AlignHCenter);
	setAllColumnsShowFocus(true);
	setFullWidth(true);
	setSorting(-1);
	
	connect(this, SIGNAL(selectionChanged()), this, SLOT(selectionChanged()));
	connect(this, SIGNAL(contextMenuRequested(QListViewItem *, const QPoint &, int)), this, SLOT(contextMenu(QListViewItem *, const QPoint &, int)));
	
	updateList();
}

KLDrawList::~KLDrawList()
{
	clear();
}

void KLDrawList::clearSelection()
{
	KListView::clearSelection();
	
	emit cursorChanged(0L);
}

element_extref* KLDrawList::lastRef() const
{
	KLDrawListItem *it;
	
	for (it = dynamic_cast<KLDrawListItem *>(selectedItem()); it; it = dynamic_cast<KLDrawListItem *>(it->itemAbove())) {
		if (it->item()->get_type() == type_extref)
			return CAST_AS_EXTREF(it->item());
	}
	
	return 0L;
}

void KLDrawList::updateList()
{
	clear();
	
	if (!(*activeDocument_)) return;
	
	model &curModel = *(*activeDocument_)->getActiveModel();
	
	int i = 0;
	KLDrawListItem *past = 0L;
	KLDrawListItem *newItem;
	for (model::iterator it = curModel.begin(); it != curModel.end(); ++it) {
		if (!past)
			newItem = new KLDrawListItem(this, i++, *it);
		else
			newItem = new KLDrawListItem(this, past, i++, *it);
		past = newItem;
	}
	
	setSelected(lastItem(), true);
	selectionChanged();
	
	if (curModel.size() == 0) {
		emit enableAction(0, false);
		emit enableAction(1, false);
	} else {
		if (currentItem() == firstChild())
			emit enableAction(0, false);
		else
			emit enableAction(0, true);
		
		if (currentItem() == lastItem())
			emit enableAction(1, false);
		else
			emit enableAction(1, true);
	}
}

void KLDrawList::selectionChanged()
{
	KLDrawListItem *i = dynamic_cast<KLDrawListItem *>(selectedItem());
	
	if (!i || i->item()->get_type() != type_extref)
		if (modifyDialog_->isShown())
			modifyDialog_->close();
	
	modifyDialog_->setCurrentItem(i);
	
	selectedItem_ = i;
	if (!i) return;
	
	ensureVisible(0, itemPos(i)+i->height());
	
	emit cursorChanged(dynamic_cast<KLDrawListItem *>(currentItem()));
}

void KLDrawList::goFirst()
{
	setSelected(firstChild(), true);
	
	emit enableAction(0, false);
	
	if (currentItem() == lastItem())
		emit enableAction(1, false);
	else
		emit enableAction(1, true);
	
	emit cursorChanged(dynamic_cast<KLDrawListItem *>(currentItem()));
}

void KLDrawList::goPrev()
{
	if (currentItem() && currentItem()->itemAbove()) {
		if (currentItem() == lastItem())
			emit enableAction(1, true);
		
		setSelected(currentItem()->itemAbove(), true);
		
		if (currentItem() == firstChild())
			emit enableAction(0, false);
		
		emit cursorChanged(dynamic_cast<KLDrawListItem *>(currentItem()));
	}
}

void KLDrawList::goNext()
{
	if (currentItem() && currentItem()->itemBelow()) {
		if (currentItem() == firstChild())
			emit enableAction(0, true);
		
		setSelected(currentItem()->itemBelow(), true);
		
		if (currentItem() == lastItem())
			emit enableAction(1, false);
		
		emit cursorChanged(dynamic_cast<KLDrawListItem *>(currentItem()));
	}
}

void KLDrawList::goLast()
{
	setSelected(lastItem(), true);
	
	emit enableAction(1, false);
	
	if (currentItem() == firstChild())
		emit enableAction(0, false);
	else
		emit enableAction(0, true);
	
	emit cursorChanged(dynamic_cast<KLDrawListItem *>(currentItem()));
}

void KLDrawList::contextMenu(QListViewItem *item, const QPoint &pos, int)
{
	KLDrawListItem *i = dynamic_cast<KLDrawListItem *>(item);
	
	if (!i) return;
	
	KPopupMenu menu(this);
	
	switch (i->item()->get_type()) {
		case type_extref:
			menu.insertItem(i18n("Edit Matrix"), 0);
		default:
			break;
	}
	
	switch (menu.exec(pos)) {
		case 0:
			modifyDialog_->setCurrentItem(dynamic_cast<KLDrawListItem *>(item));
			modifyDialog_->show();
			break;
	}
}

/******************************************************************************
 * KLDrawListItem
 ******************************************************************************/

KLDrawListItem::KLDrawListItem(KLDrawList *parent, int number, element_base *item)
	: KListViewItem(parent)
{	
	id_ = number;
	item_ = item;
	
	updateVisibleInfo();
}

KLDrawListItem::KLDrawListItem(KLDrawList *parent, KLDrawListItem *after, int number, element_base *item)
	: KListViewItem(parent, after)
{	
	id_ = number;
	item_ = item;
	
	updateVisibleInfo();
}

void KLDrawListItem::increaseFollowing(int n)
{
	KLDrawListItem *i;
	
	for (i = this; i != 0L; i = dynamic_cast<KLDrawListItem *>(i->itemBelow())) {
		i->setId(i->id() + n);
		i->updateVisibleInfo();
	}
}

void KLDrawListItem::swapItem(KLDrawListItem *i)
{
	element_base *tmpelem = i->item();
	int tmpid = i->id();
	
	i->setItem(item());
	i->setId(id());
	setItem(tmpelem);
	setId(tmpid);
	
	updateVisibleInfo();
}

void KLDrawListItem::updateVisibleInfo()
{
	setText(0, QString("%1").arg(id_ + 1));
	setText(1, QString::fromLocal8Bit(item_->serialize().c_str()));
}

/* Represent type of this item by colors
 * Color legend:
 *  Metadata : White
 *  Primitives : Green
 *  Part (or primitive part) reference : Red
 *  Submodel reference : Blue
 */
void KLDrawListItem::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align)
{
	QColorGroup _cg = cg;
	QColor rc;
	QListView *lv = listView();
	
	ldraw::type t = item_->get_type();
	
	if (t == ldraw::type_line          || t == ldraw::type_triangle ||
	    t == ldraw::type_quadrilateral || t == ldraw::type_condline) {
		rc = QColor(198, 255, 195);
	} else if (t == ldraw::type_extref) {
		ldraw::element_extref *o = CAST_AS_EXTREF(item_);
		if (o->get_model() && o->get_model()->is_submodel(o->get_model()->parent()))
			rc = QColor(218, 228, 255);
		else if (!o->get_model()) { /* unresolved model */
			rc = QColor(0, 0, 0);
			_cg.setColor(QColorGroup::Text, QColor(255, 255, 0));
		}
		else
			rc = QColor(255, 211, 215);
	} else {
		rc = QColor(255, 255, 255);
	}
	
	_cg.setColor((lv->viewport()->backgroundMode() == Qt::FixedColor) ?
		QColorGroup::Background : QColorGroup::Base, rc);
	QListViewItem::paintCell(p, _cg, column, width, align);
}
