// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <cstdlib>
#include <unistd.h>

#include <qvbox.h>
#include <qsplitter.h>
#include <qimage.h>
#include <kapplication.h>
#include <kfiledialog.h>
#include <kfileitem.h>
#include <kio/netaccess.h>
#include <qlayout.h>
#include <klocale.h>
#include <kaction.h>
#include <kmessagebox.h>
#include <kstdaction.h>
#include <ktabbar.h>
#include <ktabwidget.h>

#include "colorpicker.h"
#include "kldrawconfig.h"
#include "renderwidget.h"
#include "dbmanager.h"
#include "document.h"
#include "editor.h"
#include "list.h"
#include "libldr.h"
#include "modelproperties.h"
#include "partlist.h"
#include "partrenderer.h"
#include "kldraw.h"
#include "partrenderer.h"
#include "submodellist.h"
#include "transformdialog.h"

// XXX
#include "povrayrenderwidget.h"
#include "povrayrenderparameters.h"

#include "mainwindow.h"

using namespace std;
using namespace ldraw;

KLDrawMainWindow *KLDrawMainWindow::instance_ = 0L;

KLDrawMainWindow::KLDrawMainWindow(QWidget *parent, const char *name) : KMainWindow(parent, name)
{
	instance_ = this;
	
	blankFiles_ = 0;
	activeDocument_ = 0L;
	cursor_ = 0L;
	dragResources_.dragging = false;
	dragResources_.dragId = -1;
	
	// Initialize libLDR
	if (!initializeLDraw()) {
		KMessageBox::error(this, i18n("<qt>Failed to find LDraw root directory. Check your installation.<br />KLDraw requires <i>LDraw</i> part library. To obtain more information about LDraw, Visit <i>http://www.ldraw.org</i>.</qt>"));
		close();
		return;
	}
	ldraw::set_unlink_mode(unlink_parts);
	
	db_ = new KLDrawDBManager();
	editor_ = new KLDrawEditor(this);
	modelPropertiesWidget_ = new KLDrawModelProperties;
	
	setupLayout();
	setupConnections();
	initializeMenus();
	setupGUI();
	
	connect(colorEditAction->getWidget(), SIGNAL(colorPicked(const ldraw::color_entity *)), editor_, SLOT(changeColor(const ldraw::color_entity *)));
	
	clearWState(WState_Polished);
	
	changeCaption();
	activate(false);
}

KLDrawMainWindow::~KLDrawMainWindow()
{
	delete db_;
	
	ldraw::cleanup_refmap();
	
	openRecentAction->saveEntries(kapp->config());
	KLDrawConfig::self()->writeConfig();
}

void KLDrawMainWindow::newFile()
{
	addDocument(new KLDrawDocument(++blankFiles_, this));
	
	if (tabBar_->count() == 1)
		emit editingEnabled(true);
}

void KLDrawMainWindow::openFile()
{
	KURL url = KFileDialog::getOpenURL(QString::null,
		QString("*.ldr *.mpd *.dat|") + i18n("LDraw Model Files (*.ldr *.mpd *.dat)") + "\n" +
		"*|" + i18n("All Files"), this
	);
	
	if (url.isEmpty()) return;
	
	openRecentAction->addURL(url);
	loadURL(url);
}

void KLDrawMainWindow::openRecent(const KURL &url)
{
	loadURL(url);
}

void KLDrawMainWindow::saveFile()
{
	if (!activeDocument_) return;
	
	saveDocument(activeDocument_);
}

void KLDrawMainWindow::saveFileAs()
{
	if (!activeDocument_) return;
	
	saveDocumentAs(activeDocument_);
}

void KLDrawMainWindow::closeFile()
{
	if (!activeDocument_) return; // nothing to do
	
	if (!checkTab(activeDocument_)) // do not close if canceled
		return;
	
	documentList_.remove(activeDocument_);
	tabBar_->removeTab(activeDocument_);
	
	if (tabBar_->count() > 0)
		changeActiveSelection(tabBar_->currentTab());
	else {
		activeDocument_ = 0L;
		cursor_ = 0L;
		emit documentChanged();
		emit activeModelChanged();
		emit editingEnabled(false);
	}
}

void KLDrawMainWindow::changeActiveSelection(int id)
{
	activeDocument_ = dynamic_cast<KLDrawDocument *>(tabBar_->tab(id));
	
	emit documentChanged();
	emit activeModelChanged();
	
	changeCursor(dynamic_cast<KLDrawListItem *>(listWidget_->lastItem()));
}

void KLDrawMainWindow::activate(bool b)
{
	for (int i = 0; i < 4; ++i)
		renderWidget_[i]->setAcceptDrops(b);
	
	listWidget_->setEnabled(b);
	partList_->setEnabled(b);
	submodelList_->setEnabled(b);
	
	saveAction->setEnabled(b);
	saveAsAction->setEnabled(b);
	
	rotateXCWAction->setEnabled(b);
	rotateXCCWAction->setEnabled(b);
	rotateYCWAction->setEnabled(b);
	rotateYCCWAction->setEnabled(b);
	rotateZCWAction->setEnabled(b);
	rotateZCCWAction->setEnabled(b);
	
	viewFirstAction->setEnabled(b);
	viewPrevAction->setEnabled(b);
	viewNextAction->setEnabled(b);
	viewLastAction->setEnabled(b);
	zoomResetAction->setEnabled(b);
	perspResetAction->setEnabled(b);
	
	deleteAction->setEnabled(b);
	editModelAction->setEnabled(b);
	colorEditAction->setEnabled(b);
	
	submodelNewAction->setEnabled(b);
	submodelModifyAction->setEnabled(b);
	submodelDeleteAction->setEnabled(b);
}

void KLDrawMainWindow::changeCaption()
{
	if (!activeDocument_)
		setCaption(i18n("KLDraw"));
	else {
		setCaption(i18n("%1 - %2%3%4").arg(QString(activeDocument_->getModel()->get_model()->name().c_str())).arg(QString(activeDocument_->getModel()->get_model()->desc().c_str())).arg(activeDocument_->canSave() ? i18n(" [modified]"):"").arg(activeDocument_->readOnly() ? i18n(" [read only]"):""));
	}
}

void KLDrawMainWindow::resetZoom()
{
	if (activeDocument_)
		activeDocument_->recalibrateScreenDimension();
	
	emit refresh();
}

void KLDrawMainWindow::resetPersp()
{
	if (activeDocument_)
		activeDocument_->resetPerspective();
	
	emit refresh();
}

void KLDrawMainWindow::changeActiveModel(QListViewItem *item)
{
	KLDrawSubmodelItem *i = dynamic_cast<KLDrawSubmodelItem *>(item);
	
	// TODO is this model is in this document?
	if (activeDocument_) {
		activeDocument_->setActiveModel(i->getModel());
		activeDocument_->recalibrateScreenDimension();
		
		emit activeModelChanged();
		
		changeCursor(dynamic_cast<KLDrawListItem *>(listWidget_->lastItem()));
	}
}

void KLDrawMainWindow::enableAction(int id, bool enable)
{	
	switch (id) {
		case 0:
			viewFirstAction->setEnabled(enable);
			viewPrevAction->setEnabled(enable);
			return;
			
		case 1:
			viewLastAction->setEnabled(enable);
			viewNextAction->setEnabled(enable);
			return;
			
	}
}

void KLDrawMainWindow::changeCursor(KLDrawListItem *item)
{
	cursor_ = item;
}

void KLDrawMainWindow::openModelPropertiesDialog()
{
	modelPropertiesWidget_->setModel(activeDocument_->getActiveModel());
	
	modelPropertiesWidget_->show();
}

void KLDrawMainWindow::synchronizeViews()
{
	renderWidget_[0]->repaint();
	renderWidget_[1]->repaint();
	renderWidget_[2]->repaint();
	renderWidget_[3]->repaint();
}

void KLDrawMainWindow::dispatchDragStart(const ldraw::vector3d &bboxmin, const ldraw::vector3d &bboxmax)
{
	dragResources_.dragBBoxMin = bboxmin;
	dragResources_.dragBBoxMax = bboxmax;
	dragResources_.dragging = true;
	
	element_extref *elem = KLDrawList::instance()->lastRef();
	if (elem == 0L) {
		dragResources_.lastMatrix = matrix();
		dragResources_.lastColor = color(0);
	} else {
		dragResources_.lastMatrix = elem->get_matrix();
		dragResources_.lastColor = elem->get_color();
	}
	
	renderWidget_[0]->dragStart();
	renderWidget_[1]->dragStart();
	renderWidget_[2]->dragStart();
	renderWidget_[3]->dragStart();
}

void KLDrawMainWindow::dispatchDragStop()
{
	renderWidget_[0]->dragStop();
	renderWidget_[1]->dragStop();
	renderWidget_[2]->dragStop();
	renderWidget_[3]->dragStop();
}

void KLDrawMainWindow::submodelNew()
{
	sorry();
}

void KLDrawMainWindow::submodelModify()
{
	// Exit if current model is main model
	if (!activeDocument_ || activeDocument_->getActiveModel() == activeDocument_->getModel()->get_model())
		return;
	
	sorry();
}

void KLDrawMainWindow::submodelDelete()
{
	// Exit if current model is main model
	if (!activeDocument_ || activeDocument_->getActiveModel() == activeDocument_->getModel()->get_model())
		return;
	
	sorry();
}

void KLDrawMainWindow::render()
{
	KLDrawPOVRayRenderParameters *n = new KLDrawPOVRayRenderParameters();
	(void)new KLDrawPOVRayRenderWidget(n, *activeDocument_, 0L);
}

bool KLDrawMainWindow::initializeLDraw()
{
	if (ldraw::init_refmap()) // First attempt: guess ldraw path from $LDRAWDIR environment variable
		return true;
	else if (ldraw::init_refmap("/usr/share/ldraw")) // Second attempt: /usr/share/ldraw
		return true;
	else if (ldraw::init_refmap("/usr/share/LDRAW")) // Third attempt: /usr/share/LDRAW (uppercase)
		return true;
	else if (ldraw::init_refmap("/usr/local/share/LDRAW")) // Fourth attempt: /usr/local/share/ldraw
		return true;
	else if (ldraw::init_refmap("/usr/local/share/LDRAW")) // Fifth attempt: /usr/local/share/LDRAW (uppercase)
		return true;
	else if (ldraw::init_refmap("~/.ldraw")) // Sixth attempt: ~/.ldraw (home directory)
		return true;
	else // ...And return false when all of these attempts have failed
		return false;
}

void KLDrawMainWindow::initializeMenus()
{
	KActionCollection *ac = actionCollection();
	
	/**
	 * File
	 **/
	newAction = KStdAction::openNew(this, SLOT(newFile()), ac);
	openAction = KStdAction::open(this, SLOT(openFile()), ac);
	openRecentAction = KStdAction::openRecent(this, SLOT(openRecent(const KURL &)), ac);
	openRecentAction->setMaxItems(10);
	openRecentAction->loadEntries(kapp->config());
	saveAction = KStdAction::save(this, SLOT(saveFile()), ac);
	saveAsAction = KStdAction::saveAs(this, SLOT(saveFileAs()), ac);
	closeAction = KStdAction::close(this, SLOT(closeFile()), ac);
	quitAction = KStdAction::quit(this, SLOT(close()), ac);
	
	/**
	 * Edit
	 **/
	// Grid actionss
	gridSparseAction = new KRadioAction(i18n("Sparse grid"), "grid_sparse", KShortcut("Ctrl+1"), editor_, SLOT(gridSparse()), ac, "grid_sparse");
	gridStandardAction = new KRadioAction(i18n("Standard grid"), "grid_standard", KShortcut("Ctrl+2"), editor_, SLOT(gridStandard()), ac, "grid_standard");
	gridDenseAction = new KRadioAction(i18n("Dense grid"), "grid_dense", KShortcut("Ctrl+3"), editor_, SLOT(gridDense()), ac, "grid_dense");
	gridFineAction = new KRadioAction(i18n("Fine grid"), "grid_fine", KShortcut("Ctrl+4"), editor_, SLOT(gridFine()), ac, "grid_fine");
	gridSparseAction->setExclusiveGroup("grid");
	gridStandardAction->setExclusiveGroup("grid");
	gridDenseAction->setExclusiveGroup("grid");
	gridFineAction->setExclusiveGroup("grid");
	gridStandardAction->setChecked(true);
	
	// Rotate action
	rotateXCCWAction = new KAction(i18n("Rotate by x-axis counterclockwise"), "rotate_ccw", editor_, SLOT(rotateXCCW()), ac, "rotate_x_ccw");
	rotateXCWAction = new KAction(i18n("Rotate by x-axis clockwise"), "rotate_cw", editor_, SLOT(rotateXCW()), ac, "rotate_x_cw");
	rotateYCCWAction = new KAction(i18n("Rotate by y-axis counterclockwise"), "rotate_ccw", editor_, SLOT(rotateYCCW()), ac, "rotate_y_ccw");
	rotateYCWAction = new KAction(i18n("Rotate by y-axis clockwise"), "rotate_cw", editor_, SLOT(rotateYCW()), ac, "rotate_y_cw");
	rotateZCCWAction = new KAction(i18n("Rotate by z-axis counterclockwise"), "rotate_ccw", editor_, SLOT(rotateZCCW()), ac, "rotate_z_ccw");
	rotateZCWAction = new KAction(i18n("Rotate by z-axis clockwise"), "rotate_cw", editor_, SLOT(rotateZCW()), ac, "rotate_z_cw");
	
	// Move action
	moveXMAction = new KAction(i18n("Move by x-axis -"), KShortcut("Left"), editor_, SLOT(moveXM()), ac, "moveXMAction");
	moveXPAction = new KAction(i18n("Move by x-axis +"), KShortcut("Right"), editor_, SLOT(moveXP()), ac, "moveXPAction");
	moveYMAction = new KAction(i18n("Move by y-axis -"), KShortcut("Shift+Up"), editor_, SLOT(moveYM()), ac, "moveYMAction");
	moveYPAction = new KAction(i18n("Move by y-axis +"), KShortcut("Shift+Down"), editor_, SLOT(moveYP()), ac, "moveYPAction");
	moveZMAction = new KAction(i18n("Move by z-axis -"), KShortcut("Down"), editor_, SLOT(moveZM()), ac, "moveZMAction");
	moveZPAction = new KAction(i18n("Move by z-axis +"), KShortcut("Up"), editor_, SLOT(moveZP()), ac, "moveZPAction");
	
	// Miscellaneous
	deleteAction = new KAction(i18n("Delete"), "editdelete", KShortcut("Delete"), editor_, SLOT(remove()), ac, "delete");
	editModelAction = new KAction(i18n("Model Properties"), "edit", KShortcut(), this, SLOT(openModelPropertiesDialog()), ac, "modelproperties");
	colorEditAction = new KLDrawColorPickerAction(ac);
	
	/**
	 * Submodel
	 **/
	submodelNewAction = new KAction(i18n("New Submodel..."), "filenew", KShortcut(), this, SLOT(submodelNew()), ac, "submodel_new");
	submodelModifyAction = new KAction(i18n("Submodel Properties"), "edit", KShortcut(), this, SLOT(submodelModify()), ac, "submodel_modify");
	submodelDeleteAction = new KAction(i18n("Delete Submodel"), "fileclose", KShortcut(), this, SLOT(submodelDelete()), ac, "submodel_delete");
	
	/**
	 * View
	 **/
	viewFirstAction = new KAction(i18n("Move to First"), "start", KShortcut("Home"), listWidget_, SLOT(goFirst()), ac, "move_to_first");
	viewPrevAction = new KAction(i18n("Move Backward"), "back", KShortcut("PageUp"), listWidget_, SLOT(goPrev()), ac, "move_backward");
	viewNextAction = new KAction(i18n("Move Forward"), "forward", KShortcut("PageDown"), listWidget_, SLOT(goNext()), ac, "move_forward");
	viewLastAction = new KAction(i18n("Move to Last"), "finish", KShortcut("End"), listWidget_, SLOT(goLast()), ac, "move_to_last");
	
	zoomResetAction = new KAction(i18n("Zoom to Fit"), "viewmagfit", KShortcut("R"), this, SLOT(resetZoom()), ac, "zoom_reset");
	perspResetAction = new KAction(i18n("Reset Perspective"), 0, KShortcut("P"), this, SLOT(resetPersp()), ac, "persp_reset");
	
	renderAction = new KAction(i18n("Render"), "render", KShortcut("Ctrl+R"), this, SLOT(render()), ac, "render");
	//renderOptionsAction;
	
	activate(false);
}

// Setup layout
void KLDrawMainWindow::setupLayout()
{
	statusBar();
	
	QVBox *vlayout = new QVBox(this);
	
	QSplitter *hlayout = new QSplitter(vlayout);
	
	tabWidget_ = new KTabWidget(hlayout);
	submodelList_ = new KLDrawSubmodelList();
	partList_ = new KLDrawPartList();
	tabWidget_->insertTab(partList_, i18n("Parts"));
	tabWidget_->insertTab(submodelList_, i18n("Submodels"));
	
	QSplitter *topSplitter = new QSplitter(QSplitter::Vertical, hlayout);
	
	listWidget_ = new KLDrawList(topSplitter);
	
	QSplitter *vsplitter = new QSplitter(QSplitter::Horizontal, topSplitter);
	
	QSplitter *hsplitter[2];
	hsplitter[0] = new QSplitter(QSplitter::Vertical, vsplitter);
	hsplitter[1] = new QSplitter(QSplitter::Vertical, vsplitter);
	
	renderWidget_[0] = new KLDrawRenderWidget(&activeDocument_, KLDrawRenderWidget::viewportFront, 0, hsplitter[0]);
	renderWidget_[1] = new KLDrawRenderWidget(&activeDocument_, KLDrawRenderWidget::viewportTop  , 1, hsplitter[0]);
	renderWidget_[2] = new KLDrawRenderWidget(&activeDocument_, KLDrawRenderWidget::viewportLeft , 2, hsplitter[1]);
	renderWidget_[3] = new KLDrawRenderWidget(&activeDocument_, KLDrawRenderWidget::viewportFree , 3, hsplitter[1]);
	
	tabBar_ = new KTabBar(vlayout);
	
	tabBar_->setShape(QTabBar::RoundedBelow);
	
	hlayout->setResizeMode(tabWidget_, QSplitter::KeepSize);
	tabWidget_->resize(250, submodelList_->height());
	
	topSplitter->setResizeMode(listWidget_, QSplitter::KeepSize);
	listWidget_->resize(listWidget_->width(), 130);
	
	setCentralWidget(vlayout);
}

void KLDrawMainWindow::setupConnections()
{
	QObject::connect(tabBar_, SIGNAL(selected(int)), this, SLOT(changeActiveSelection(int)));
	
	QObject::connect(this, SIGNAL(activeModelChanged()), listWidget_, SLOT(updateList()));
	QObject::connect(this, SIGNAL(documentChanged()), submodelList_, SLOT(updateList()));
	QObject::connect(this, SIGNAL(documentChanged()), this, SLOT(changeCaption()));
	QObject::connect(editor_, SIGNAL(modelModified()), this, SLOT(changeCaption()));
	
	for (int i = 0; i < 4; ++i) {
		renderWidget_[i]->setControlled(true);
		QObject::connect(this, SIGNAL(activeModelChanged()), renderWidget_[i], SLOT(updateGL()));
		QObject::connect(this, SIGNAL(refresh()), renderWidget_[i], SLOT(updateGL()));
		QObject::connect(listWidget_, SIGNAL(selectionChanged()), renderWidget_[i], SLOT(updateGL()));
		QObject::connect(editor_, SIGNAL(needUpdate()), renderWidget_[i], SLOT(updateGL()));
		
		QObject::connect(renderWidget_[i], SIGNAL(objectDragging()), this, SLOT(synchronizeViews()));
		QObject::connect(renderWidget_[i], SIGNAL(startDragging(const ldraw::vector3d &, const ldraw::vector3d &)), this, SLOT(dispatchDragStart(const ldraw::vector3d &, const ldraw::vector3d &)));
		QObject::connect(renderWidget_[i], SIGNAL(stopDragging()), this, SLOT(dispatchDragStop()));
	}
	
	QObject::connect(submodelList_, SIGNAL(executed(QListViewItem *)), this, SLOT(changeActiveModel(QListViewItem *)));
	
	QObject::connect(this, SIGNAL(editingEnabled(bool)), this, SLOT(activate(bool)));
	
	QObject::connect(listWidget_, SIGNAL(enableAction(int, bool)), this, SLOT(enableAction(int, bool)));
	QObject::connect(listWidget_, SIGNAL(cursorChanged(KLDrawListItem *)), this, SLOT(changeCursor(KLDrawListItem *)));
	QObject::connect(editor_, SIGNAL(cursorChanged(KLDrawListItem *)), this, SLOT(changeCursor(KLDrawListItem *)));
}

void KLDrawMainWindow::loadURL(const KURL &location)
{
	QString tempFile;
	if (!KIO::NetAccess::download(location, tempFile, this)) {
		KMessageBox::error(this, i18n("<qt>Could not open <i>\"%1\"</i>: %2</qt>").arg(location.fileName()).arg(KIO::NetAccess::lastErrorString()));
		return;
	}
	for (list<KLDrawDocument *>::iterator it = documentList_.begin(); it != documentList_.end(); ++it) {
		if ((*it)->location() == location) {
			KMessageBox::error(this, i18n("<qt>File <i>\"%1\"</i> is already on the workspace.</qt>").arg(location.filename()));
			return;
		}
	}
	
	KLDrawDocument *newDocument = new KLDrawDocument(tempFile, this);
	// TODO Exception handling
	
	if (tabBar_->count() == 0)
		emit editingEnabled(true);
	
	KIO::UDSEntry entry;
	if (!KFileItem(entry, location).isWritable())
		newDocument->setReadOnly(true);
	newDocument->setLocation(location);
	addDocument(newDocument);
}

bool KLDrawMainWindow::saveURL(KLDrawDocument *d, const KURL &location, bool newName)
{
	if (d->readOnly() && !newName) {
		KMessageBox::error(this, i18n("<qt>Current document is read-only.</qt>"));
		return false;
	}
	
	string tempFile = string("/tmp/kldraw_save_") + getenv("USER");
	string newFileName = string(location.fileName().local8Bit());
	model_multipart *mainModel = d->getModel();
	
	if (newName)
		mainModel->get_model()->name() = newFileName;
	
	if (!mainModel->save_dat(tempFile, true)) {
		KMessageBox::sorry(this, i18n("<qt>Could not save temporary file.</qt>"));
		return false;
	}
	
	if (!KIO::NetAccess::upload(QString(tempFile.c_str()), location, this)) {
		KMessageBox::sorry(this, i18n("<qt>Error saving file: %1</qt>").arg(KIO::NetAccess::lastErrorString()));
		unlink(tempFile.c_str());
		return false;
	}
	
	d->setReadOnly(false);
	d->setSaveable(false);
	d->setFirstSave(false);
	
	if (newName)
		d->setLocation(location);
	
	if (d == activeDocument_)
		changeCaption();
	
	unlink(tempFile.c_str());
	
	return true;
}

void KLDrawMainWindow::addDocument(KLDrawDocument *newDocument)
{
	documentList_.push_back(newDocument);
	int id = tabBar_->addTab(newDocument);
	tabBar_->setCurrentTab(id);
	
	if (tabBar_->count() == 1)
		changeActiveSelection(id);
}

bool KLDrawMainWindow::queryClose()
{
	for (list<KLDrawDocument *>::iterator it = documentList_.begin(); it != documentList_.end(); ++it) {
		if (!checkTab(*it))
			return false;
	}
	
	return true;
}

bool KLDrawMainWindow::checkTab(KLDrawDocument *d)
{
	if (d->canSave() && !d->readOnly()) {
		switch (KMessageBox::warningYesNoCancel(this, i18n("<qt>The document \"%1\" has been modified.<br/><br/>Do you want to save your changes or discard them?</qt>").arg(QString(d->getModel()->get_model()->name().c_str())))) {
			case KMessageBox::Yes:
				return saveDocument(d);
			case KMessageBox::No:
				return true;
			default:
				return false;
		}
	} else
		return true;
}

bool KLDrawMainWindow::saveDocument(KLDrawDocument *d)
{
	if (activeDocument_->firstSave())
		return saveDocumentAs(d);
	else
		return saveURL(d, d->location(), false);
}

bool KLDrawMainWindow::saveDocumentAs(KLDrawDocument *d)
{
	KURL url = KFileDialog::getSaveURL(QString(d->getModel()->get_model()->name()),
		QString("*.ldr|") + i18n("LDraw Model File (*.ldr)") + "\n" + 
		"*|" + i18n("All Files"), this
	);
	
	// when no url specified
	if (url.isEmpty()) return false;
	
	// if the url already exists
	if (KIO::NetAccess::exists(url, false, this)) {
		if (KMessageBox::warningYesNo(this, i18n("A file named \"%1\" already exists. Are you sure you want to overwrite it?").arg(url.fileName())) == KMessageBox::No)
			return false;
	}
	
	return saveURL(d, url, true);
}

void KLDrawMainWindow::sorry()
{
	KMessageBox::sorry(this, i18n("Not implemented. sorry :("));
}
