// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <qdragobject.h>
#include <qheader.h>
#include <qimage.h>
#include <qsplitter.h>
#include <qtooltip.h>
#include <klistview.h>
#include <klocale.h>
#include <kiconview.h>
#include <ktoolbarbutton.h>

#include "kldraw.h"
#include "dbmanager.h"
#include "clicklineedit.h"
#include "fixedtoolbar.h"

#include "partlist.h"

using namespace ldraw;

/////////////////////////////////////////////////////////////////////////////////////////
// KLDrawPartList
/////////////////////////////////////////////////////////////////////////////////////////

KLDrawPartList::KLDrawPartList(QWidget *parent, const char *name)
	: QVBox(parent, name)
{
	thumbnailCache_.setCacheLimit(4096); // TODO configurable
	db_ = KLDrawDBManager::instance();
	
	// Setup UI
	setSpacing(4);
	
	KToolBar *searchBar = new KLDrawFixedToolBar(this);
	
	eraseButton_ = new KToolBarButton("locationbar_erase", 0, searchBar);
	searchEdit_ = new KLDrawClickLineEdit(i18n("Enter search terms here"), searchBar);
	searchEdit_->installEventFilter(this);
	searchBar->setStretchableWidget(searchEdit_);
	
	searchEdit_->setFrame(QFrame::Sunken);
	connect(eraseButton_, SIGNAL(clicked()), searchEdit_, SLOT(clear()));
	
	QToolTip::add(eraseButton_, i18n("Clear filter"));
	QToolTip::add(searchEdit_, i18n("Enter space-separated terms to filter collection"));
	
	QSplitter *vsplitter = new QSplitter(QSplitter::Vertical, this);
	
	partList_ = new KLDrawPartListView(vsplitter);
	iconList_ = new KLDrawPartIconView(this, vsplitter);
	vsplitter->setResizeMode(partList_, QSplitter::KeepSize);
	partList_->resize(partList_->width(), 480);
	
	connect(partList_, SIGNAL(selectionChanged(QListViewItem *)), iconList_, SLOT(updateView(QListViewItem *)));
	connect(iconList_, SIGNAL(selectionChanged(QIconViewItem *)), iconList_, SLOT(selectNew(QIconViewItem *)));
	connect(iconList_, SIGNAL(pointItemInList(KLDrawPartListViewItem *)), partList_, SLOT(selectItem(KLDrawPartListViewItem *)));
	
	initializeItems();
}

KLDrawPartList::~KLDrawPartList()
{
	thumbnailCache_.clear();
}

void KLDrawPartList::setEnabled(bool b)
{
	partList_->setEnabled(b);
	iconList_->setEnabled(b);
	eraseButton_->setEnabled(b);
	searchEdit_->setEnabled(b);
}

void KLDrawPartList::initializeItems()
{
	QStringList l = db_->query("SELECT family,desc,filename,partno,minx,maxx,miny,maxy,minz,maxz FROM parts;");
	
	for (QStringList::iterator it = l.begin(); it != l.end(); ) {
		const QString &family = *(it++);
		const QString &desc = *(it++);
		const QString &filename = *(it++);
		const QString &partNo = *(it++);
		float minx = (*(it++)).toFloat();
		float maxx = (*(it++)).toFloat();
		float miny = (*(it++)).toFloat();
		float maxy = (*(it++)).toFloat();
		float minz = (*(it++)).toFloat();
		float maxz = (*(it++)).toFloat();
		vector3d min(minx, miny, minz);
		vector3d max(maxx, maxy, maxz);
		
		partList_->registerItem(family, desc, partNo, filename, min, max);
	}
	
	partList_->getRoot()->setOpen(true);
}

/////////////////////////////////////////////////////////////////////////////////////////
// KLDrawPartListView
/////////////////////////////////////////////////////////////////////////////////////////

KLDrawPartListView::KLDrawPartListView(QWidget *parent, const char *name)
	: KListView(parent, name)
{
	addColumn(i18n("Parts"));
	addColumn(i18n("Part no."));
	addColumn(i18n("Dimension"));
	
	setResizeMode(QListView::LastColumn);
	setItemsMovable(false);
	setSorting(0);
	setShowSortIndicator(true);
	setRootIsDecorated(true);
	setColumnWidthMode(0, QListView::Manual);
	setColumnWidthMode(1, QListView::Manual);
	setColumnWidthMode(2, QListView::Manual);
	setColumnWidth(0, 300);
	setColumnWidth(1, 96);
	setColumnWidth(2, 80);
	
	partMain_ = new KListViewItem(this);
	partFavorite_ = new KListViewItem(this, partMain_);
	partMain_->setText(0, i18n("All parts"));
	partMain_->setSelectable(false);
	partFavorite_->setText(0, i18n("Favorite parts"));
	partMain_->setSelectable(false);
}

KLDrawPartListView::~KLDrawPartListView()
{
	
}

void KLDrawPartListView::selectItem(KLDrawPartListViewItem *item)
{
	QListViewItem *i = static_cast<QListViewItem *>(item);
	
	ensureItemVisible(i);
	setSelected(i, true);
}

// Insert an item into the partlistview
void KLDrawPartListView::registerItem(const QString &category, const QString &desc, const QString &partNo, const QString &filename, const vector3d &min, const vector3d &max)
{
	KListViewItem *i;
	
	if (!categoryList_.contains(category)) {
		i = new KListViewItem(partMain_);
		i->setText(0, category);
		i->setExpandable(true);
		categoryList_.insert(category, i);
	} else
		i = categoryList_[category];
	
	KLDrawPartListViewItem *ni = new KLDrawPartListViewItem(i);
	ni->setText(0, desc);
	ni->setText(1, partNo);
	ni->setFilename(filename);
	ni->setCategory(category);
	ni->setDimension(min, max);
}

QDragObject* KLDrawPartListView::dragObject()
{
	KLDrawPartListViewItem *cur = dynamic_cast<KLDrawPartListViewItem *>(selectedItem());
	
	// Poor serialization method
	QString serialized = QString("%1:%2:%3:%4:%5:%6:%7").arg(cur->filename()).arg(cur->dimMin().x()).arg(cur->dimMin().y()).arg(cur->dimMin().z()).arg(cur->dimMax().x()).arg(cur->dimMax().y()).arg(cur->dimMax().z());
	
	QTextDrag *drag = new QTextDrag(serialized, this);
	return drag;
}

/////////////////////////////////////////////////////////////////////////////////////////
// KLDrawPartListViewItem
/////////////////////////////////////////////////////////////////////////////////////////

KLDrawPartListViewItem::KLDrawPartListViewItem(KListViewItem *category)
	: KListViewItem(category)
{
	setDragEnabled(true);
}

/////////////////////////////////////////////////////////////////////////////////////////
// KLDrawPartIconView
/////////////////////////////////////////////////////////////////////////////////////////

KLDrawPartIconView::KLDrawPartIconView(KLDrawPartList *top, QWidget *parent, const char *name)
	: KIconView(parent, name)
{
	top_ = top;
	setSelectionMode(QIconView::Single);
	setSorting(false);
	setItemsMovable(false);
}

KLDrawPartIconView::~KLDrawPartIconView()
{
	
}

void KLDrawPartIconView::updateView(QListViewItem *item)
{
	QListViewItem *category;
	KLDrawPartListViewItem *cursor;
	
	if (item->depth() == 1) {
		category = item;
		cursor = 0L; 
	} else if (item->depth() == 2) {
		category = item->parent();
		cursor = dynamic_cast<KLDrawPartListViewItem *>(item);
	} else
		return;
	
	if (currentCategory_ != category->text(0)) {
		KLDrawPartListViewItem *child = dynamic_cast<KLDrawPartListViewItem *>(category->firstChild());
		
		currentCategory_ = category->text(0);
		clear();
		
		KLDrawPartIconViewItem *prevItem = 0L;
		KLDrawPartIconViewItem *newItem;
		QString key;
		for (; child; child = dynamic_cast<KLDrawPartListViewItem *>(child->nextSibling())) {
			key = child->filename();
			
			QPixmap pm, partpm;
			if (!top_->thumbnailCache_.find(key, pm)) {
				QPixmap partpm(KLDraw::saveLocation("partimgs/")+key+".png");
				if (partpm.isNull())
					continue;
				
				pm = QPixmap(partpm.convertToImage().smoothScale(64, 64, QImage::ScaleMin));  // TODO configurable
				top_->thumbnailCache_.insert(key, pm);
			}
			
			if (!prevItem)
				newItem = new KLDrawPartIconViewItem(child, this);
			else
				newItem = new KLDrawPartIconViewItem(child, this, prevItem);
			
			newItem->setSelectable(true);
			newItem->setKey(key);
			newItem->setText(key.section('.', 0, 0));
			newItem->setPixmapSize(QSize(64, 64)); // TODO configurable
			newItem->setPixmap(pm);
			
			prevItem = newItem;
		}
	}
	
	QIconViewItem *iptr;
	if (cursor) {
		for (iptr = firstItem(); iptr != 0L; iptr = iptr->nextItem()) {
			if (iptr->key() == cursor->filename()) {
				ensureItemVisible(iptr);
				iptr->setSelected(true);
				break;
			}
		}
	}
}

void KLDrawPartIconView::selectNew(QIconViewItem *item)
{
	if (!item) return;
	
	KLDrawPartIconViewItem *i = dynamic_cast<KLDrawPartIconViewItem *>(item);
	
	emit pointItemInList(i->listItem());
}

QDragObject* KLDrawPartIconView::dragObject()
{
	KLDrawPartListViewItem *cur = dynamic_cast<KLDrawPartIconViewItem *>(currentItem())->listItem();
	
	QString serialized = QString("%1:%2:%3:%4:%5:%6:%7").arg(cur->filename()).arg(cur->dimMin().x()).arg(cur->dimMin().y()).arg(cur->dimMin().z()).arg(cur->dimMax().x()).arg(cur->dimMax().y()).arg(cur->dimMax().z());
	
	QTextDrag *drag = new QTextDrag(serialized, this);
	return drag;
}

/////////////////////////////////////////////////////////////////////////////////////////
// KLDrawPartIconViewItem
/////////////////////////////////////////////////////////////////////////////////////////

KLDrawPartIconViewItem::KLDrawPartIconViewItem(KLDrawPartListViewItem *item, KLDrawPartIconView *parent)
	: KIconViewItem(parent)
{
	item_ = item;
}

KLDrawPartIconViewItem::KLDrawPartIconViewItem(KLDrawPartListViewItem *item, KLDrawPartIconView *parent, KLDrawPartIconViewItem *after)
	: KIconViewItem(parent, after)
{
	item_ = item;
}
