// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <iostream>

#include <math.h>

#include <qgl.h>

#include "document.h"
#include "libldr.h"

#include "partrenderer.h"

using namespace ldraw;

KLDrawPartRenderer::KLDrawPartRenderer(int width, int height)
{
	width_ = width;
	height_ = height;
	
	setStudMode(studNormal);
	initializeColorStack(7);	
	
	// Initializes up OpenGL context
	QGLFormat format;
	format.setDirectRendering(false);
	format.setDoubleBuffer(false);
	
	renderedImage_ = QPixmap(width_, height_, 24);
	
	context_ = new QGLContext(format, &renderedImage_);
	context_->create();
	
	context_->makeCurrent();
	
	// Initialize GL
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClearDepth(1.0f);
	glShadeModel(GL_SMOOTH);
	
	// Enable depth test
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	// Set blend enabled
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glMatrixMode(GL_PROJECTION);
	glViewport(0, 0, width_, height_);
}

KLDrawPartRenderer::~KLDrawPartRenderer()
{
	delete context_;
}

void KLDrawPartRenderer::setNewSize(int width, int height)
{
	width_ = width;
	height_ = height;
	
	renderedImage_ = QPixmap(width_, height_);
	
	context_->makeCurrent();
	
	glViewport(0, 0, width_, height_);
}

QPixmap KLDrawPartRenderer::render(model *m, bool crop)
{
	context_->makeCurrent();
	
	glDrawBuffer(GL_FRONT);
	
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	// Setup viewport
	// TODO reuse code
	
	m->get_dimension();
	
	vector3d &min = m->dimension().first;
	vector3d &max = m->dimension().second;
	
	KLDrawViewport vp, dvp;
	vp.left   = 1e30;
	vp.right  = -1e30;
	vp.top    = 1e30;
	vp.bottom = -1e30;
	
	vector3d transformedVec[8];
	transformedVec[0] = isometricProjectionMatrix * min;
	transformedVec[1] = isometricProjectionMatrix * vector3d(min.x(), min.y(), max.z());
	transformedVec[2] = isometricProjectionMatrix * vector3d(min.x(), max.y(), min.z());
	transformedVec[3] = isometricProjectionMatrix * vector3d(min.x(), max.y(), max.z());
	transformedVec[4] = isometricProjectionMatrix * vector3d(max.x(), min.y(), min.z());
	transformedVec[5] = isometricProjectionMatrix * vector3d(max.x(), min.y(), max.z());
	transformedVec[6] = isometricProjectionMatrix * vector3d(max.x(), max.y(), min.z());
	transformedVec[7] = isometricProjectionMatrix * max;
	
	for (int i = 0; i < 8; i++) {
		if (transformedVec[i].x() < vp.left)
			vp.left = transformedVec[i].x();
		if (transformedVec[i].x() > vp.right)
			vp.right = transformedVec[i].x();
		if (transformedVec[i].y() > vp.bottom)
			vp.bottom = transformedVec[i].y();
		if (transformedVec[i].y() < vp.top)
			vp.top = transformedVec[i].y();
	}
	
	float fxlen = fabs(vp.right  - vp.left)*0.5f; 
	float fylen = fabs(vp.bottom - vp.top )*0.5f;
	
	vp.left   -= fxlen*0.1f;
	vp.right  += fxlen*0.1f;
	vp.top    -= fylen*0.1f;
	vp.bottom += fylen*0.1f;
	vp.aspectRatio = fxlen/fylen;
	dvp = vp;
	
	float median, d;
	
	if (fabs(vp.bottom-vp.top)*((float)width_/height_) >= fabs(vp.right-vp.left)) {
		median = (vp.right + vp.left) * 0.5f;
		d = fabs((vp.right-median)*((float)width_/height_))/vp.aspectRatio;
		vp.left = median - d;
		vp.right = median + d;
		glOrtho(vp.left, vp.right, vp.bottom, vp.top, 10000.0f, -10000.0f);
	} else {
		median = (vp.top + vp.bottom) * 0.5f;
		d = fabs((vp.bottom-median)/((float)width_/(float)height_))*vp.aspectRatio;
		vp.bottom = median + d;
		vp.top = median - d;
		glOrtho(vp.left, vp.right, vp.bottom, vp.top, 10000.0f, -10000.0f);
	}
	
	// Draw to pixbuf
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glLoadMatrixf(isometricProjectionMatrix.transpose().get_pointer());
	
	glLineWidth(2.0f);
	drawModel(m, 0);
	
	glFlush();
	
	if (crop) {
		int w, h;
		float xl = fabs(dvp.right-dvp.left);
		float yl = fabs(dvp.bottom-dvp.top);
		
		if (xl >= yl) {
			w = width_;
			h = (int)(width_ * (yl/xl));
		} else {
			h = height_;
			w = (int)(height_ * (xl/yl));
		}
		
		QPixmap croppedpm(w, h);
		croppedpm.fill(); // Fill to white
		copyBlt(&croppedpm, 0, 0, &renderedImage_, width_/2 - w/2, height_/2 - h/2, w, h);
		
		return croppedpm;
	} else
		return renderedImage_;
}
