// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <cmath>
#include <sstream>

#include <qstring.h>

#include "kldraw.h"
#include "povrayrenderparameters.h"

#include "povrayexporter.h"

#define RADIAN(n) (n / 180.0f * M_PI)

/* POV-Ray serializer
 * Exports a LDraw model to POV-Ray format.
 * Large portion of serialization format has been taken from L3P.
 * L3P by Lars C. Hassing. (http://www.hassings.dk/l3/l3p.html) 
 */

using namespace ldraw;
using namespace std;

void KLDrawPOVRayExporter::exportModel(QTextStream &stream, KLDrawDocument &doc)
{
	model *m = doc.getModel()->get_model();
	
	fillHeader(stream);
	fillColors(stream, m);
	fillParts(stream, m);
	fillFooter(stream, m);
}

QString KLDrawPOVRayExporter::convertNameFormat(const string &s)
{
	QString cs = QString::fromLocal8Bit(s.c_str());
	
	cs = cs.lower();
	cs = cs.insert(0, '_');
	cs = cs.replace('.', "_dot_");
	cs = cs.replace('-', "_dash_");
	cs = cs.replace('/', "_slash_");
	cs = cs.replace('\\', "_slash_");
	cs = cs.replace('#', "_sharp_");
	
	return cs;
}

void KLDrawPOVRayExporter::fillHeader(QTextStream &o)
{
	// LEGO Logo
	
	o << "#declare L3Logo = union {\n";
	o << "\tsphere { <-59, 0, -96>, 6}\n";
	o << "\tcylinder { <-59, 0, -96>, <59, 0, -122>, 6 open}\n";
	o << "\tsphere { <59, 0, -122>, 6 }\n";
	o << "\tcylinder { <59, 0, -122>, <59, 0, -84>, 6 open}\n";
	o << "\tsphere { <59 ,0 ,-84>, 6 }\n\n";
	
	o << "\tsphere { <-59, 0, -36>, 6 }\n";
	o << "\tcylinder { <-59, 0, -36>, <-59, 0, 1>, 6 open}\n";
	o << "\tsphere { <-59, 0, 1>, 6 }\n";
	o << "\tcylinder { <0, 0, -49>, <0, 0, -25>, 6 open }\n";
	o << "\tsphere { <0, 0, -25>, 6 }\n";
	o << "\tsphere { <59, 0, -62>, 6 }\n";
	o << "\tcylinder { <59, 0, -62>, <59, 0, -24>, 6 open }\n";
	o << "\tsphere { <59, 0, -24>, 6 }\n";
	o << "\tcylinder { <-59 ,0 ,-36>, <59, 0, -62>, 6 open }\n\n";
	
	o << "\tsphere { <-35.95, 0, 57>, 6 }\n";
	o << "\ttorus { 18.45, 6 clipped_by { plane { <40,0,-9>, 0 } } translate <-40, 0, 39> }\n";
	o << "\tcylinder { <-44.05, 0, 21>, <35.95, 0, 3>, 6 open }\n";
	o << "\ttorus { 18.45, 6 clipped_by { plane { <-40, 0, 9>, 0 } } translate <40, 0, 21> }\n";
	o << "\tcylinder { <44.05, 0, 39>, <0, 0, 49>, 6 open }\n";
	o << "\tsphere { <0, 0, 49>, 6 }\n";
	o << "\tcylinder { <0, 0, 49>, <0, 0, 34>, 6 open }\n";
	o << "\tsphere { <0, 0, 34>, 6 }\n\n";
	
	o << "\ttorus { 18.45, 6 clipped_by { plane { <40, 0, -9>, 0 } } translate <-40, 0, 99> }\n";
	o << "\tcylinder { <-44.05, 0, 81>, <35.95, 0, 63>, 6 open }\n";
	o << "\ttorus { 18.45, 6 clipped_by { plane { <-40, 0, 9>,0 } } translate <40, 0, 81> }\n";
	o << "\tcylinder { <44.05, 0, 99>,<-35.95, 0, 117>, 6 open }\n\n";
	
	o << "\tscale 4.5 / 128\n";
	o << "}\n\n";
}

void KLDrawPOVRayExporter::fillColors(QTextStream &o, model *m)
{
	list<int> colorList;
	
	fillColorsRecursive(o, m, colorList);
}

void KLDrawPOVRayExporter::fillColorsRecursive(QTextStream &o, model *m, list<int> &l)
{
	bool found;
	int i = 0;
	
	for (model::iterator it = m->begin(); it != m->end(); ++i) {
		int c;
		
		found = false;
		
		if (i == 0) // At least, color '0' must be exported.
			c = 0;
		else if ((*it)->get_type() == type_extref && CAST_AS_EXTREF(*it)->get_model()) {
			c = CAST_AS_EXTREF(*it)->get_color().get_id();
			fillColorsRecursive(o, CAST_AS_EXTREF(*it)->get_model(), l);
		} else if ((*it)->get_type() == type_triangle)
			c = CAST_AS_TRIANGLE(*it)->get_color().get_id();
		else if ((*it)->get_type() == type_quadrilateral)
			c = CAST_AS_QUADRILATERAL(*it)->get_color().get_id();
		else {
			if (i > 0)
				++it;
			continue;
		}
		
		// Traveling down list, find every color used in the model
		// if a color has not been written, write it.
		for (list<int>::iterator it2 = l.begin(); it2 != l.end(); ++it2) {
			if ((*it2) == c) {
				found = true;
				break;
			}
		}
		// Serialize!
		if (!found) {
			color ce(c);
			material_type material = ce.get_entity()->material;
			const unsigned char *rgba = ce.get_entity()->rgba;
			float ambient;
			float diffuse;
			float reflection;
			int phong_size;
			
			switch (material) {
				case material_clear:
					ambient = 0.3f;
					diffuse = 0.6f;
					reflection = 0.25f;
					phong_size = 60;
					break;
				case material_metallic:
					ambient = 0.25f;
					diffuse = 0.6f;
					reflection = 0.6f;
					break;
				default:
					ambient = 0.15f;
					diffuse = 0.4f;
					reflection = 0.05f;
					phong_size = 20;
			}
			
			l.push_back(c);
			
			o << "#declare Color" << c << " = material {\n";
			o << "\ttexture {\n";
			o << "\t\tpigment { " << (material != material_clear ? "rgb" : "rgbf") << " <" << (float)rgba[0]/255.0f << ", " << (float)rgba[1]/255.0f << ", " << (float)rgba[2]/255.0f;
			if (material == material_clear)
				o << ", " << (float)rgba[3]/255.0f;
			o << "> }\n";
			o << "\t\tfinish { ambient " << ambient << " diffuse " << diffuse << " reflection " << reflection << " ";
			if (material != material_metallic)
				o << "phong 0.3 phong_size " << phong_size << " ";
			else
				o << "brilliance 5 metallic specular 0.7 roughness 1/100 ";
			if (material == material_clear)
				o << "refraction 1 ior 1.5 ";
			o << "}\n";
			if (material == material_normal)
				o << "\t\tnormal { bumps 0.1 scale 0.02 }\n";
			o << "\t}\n}\n\n";
		}
		
		if (i > 0)
			++it;
	}
}

void KLDrawPOVRayExporter::fillParts(QTextStream &o, model *m)
{
	list<model *> partList;
	map<QString, char> blackList; // Blacklist for empty model.
	
	fillPartsRecursive(o, m, partList, blackList);
	
	o << "object { " << convertNameFormat(m->name()) << " material { Color0 } }\n"; 
}

void KLDrawPOVRayExporter::fillPartsRecursive(QTextStream &o, model *m, list<model *> &l, std::map<QString, char> &bl)
{
	bool foundFlag;
	QString modelName = convertNameFormat(m->name());
	
	// Skip this if current part is in the blacklist
	if (bl.find(modelName) != bl.end())
		return;
	
	// A group of triangles and quads must be grouped into mesh { ... } block,
	// and a mesh must have one identical color.
	// If an object has only one (or less) object, it must be defined as 'object'
	// or 'union' if else.
	int count = 0;
	int cf = -1;
	bool pext = false;
	for (model::iterator it = m->begin(); it != m->end(); ++it) {
		if ((*it)->get_type() == type_extref) {
			pext = true;
			element_extref *elem = CAST_AS_EXTREF(*it);
			
			// Duplicate check
			foundFlag = false;
			
			for (list<model *>::iterator it2 = l.begin(); it2 != l.end(); ++it2) {
				if (*it2 == elem->get_model()) {
					foundFlag = true;
					break;
				}
			}
			
			if (!foundFlag)
				fillPartsRecursive(o, elem->get_model(), l, bl);
			
			if (bl.find(convertNameFormat(elem->get_model()->name())) == bl.end())
				++count;
		} else if ((*it)->get_type() == type_triangle || (*it)->get_type() == type_quadrilateral) {
			int cc = (*it)->get_type() == type_triangle ? CAST_AS_TRIANGLE(*it)->get_color().get_id() : CAST_AS_QUADRILATERAL(*it)->get_color().get_id();
			
			if (cf != cc) {
				++count;
				cf = cc;
			} else if (pext) {
				++count;
				pext = false;
			}
		}
	}
	if (!count) {
		bl[modelName] = 1;
		return; 
	}
	
	l.push_back(m);
	
	o << "#declare " << convertNameFormat(m->name()) << " = ";
	
	// If this is not a part to substitute, serialize it normally
	if (!substitute(o, convertNameFormat(m->name()))) {
		o << ((count == 1) ? "object" : "union") << " {\n";
		
		// Serialize data
		bool meshFlag = false;
		int colorFlag = -1;
		for (model::iterator it = m->begin(); it != m->end(); ++it) {
			// Process triangles and quadrilaterals.
			if ((*it)->get_type() == type_triangle || (*it)->get_type() == type_quadrilateral) {
				int c = (*it)->get_type() == type_triangle ? CAST_AS_TRIANGLE(*it)->get_color().get_id() : CAST_AS_QUADRILATERAL(*it)->get_color().get_id();
				if (colorFlag == -1) {
					colorFlag = c;
				}
				
				// If current " mesh { ... } " block has been finished and one another block is about to start,
				// Terminate the previous one and create new.
				if (colorFlag != c && meshFlag) {
					ostringstream str;
					if (colorFlag != 16 && colorFlag != 24) {
						str << "\t\tmaterial { Color" << colorFlag << " }\n";
						o << str.str().c_str();
					}
					o << "\t}\n";
					meshFlag = false;
					colorFlag = c;
				}
				
				// new block
				if (!meshFlag) {
					o << "\tmesh {\n";
					meshFlag = true;
				}
				
				// Serialize!
				if ((*it)->get_type() == type_triangle) {
					element_triangle *elem = CAST_AS_TRIANGLE(*it);
					ostringstream str;
						str << "\t\ttriangle { " << KLDraw::serializePOV(elem->pos1()) << ", " << KLDraw::serializePOV(elem->pos2()) << ", " << KLDraw::serializePOV(elem->pos3()) << " }\n";
					o << str.str().c_str();
				} else {
					element_quadrilateral *elem = CAST_AS_QUADRILATERAL(*it);
					ostringstream str;
						str << "\t\ttriangle { " << KLDraw::serializePOV(elem->pos1()) << ", " << KLDraw::serializePOV(elem->pos2()) << ", " << KLDraw::serializePOV(elem->pos3()) << " }\n";
					str << "\t\ttriangle { " << KLDraw::serializePOV(elem->pos3()) << ", " << KLDraw::serializePOV(elem->pos4()) << ", " << KLDraw::serializePOV(elem->pos1()) << " }\n";
					o << str.str().c_str();
				}
			} else if ((*it)->get_type() == type_extref) {
				element_extref *elem = CAST_AS_EXTREF(*it);
				
				if (!elem->get_model())
					continue;
				
				QString name = convertNameFormat(elem->get_model()->name());
				
				if (bl.find(name) == bl.end()) {
					if (meshFlag) {
						if (colorFlag != 16 && colorFlag != 24) {
							ostringstream str;
							str << "\t\tmaterial { Color" << colorFlag << " }\n";
							o << str.str().c_str();
						}
						
						o << "\t}\n";
							meshFlag = false;
					}
				}
				
				// POV-Ray hates matrices filled of 0.
				matrix om = elem->get_matrix().transpose();
				for (int i = 0; i < 3; ++i) {
					if (om.value(i, 0) == 0.0f && om.value(i, 1) == 0.0f && om.value(i, 2) == 0.0f) {
						om.value(i, 0) = 1.0f;
						om.value(i, 1) = 1.0f;
						om.value(i, 2) = 1.0f;
					}
				}
				
				ostringstream str;
				str << "\tobject { " << name << " matrix " << KLDraw::serializePOV(om) << " ";
				if (elem->get_color().get_id() != 16 && elem->get_color().get_id() != 24)
					str << "material { Color" << elem->get_color().get_id() << " } ";
				str << "}\n";
				o << str.str().c_str(); 
			}
		}
		
		if (meshFlag) {
			if (colorFlag != 16 && colorFlag != 24) {
				ostringstream str;
				str << "\t\tmaterial { Color" << colorFlag << " }\n";
				o << str.str().c_str();
			}
			
			o << "\t}\n";
		}
	}
	
	o << "}\n\n";
}

void KLDrawPOVRayExporter::fillFooter(QTextStream &o, model *m)
{
	m->get_dimension();
	
	setPlane(o, m);
	setBackground(o);
	setCamera(o, m);
	setLights(o, m);
}

void KLDrawPOVRayExporter::setPlane(QTextStream &o, ldraw::model *m)
{
	if (!parameter_->drawPlane())
		return;
	
	const QColor &c = parameter_->planeColor();
	
	o << "object {\n";
	o << "\tplane { y, " << (parameter_->planePositionAuto() ? m->dimension().second.y() : parameter_->planePosition()) << " hollow }\n";
	o << "\ttexture { pigment { color rgb <" << (float)c.red()/255.0f << ", " << (float)c.green()/255.0f << ", " << (float)c.blue()/255.0f << "> } finish { ambient 0.4 diffuse 0.4 } }\n";
	o << "}\n\n";
}

void KLDrawPOVRayExporter::setBackground(QTextStream &o)
{
	const QColor &c = parameter_->backgroundColor(); 
	
	o << "background { color rgb <" << (float)c.red()/255.0f << ", " << (float)c.green()/255.0f << ", " << (float)c.blue()/255.0f << "> }\n";
}

// Determine camera position
// NOTE Current viewport calculation method doesn't fit model on screen exactly.
//      Anyone have better idea?
void KLDrawPOVRayExporter::setCamera(QTextStream &o, ldraw::model *m)
{
	const float latitude = RADIAN(parameter_->cameraLatitude());
	const float longitude = RADIAN(parameter_->cameraLongitude());
	
	const vector3d center = (m->dimension().first + m->dimension().second) * 0.5f;
	const vector3d lv = m->dimension().second - center;
	const float diameter = sqrt(lv.x()*lv.x() + lv.y()*lv.y() + lv.z()*lv.z()) * 2.0f;
	const vector3d camera = vector3d(sin(latitude) * cos(longitude), -sin(longitude), cos(latitude) * cos(longitude)) * diameter * parameter_->cameraRadius() + center;
	
	o << "camera {\n";
	if (parameter_->cameraOrthographic())
		o << "\torthographic\n";
	o << "\tlocation <" << camera.x() << ", " << camera.y() << ", " << camera.z() << ">\n";
	o << "\tsky -y\n";
	o << "\tright -" << parameter_->width() << " / " << parameter_->height() << " * x\n";
	o << "\tlook_at <" << center.x() << ", " << center.y() << ", " << center.z() << ">\n";
	if (!parameter_->cameraOrthographic())
		o << "\tangle " << parameter_->cameraFOV() << "\n";
	o << "}\n\n";
}

void KLDrawPOVRayExporter::setLights(QTextStream &o, ldraw::model *m)
{
	const float longitude = RADIAN(parameter_->lightsLongitude());
	const vector3d center = (m->dimension().first + m->dimension().second) * 0.5f;
	const vector3d lv = m->dimension().second - center;
	const float diameter = sqrt(lv.x()*lv.x() + lv.y()*lv.y() + lv.z()*lv.z()) * 2.0f;
	const QColor color = parameter_->lightsColor();
	
	float latitude;
	vector3d lpos;
	for (int i = 0; i < parameter_->lights(); ++i) {
		latitude = RADIAN(parameter_->lightsLatitudeOffset()) + i * (M_PI / parameter_->lights());
		if (latitude > M_PI) latitude -= M_PI;
		lpos = vector3d(sin(latitude) * cos(longitude), -sin(longitude), cos(latitude) * cos(longitude)) * diameter * parameter_->lightsRadius() + center;
		
		o << "light_source {\n";
		o << "\t<" << lpos.x() << ", " << lpos.y() << ", " << lpos.z() << ">\n";
		o << "\tcolor rgb <" << color.red() / 255.0f << ", " << color.green() / 255.0f << ", " << color.blue() / 255.0f << "> * " << parameter_->lightsIntensity() << "\n";
		o << "}\n\n";
	}
}

bool KLDrawPOVRayExporter::substitute(QTextStream &o, QString name)
{
	if (name == QString("_stud_dot_dat")) {
		substituteStud(o);
		return true;
	} else if (name == QString("_stud2_dot_dat")) {
		substituteStud2(o);
		return true;
	} else if (name == QString("_stud3_dot_dat")) {
		substituteStud3(o);
		return true;
	} else if (name == QString("_stud4_dot_dat")) {
		substituteStud4(o);
		return true;
	} else if (name == QString("_peghole_dot_dat")) {
		substitutePeghole(o);
		return true;
	} else if (name == QString("_2_dash_4cyli_dot_dat")) {
		substitute2Dash4Cyli(o);
		return true;
	} else if (name == QString("_2_dash_4ndis_dot_dat")) {
		substitute2Dash4Ndis(o);
		return true;
	} else if (name == QString("_2_dash_4disc_dot_dat")) {
		substitute2Dash4Disc(o);
		return true;
	} else if (name == QString("_4_dash_4cyli_dot_dat")) {
		substitute4Dash4Cyli(o);
		return true;
	} else if (name == QString("_4_dash_4ndis_dot_dat")) {
		substitute4Dash4Ndis(o);
		return true;
	} else
		return false;
}

void KLDrawPOVRayExporter::substituteStud(QTextStream &o)
{
	o << "union {\n";
	o << "\tcylinder { <0, 0, 0>, <0, -4, 0>, 6 }\n";
	o << "\tobject { L3Logo translate <0, -4, 0> }\n";
}

void KLDrawPOVRayExporter::substituteStud2(QTextStream &o)
{
	o << "difference {\n";
	o << "\tcylinder { <0, 0, 0>, <0, -4, 0>, 6 }\n";
	o << "\tcylinder { <0, 0.01 ,0>, <0, -4.01, 0>, 4 }\n";
}

void KLDrawPOVRayExporter::substituteStud3(QTextStream &o)
{
	o << "cylinder {\n";
	o << "\t<0, 0, 0>, <0, -4, 0>, 4\n";
}

void KLDrawPOVRayExporter::substituteStud4(QTextStream &o)
{
	o << "difference {\n";
	o << "\tcylinder { <0, 0, 0>, <0, -4, 0>, 8 }\n";
	o << "\tcylinder { <0, 0.01 ,0>, <0, -4.01, 0>, 6 }\n";
}

void KLDrawPOVRayExporter::substitutePeghole(QTextStream &o)
{
	o << "union {\n";
	o << "\tcylinder { <0, 0, 0>, <0, 2, 0>, 8 open }\n";
	o << "\tdisc { <0, 2, 0>, <0, 1, 0>, 8, 6 }\n";
}

void KLDrawPOVRayExporter::substitute2Dash4Cyli(QTextStream &o)
{
	o << "cylinder {\n";
	o << "\t<0, 0, 0>, <0, 1, 0>, 1 open\n";
	o << "\tclipped_by {\n";
	o << "\t\tbox { <-2, -1, 0>, <2, 2, 2> }\n";
	o << "\t}\n";
}

void KLDrawPOVRayExporter::substitute2Dash4Ndis(QTextStream &o)
{
	o << "disc {\n";
	o << "\t<0, 0, 0>, <0, 1, 0>, 2, 1\n";
	o << "\tclipped_by {\n";
	o << "\t\tbox { <-1, -1, 0>, <1, 1, 1> }\n";
	o << "\t}\n";
}

void KLDrawPOVRayExporter::substitute2Dash4Disc(QTextStream &o)
{
	o << "disc {\n";
	o << "\t<0, 0, 0>, <0, 1, 0>, 1\n";
	o << "\tclipped_by {\n";
	o << "\t\tbox { <-2, -1, 0>, <2, 1, 2> }\n";
	o << "\t}\n";
}

void KLDrawPOVRayExporter::substitute4Dash4Cyli(QTextStream &o)
{
	o << "cylinder {\n";
	o << "\t<0, 0, 0>, <0, 1, 0>, 1 open\n";
}

void KLDrawPOVRayExporter::substitute4Dash4Ndis(QTextStream &o)
{
	o << "disc {\n";
	o << "\t<0, 0, 0>, <0, 1, 0>, 2, 1\n";
	o << "\tclipped_by {\n";
	o << "\t\tbox { <-1, -1, -1>, <1, 1, 1> }\n";
	o << "\t}\n";
}
