// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <qpainter.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <ktempfile.h>
#include <kprocess.h>

#include "povrayrenderparameters.h"

#include "povrayrenderwidget.h"

KLDrawPOVRayRenderWidget::KLDrawPOVRayRenderWidget(KLDrawPOVRayRenderParameters *param, KLDrawDocument &doc, QWidget *parent, const char *name)
	: QWidget(parent, name), KLDrawPOVRayExporter(param)
{
	suspended_ = false;
	
	tempFile_ = new KTempFile(QString::null, ".pov");
	//tempFile_->setAutoDelete(true);
	
	QTextStream &stream = *(tempFile_->textStream());
	stream.setEncoding(QTextStream::Latin1);
	exportModel(stream, doc);
	
	if (tempFile_->status() != 0) {
		KMessageBox::error(this, i18n("Error writing to temporary file."));
		return;
	}
	tempFile_->close();
	
	process_ = new KProcess();
	connect(process_, SIGNAL(receivedStdout(KProcess *, char *, int)),
			SLOT(povrayImage(KProcess *, char *, int)));
	connect(process_, SIGNAL(receivedStderr(KProcess *, char *, int)),
			SLOT(povrayMessage(KProcess *, char *, int)));
	connect(process_, SIGNAL(processExited(KProcess *)),
			SLOT(renderingFinished(KProcess *)));
	
	iniFile_ = param->getIniFile();
	
	*process_ << "povray";
	*process_ << QString("+I")+tempFile_->name() << "+O-" << "+FT" << "-D" << iniFile_->name();
	
	rcvHeader_ = true;
	rcvHeaderBytes_ = 0;
	rcvPixels_ = 0;
	progress_ = 0;
	numRestBytes_ = 0;
	line_ = 0;
	column_ = 0;
	skipBytes_ = 0;
	
	image_.create(parameter_->width(), parameter_->height(), 32);
	image_.setAlphaBuffer(true);
	image_.fill(qRgb(0, 0, 0));
	repaint();
	
	show();
	
	if (!process_->start(KProcess::NotifyOnExit, KProcess::AllOutput)) {
		KMessageBox::error(this, i18n("Could not execute povray. Check your installation."));
		delete process_;
		process_ = 0L;
	}
}

KLDrawPOVRayRenderWidget::~KLDrawPOVRayRenderWidget()
{
	if (process_)
		process_->kill();
	
	cleanup();
}

QSize KLDrawPOVRayRenderWidget::sizeHint() const
{
	QSize s;
	
	if (image_.isNull())
		s = QSize(200, 200);
	else
		s = image_.size();
	
	return s.expandedTo(minimumSize());
}

void KLDrawPOVRayRenderWidget::killRendering()
{
	if (process_) {
		if (suspended_)
			process_->kill(SIGCONT);
		suspended_ = false;
		process_->kill();
	}
}

void KLDrawPOVRayRenderWidget::suspendRendering()
{
	if (process_) {
		suspended_ = true;
		process_->kill(SIGSTOP);
	}
}

void KLDrawPOVRayRenderWidget::resumeRendering()
{
	if (process_) {
		process_->kill(SIGCONT);
		suspended_ = false;
	}
}

void KLDrawPOVRayRenderWidget::povrayMessage(KProcess */*proc*/, char *buffer, int buflen)
{
	QString str;
	str.setLatin1(buffer, buflen);
	povrayOutput_ += str;
	
	emit povrayMessage(str);
}

void KLDrawPOVRayRenderWidget::povrayImage(KProcess */*proc*/, char *buffer, int buflen)
{
	int index = 0;
	int i;
	int oldline = line_;
	
	if (rcvHeader_) {
		// Targa header
		while ((rcvHeaderBytes_ < 18) && (index < buflen)) {
			header_[rcvHeaderBytes_] = (unsigned char)buffer[index];
			++rcvHeaderBytes_;
			++index;
		}
		
		if (rcvHeaderBytes_ == 18) {
			// Header received
			rcvHeader_ = false;
			skipBytes_ = header_[0];
			bytespp_ = header_[16] / 8;
		}
	}
	
	if (skipBytes_ > 0) {
		int skip = buflen - index;
		if (skip > skipBytes_)
			skip = skipBytes_;
		skipBytes_ -= skip;
		index += skip;
	}
	
	if ((numRestBytes_ > 0) && (index < buflen)) {
		while ((numRestBytes_ < bytespp_) && (index < buflen)) {
			restBytes_[numRestBytes_] = (unsigned char)buffer[index];
			++index;
			++numRestBytes_;
		}
		if (numRestBytes_ == bytespp_) {
			numRestBytes_ = 0;
			if (bytespp_ == 4)
				setPixel(column_, line_, qRgba(restBytes_[2], restBytes_[1], restBytes_[0], restBytes_[3]));
			else
				setPixel(column_, line_, qRgb(restBytes_[2], restBytes_[1], restBytes_[0]));
			++column_;
			++rcvPixels_;
			if (column_ == parameter_->width()) {
				column_ = 0;
				++line_;
			}
		}
	}
	
	if (index < buflen) {
		int num = (buflen - index) / bytespp_;
		for (i = 0; i < num; ++i) {
			if (bytespp_ == 4)
				setPixel(column_, line_, qRgba(buffer[index+2], buffer[index+1], buffer[index], buffer[index+3]));
			else
				setPixel(column_, line_, qRgb(buffer[index+2], buffer[index+1], buffer[index]));
			
			index += bytespp_;
			
			++column_;
			++rcvPixels_;
			if (column_ == parameter_->width()) {
				column_ = 0;
				++line_;
			}
		}
		
		numRestBytes_ = buflen - index;
		for (i = 0; i < numRestBytes_; ++i) {
			restBytes_[i] = buffer[index];
			++index;
		}
	}
	
	if (line_ != oldline) {
		QPainter paint(this);
		int offset = 0;
		paint.drawImage(0, offset + oldline, image_.copy(0, offset + oldline, image_.width(), offset + line_ - oldline));
		emit lineFinished(line_ - 1);
	}
	
	int oldProgress = progress_;
	int numPixels = 0;
	numPixels = parameter_->width() * parameter_->height();
	
	progress_ = rcvPixels_ * 100 / numPixels;
	
	if (progress_ != oldProgress)
		emit progress(progress_);
}

void KLDrawPOVRayRenderWidget::renderingFinished(KProcess */*proc*/)
{
	if (process_->normalExit()) {
		emit finished(process_->exitStatus());
	} else {
		KMessageBox::sorry(this, i18n("POV-Ray exited abnormally."));
		emit finished(-1000);
	}
	
	cleanup();
}

void KLDrawPOVRayRenderWidget::paintEvent(QPaintEvent *e)
{
	QPainter pe(this);
	
	pe.drawImage(e->rect().left(), e->rect().top(), image_,
			e->rect().left(), e->rect().top(), e->rect().width(),
			e->rect().height());
}

void KLDrawPOVRayRenderWidget::setPixel(int x, int y, uint c)
{
	if (x >= 0 && x < image_.width() && y >= 0 && y < image_.height())
		image_.setPixel(x, y, c);
}

void KLDrawPOVRayRenderWidget::cleanup()
{
	if (process_) {
		delete process_;
		process_ = 0L;
	}
	
	if (tempFile_) {
		delete tempFile_;
		process_ = 0L;
	}
	
	if (iniFile_) {
		delete iniFile_;
		iniFile_ = 0L;
	}
}
