// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <math.h>

#include "document.h"

#include "renderer.h"

using namespace std;
using namespace ldraw;

KLDrawRenderer::KLDrawRenderer()
{	
	studMode = studSquare;
	shading = false;
	boundingBox = false;
	
	// Initialize color stack
	colorStack.push(&color_chart[0]);
}

KLDrawRenderer::~KLDrawRenderer()
{
}

void KLDrawRenderer::initializeColorStack(int id)
{
	while (!colorStack.empty())
		colorStack.pop();
	
	colorStack.push(color(id).get_entity());
}

// rendering code
void KLDrawRenderer::drawModel(model *m, int depth, int cursor)
{
	// Obtain current modelview matrix for conditional line calculation
	float *tmp = new float[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, tmp);
	matrix proj = matrix(tmp).transpose();
	delete tmp;
	
	// Iterate!
	int i = 0;
	for (model::iterator it = m->begin(); it != m->end(); ++it) {
		type elemtype = (*it)->get_type();
		
		if (elemtype == type_line) {
			// Line
			element_line *l = CAST_AS_LINE(*it);
			renderLine(*l);
		} else if(elemtype == type_triangle) {
			// Triangle
			element_triangle *l = CAST_AS_TRIANGLE(*it);
			renderTriangle(*l);
		} else if(elemtype == type_quadrilateral) {
			// Quadrilateral
			element_quadrilateral *l = CAST_AS_QUADRILATERAL(*it);
			renderQuadrilateral(*l);
		} else if (elemtype == type_condline) {
			// Conditional line
			element_condline *l = CAST_AS_CONDLINE(*it);
			renderCondLine(*l, proj);
		} else if (elemtype == type_extref) {
			// External reference
			element_extref *l = CAST_AS_EXTREF(*it);
			
			// Push appropriate color into the color stack.
			if (l->get_color().get_id() == 16)
				colorStack.push(colorStack.top());
			else if (l->get_color().get_id() == 24)
				colorStack.push(color::get_complement(colorStack.top()->id1));
			else
				colorStack.push(l->get_color().get_entity());
				
			// Affine transform
			glPushMatrix();
			glMultMatrixf(l->get_matrix().transpose().get_pointer());
			
			if (l->filename().find("stu") != string::npos) {
				renderStud(l->get_model());
			} else if (l->get_model())
				drawModel(l->get_model(), depth+1); // Recurse
				
			if (l->get_model() && !depth && cursor == i) {
				element_extref *l = CAST_AS_EXTREF(*it);
				if (!l->get_model()->has_dimension())
					l->get_model()->get_dimension();
				renderBoundingBox(l->get_model()->dimension().first, l->get_model()->dimension().second, l->get_model()->is_submodel(l->get_model()->parent()) ? 2 : 1);
			}
			
			glPopMatrix();
			colorStack.pop();
		}
		
		++i;
	}
}

/* Renders bounding box recursively */
void KLDrawRenderer::drawBoundingBoxes(model *m, int depth, int bypassId)
{
	// Iterate!
	int i = 0;
	for (model::iterator it = m->begin(); it != m->end(); ++it) {
		if (i == bypassId) {
			++i;
			continue;
		}
		
		type elemtype = (*it)->get_type();
			
		if (elemtype == type_extref && (!depth || m->is_submodel(m->parent()))) {
			element_extref *l = CAST_AS_EXTREF(*it);
			
			if (l->get_model() && !l->get_model()->has_dimension())
				l->get_model()->get_dimension();
			
			glPushMatrix();
			glMultMatrixf(l->get_matrix().transpose().get_pointer());
			
			if (l->get_model() && !l->get_model()->is_submodel(m->parent()))
				renderBoundingBox(l->get_model()->dimension().first, l->get_model()->dimension().second, 3);
			
			if (l->get_model())
				drawBoundingBoxes(l->get_model(), depth+1);
			
			glPopMatrix();
		}
		
		++i;
	}
}

// Get current color
void KLDrawRenderer::getColor(const color &c, unsigned char *array)
{
	if (c.get_id() == 16)
		memcpy((unsigned char *)array, (unsigned char *)colorStack.top()->rgba, sizeof(unsigned char)*4);
	else if (c.get_id() == 24)
		memcpy((unsigned char *)array, (unsigned char *)color::get_complement(colorStack.top()->id1)->rgba, sizeof(unsigned char)*4);
	else
		memcpy((unsigned char *)array, (unsigned char *)c.get_entity()->rgba, sizeof(unsigned char)*4);
	
	if (colorStack.top()->material == material_clear)
		array[3] = ASSUMED_TRANSPARENT_ALPHA; // To make transparency more vivid
}

// Draw a line
void KLDrawRenderer::renderLine(element_line &l)
{
	unsigned char rgba[4];
	
	getColor(l.get_color(), rgba);
	
	glColor4ubv(rgba);
	
	glBegin(GL_LINES);
	glVertex3fv(l.pos1().get_pointer());
	glVertex3fv(l.pos2().get_pointer());
	glEnd();
}

// Draw a triangle
void KLDrawRenderer::renderTriangle(element_triangle &l)
{
	unsigned char rgba[4];
	
	getColor(l.get_color(), rgba);
	
	glColor4ubv(rgba);
	
	glBegin(GL_TRIANGLES);
	if (shading) {
		vector3d normal = calculateNormal(l.pos1(), l.pos2(), l.pos3());
		glNormal3fv(normal.get_pointer());
	}
	glVertex3fv(l.pos1().get_pointer());
	glVertex3fv(l.pos2().get_pointer());
	glVertex3fv(l.pos3().get_pointer());
	glEnd();
}

// Draw a quadrilateral
void KLDrawRenderer::renderQuadrilateral(element_quadrilateral &l)
{
	unsigned char rgba[4];
	
	getColor(l.get_color(), rgba);
	
	glColor4ubv(rgba);
	
	glBegin(GL_QUADS);
	if (shading) {
		vector3d normal = calculateNormal(l.pos1(), l.pos2(), l.pos3());
		glNormal3fv(normal.get_pointer());
	}
	glVertex3fv(l.pos1().get_pointer());
	glVertex3fv(l.pos2().get_pointer());
	glVertex3fv(l.pos3().get_pointer());
	glVertex3fv(l.pos4().get_pointer());
	glEnd();
}

// Draw a conditional line
// Draw v1v2 if the projections of both v1v2 and v3v4 are parallel
void KLDrawRenderer::renderCondLine(element_condline &l, const matrix &projectionMatrix)
{
	unsigned char rgba[4];
	
	getColor(l.get_color(), rgba);
	
	glColor4ubv(rgba);
	
	// Linear transform
	vector3d tv1 = projectionMatrix * l.pos1();
	vector3d tv2 = projectionMatrix * l.pos2();
	vector3d tv3 = projectionMatrix * l.pos3();
	vector3d tv4 = projectionMatrix * l.pos4();
	
	// Turn them into direction vector
	vector3d v1 = vector3d(fabs(tv2.x()-tv1.x()), fabs(tv2.y()-tv1.y()), 0.0f).normalize();
	vector3d v2 = vector3d(fabs(tv4.x()-tv3.x()), fabs(tv4.y()-tv3.y()), 0.0f).normalize();
	
	// If parallel (approximately)
	if (fabs(v2.x()-v1.x()) <= EPSILON && fabs(v2.y()-v1.y()) <= EPSILON) {
		glBegin(GL_LINES);
		glVertex3fv(l.pos1().get_pointer());
		glVertex3fv(l.pos2().get_pointer());
		glEnd();
	}
}

// Draw a bounding box
void KLDrawRenderer::renderBoundingBox(const vector3d &min, const vector3d &max, int highlight)
{
	if (highlight == 1) {
		glColor3ub(255, 0, 0);
		glLineWidth(2.0f);
		glDepthFunc(GL_ALWAYS);
	} else if (highlight == 2) {
		glColor3ub(0, 0, 255);
		glLineWidth(2.0f);
		glDepthFunc(GL_ALWAYS);
	} else if (highlight == 3) {
		glColor3ub(0, 0, 0);
		glLineWidth(1.0f);
	}
	
	glBegin(GL_LINES);
	
	glVertex3f(min.x(), min.y(), min.z());
	glVertex3f(max.x(), min.y(), min.z());
	
	glVertex3f(min.x(), min.y(), min.z());
	glVertex3f(min.x(), max.y(), min.z());
	
	glVertex3f(min.x(), min.y(), min.z());
	glVertex3f(min.x(), min.y(), max.z());
	
	glVertex3f(max.x(), max.y(), max.z());
	glVertex3f(min.x(), max.y(), max.z());
	
	glVertex3f(max.x(), max.y(), max.z());
	glVertex3f(max.x(), min.y(), max.z());
	
	glVertex3f(max.x(), max.y(), max.z());
	glVertex3f(max.x(), max.y(), min.z());
	
	glVertex3f(max.x(), min.y(), max.z());
	glVertex3f(max.x(), min.y(), min.z());
	
	glVertex3f(min.x(), max.y(), max.z());
	glVertex3f(min.x(), max.y(), min.z());
	
	glVertex3f(max.x(), max.y(), min.z());
	glVertex3f(max.x(), min.y(), min.z());
	
	glVertex3f(min.x(), max.y(), max.z());
	glVertex3f(min.x(), min.y(), max.z());
	
	glVertex3f(max.x(), min.y(), max.z());
	glVertex3f(min.x(), min.y(), max.z());
	
	glVertex3f(max.x(), max.y(), min.z());
	glVertex3f(min.x(), max.y(), min.z());
	
	glEnd();
	
	if (highlight)
		glDepthFunc(GL_LESS);
}

void KLDrawRenderer::renderStud(model *l)
{
	// Stud is the most frequently rendered part on every models.
	// And it is quite complex to render, causing major speed drop during entire model rendering process.
	// Therefore, substituting stud for simpler primitives (such as line, square) boosts speed a lot.
	
	switch (studMode) {
		case studNormal:
			// Render stud as normal
			drawModel(l, -1);
			return;
			
		case studLine:
			// Draw a line
			glColor4ubv(color::get_complement(colorStack.top()->id1)->rgba);
			glBegin(GL_LINES);
			glVertex3f(0.0f, 0.0f, 0.0f);
			glVertex3f(0.0f, -4.0f, 0.0f);
			glEnd();
			return;
			
		case studSquare:
			// Draw a square
			glColor4ubv(color::get_complement(colorStack.top()->id1)->rgba);
			glBegin(GL_LINES);
			glVertex3f(-6.0f, -4.0f, -6.0f);
			glVertex3f(6.0f, -4.0f, -6.0f);
			
			glVertex3f(6.0f, -4.0f, -6.0f);
			glVertex3f(6.0f, -4.0f, 6.0f);
			
			glVertex3f(6.0f, -4.0f, 6.0f);
			glVertex3f(-6.0f, -4.0f, 6.0f);
			
			glVertex3f(-6.0f, -4.0f, 6.0f);
			glVertex3f(-6.0f, -4.0f, -6.0f);
			glEnd();
			return;
	}
}

// Calculates normal vector
vector3d KLDrawRenderer::calculateNormal(const vector3d &v1, const vector3d &v2, const vector3d &v3)
{
	vector3d d1, d2, out;
	
	d1 = v2 - v1;
	d2 = v3 - v2;
	
	out = vector3d::cross_product(d1, d2).normalize();
	
	return out;
}
