// KLDraw - An interactive LDraw modeler for KDE
// Copyright (c)2006 Park "segfault" J. K. <nemesis@planetmono.org>

#include <qdragobject.h>
#include <qpainter.h>
#include <qpoint.h>
#include <qevent.h>
#include <qimage.h>
#include <qbitmap.h>
#include <klocale.h>

#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "mainwindow.h"
#include "libldr.h"
#include "document.h"
#include "editor.h"
#include "list.h"
#include "kldraw.h"

#include "renderwidget.h"

using namespace ldraw;

KLDrawRenderWidget::KLDrawRenderWidget(KLDrawDocument **ptr, ViewportMode viewport, int scrid, QWidget *parent)
	: QGLWidget(parent), KLDrawRenderer()
{
	activeDocument_ = ptr;
	controlled = false;
	screenId = scrid;
	
	panning = false;
	rotating = false;
	drawGrid = true;
	viewportText = 0L;
	arcBall = 0L;
	
	dragResources_ = KLDrawMainWindow::self()->getDragResources();
	editor = KLDrawMainWindow::self()->getEditor();
	
	// Workaround
	stretched = new KLDrawViewport;
	
	initViewport(viewportNull, viewport);
	
	setAcceptDrops(true);
}

KLDrawRenderWidget::~KLDrawRenderWidget()
{
	delete stretched;
	
	if (viewportText)
		delete viewportText;
}

vector3d KLDrawRenderWidget::viewportCoordinate(const QPoint &dim) const
{
	float xr = stretched->left + ((stretched->right - stretched->left) / width()) * dim.x();
	float yr = stretched->top + ((stretched->bottom - stretched->top) / height()) * dim.y();
	
	switch (viewportMode) {
		case viewportFront:
			return vector3d(xr, yr, 0.0f);
		case viewportBack:
			return vector3d(-xr, yr, 0.0f);
		case viewportTop:
			return vector3d(xr, 0.0f, -yr);
		case viewportBottom:
			return vector3d(-xr, 0.0f, yr);
		case viewportLeft:
			return vector3d(0.0f, yr, -xr);
		case viewportRight:
			return vector3d(0.0f, yr, xr);
		default:
			return vector3d(0.0f, 0.0f, 0.0f);
	}
}

QPoint KLDrawRenderWidget::globalCoordinate(const ldraw::vector3d &vec) const
{
	float xc, yc;
	
	switch (viewportMode) {
		case viewportTop:
			xc = vec.x();
			yc = -vec.z();
			break;
		case viewportBottom:
			xc = -vec.x();
			yc = vec.z();
			break;
		case viewportLeft:
			xc = -vec.z();
			yc = vec.y();
			break;
		case viewportRight:
			xc = vec.z();
			yc = vec.y();
			break;
		case viewportFront:
			xc = vec.x();
			yc = vec.y();
			break;
		case viewportBack:
			xc = -vec.x();
			yc = vec.y();
			break;
		default:
			xc = 0.0f;
			yc = 0.0f;
	}
	
	return QPoint(
		(int)((xc - stretched->left) / (stretched->right - stretched->left) * width()),
		(int)((yc - stretched->top) / (stretched->bottom - stretched->top) * height())
	);
}

int KLDrawRenderWidget::depthAxis() const
{
	switch (viewportMode) {
		case viewportTop:
		case viewportBottom:
			return 1; // y-axis
		case viewportFront:
		case viewportBack:
			return 2; // z-axis
		case viewportLeft:
		case viewportRight:
			return 0; // x-axis
		default:
			return -1;
	}
}

void KLDrawRenderWidget::dragStart()
{
	boundingBox = true;
	
	repaint();
}

void KLDrawRenderWidget::dragStop()
{
	boundingBox = false;
	dragResources_->dragging = false;
	
	repaint();
}

void KLDrawRenderWidget::initViewport(ViewportMode oldMode, ViewportMode newMode)
{
	viewportMode = newMode;
	
	// Initialize arcball rotation
	if (oldMode != viewportFree && newMode == viewportFree)
		arcBall = new KLDrawArcBallRotation;
	else if (oldMode == viewportFree && newMode != viewportFree) {
		delete arcBall;
		arcBall = 0L;
	}
	
	// Prepare for text that showed on the corner of the viewport
	if (oldMode == newMode)
		return;
	else {
		QString viewportName;
		
		switch (newMode) {
			case viewportTop:
				viewportName = i18n("Top");
				break;
			case viewportBottom:
				viewportName = i18n("Bottom");
				break;
			case viewportFront:
				viewportName = i18n("Front");
				break;
			case viewportBack:
				viewportName = i18n("Back");
				break;
			case viewportLeft:
				viewportName = i18n("Left");
				break;
			case viewportRight:
				viewportName = i18n("Right");
				break;
			case viewportFree:
				viewportName = i18n("3D");
				break;
			default:
				viewportName = i18n("(null)");
		}
		
		QFont font;
		QFontMetrics metrics(font);
		QBitmap bm(((metrics.boundingRect(viewportName).width()+32) % 32) * 32, metrics.height(), true);
		QPainter p(&bm, 10);
		p.setFont(font);
		p.drawText(bm.rect(), Qt::AlignLeft | Qt::AlignBottom, viewportName);
		p.end();
		
		if (viewportText)
			delete viewportText;
		
		viewportText = new QImage(bm.convertToImage().mirror().convertBitOrder(QImage::BigEndian));
	}
}

void KLDrawRenderWidget::resize(int width, int height)
{
	glViewport(0, 0, width, height);
	windowWidth = (float)width;
	windowHeight = (float)height;
	windowLength = sqrt(pow((double)windowWidth, 2.0) + pow((double)windowHeight, 2.0));
	
	// Reset arcball
	if (arcBall && windowWidth > 1.0f && windowHeight > 1.0f) {
		arcBall->adjustedWidth  = 1.0f / ((windowWidth  - 1.0f) * 0.5f);
		arcBall->adjustedHeight = 1.0f / ((windowHeight - 1.0f) * 0.5f);
	}
	
	setViewport();
}

void KLDrawRenderWidget::setViewport()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	const KLDrawViewport *viewport;
	
	if (activeDocument_ && *activeDocument_) {
		switch (viewportMode) {
			case viewportTop:
				viewport = &((*activeDocument_)->getViewport(0));
				break;
				
			case viewportBottom:
				viewport = &((*activeDocument_)->getViewport(1));
				break;
				
			case viewportFront:
				viewport = &((*activeDocument_)->getViewport(2));
				break;
				
			case viewportBack:
				viewport = &((*activeDocument_)->getViewport(3));
				break;
				
			case viewportLeft:
				viewport = &((*activeDocument_)->getViewport(4));
				break;
				
			case viewportRight:
				viewport = &((*activeDocument_)->getViewport(5));
				break;
				
			default: // Free-view
				viewport = &((*activeDocument_)->getViewport(6));
		}
	} else {
		viewport = &defaultViewport;
	}
	
	float median, d;
	
	if (fabs(viewport->bottom-viewport->top)*(windowWidth/windowHeight) >= fabs(viewport->right-viewport->left)) {
		median = (viewport->right + viewport->left) * 0.5f;
		d = fabs((viewport->right-median)*(windowWidth/windowHeight))/viewport->aspectRatio;
		stretched->left = median - d;
		stretched->right = median + d;
		stretched->bottom = viewport->bottom;
		stretched->top = viewport->top;
		glOrtho(stretched->left, stretched->right, stretched->bottom, stretched->top, 10000.0f, -10000.0f);
	} else {
		median = (viewport->top + viewport->bottom) * 0.5f;
		d = fabs((viewport->bottom-median)/(windowWidth/windowHeight))*viewport->aspectRatio;
		stretched->left = viewport->left;
		stretched->right = viewport->right;
		stretched->bottom = median + d;
		stretched->top = median - d;
		glOrtho(stretched->left, stretched->right, stretched->bottom, stretched->top, 10000.0f, -10000.0f);
	}
}

// OpenGL environment initialization
void KLDrawRenderWidget::initializeGL()
{
	GLfloat ambient[] = { 0.45f, 0.45f, 0.45f, 1.0f };
	GLfloat diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
	GLfloat lightpos[] = { 300.0f, -500.0f, 150.0f, 1.0f };
	
	// Set default color and depth
	glClearColor(0.97f, 0.97f, 0.97f, 1.0f);
	glClearDepth(1.0f);
	glShadeModel(GL_SMOOTH);
	
	// Enable depth test
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	// Set blend enabled
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	// Set light
	glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT1, GL_POSITION, lightpos);
	glEnable(GL_LIGHT1);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
}

// Screen refresh entrypoint
void KLDrawRenderWidget::paintGL()
{	
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	if (!(*activeDocument_))
		return;
	
	setViewport();
	glMatrixMode(GL_MODELVIEW);
	
	glPushMatrix();
	glLoadIdentity();
	
	rotate();
	
	model *m =(*activeDocument_)->getActiveModel();
	
	if (!m->has_dimension())
		m->get_dimension();
	
	if (!boundingBox) {
		glLineWidth(1.0 + windowLength/2400.0);
		if (m->size() > 0 && KLDrawList::instance()->selected()) {
			drawModel(m, 0, KLDrawList::instance()->selected()->id());
		}
	} else {
		if (m->size() > 0 && KLDrawList::instance()->selected()) {
			drawBoundingBoxes(m, 0, dragResources_->dragId);
		}
	}
	
	renderGrid(20.0f, 20.0f, 50, 50, 0.0f, 0.0f, 0.0f);
	
	if (dragResources_->dragging) {
		glMultMatrixf(dragResources_->dragPos.transpose().get_pointer());
		glColor3f(0.0f, 1.0f, 0.0f);
		glLineWidth(2.0f);
		renderBoundingBox(dragResources_->dragBBoxMin, dragResources_->dragBBoxMax);
	}
	
	glPopMatrix();
	
	renderViewportName();
}

// Screen recalibration
void KLDrawRenderWidget::resizeGL(int width, int height)
{
	resize(width, height);
}

void KLDrawRenderWidget::mousePressEvent(QMouseEvent *event)
{
	if (!(*activeDocument_)) return;
	
	if (event->button() & Qt::MidButton) {
		panning = true;
		boundingBox = true;
		lastPos = event->pos();
		updateGL();
	} else if (arcBall && event->button() & Qt::LeftButton) {
		rotating = true;
		boundingBox = true;
		
		(*activeDocument_)->getLastRotation() = (*activeDocument_)->getFreeViewport();
		
		arcBall->stVec = mapToSphere(event->pos());
	} else if (!arcBall && event->button() & Qt::LeftButton) {
		if (!KLDrawMainWindow::self()->getCurrentCursor() || KLDrawMainWindow::self()->getCurrentCursor()->item()->get_type() != type_extref)
			return;
		
		element_extref *o = CAST_AS_EXTREF(KLDrawMainWindow::self()->getCurrentCursor()->item());
		
		if (!o->get_model())
			return;
		
		if (!o->get_model()->has_dimension())
			o->get_model()->get_dimension();
		
		// bbox-based collision detection
		vector3d minVec = o->get_matrix() * o->get_model()->dimension().first;
		vector3d maxVec = o->get_matrix() * o->get_model()->dimension().second;
		QPoint pos1 = globalCoordinate(minVec);
		QPoint pos2 = globalCoordinate(maxVec);
		int rectL = pos1.x() < pos2.x() ? pos1.x() : pos2.x();
		int rectT = pos1.y() < pos2.y() ? pos1.y() : pos2.y();
		int rectW = pos1.x() > pos2.x() ? pos1.x() - pos2.x() : pos2.x() - pos1.x();
		int rectH = pos1.y() > pos2.y() ? pos1.y() - pos2.y() : pos2.y() - pos1.y();
		if (!QRect(rectL, rectT, rectW, rectH).contains(event->pos()))
			return;
		
		moveObject(event->pos());
		
		dragResources_->dragId = KLDrawMainWindow::self()->getCurrentCursor()->id();
		dragResources_->dragObject = o;
		
		emit startDragging(o->get_model()->dimension().first, o->get_model()->dimension().second);
	}
}

void KLDrawRenderWidget::mouseMoveEvent(QMouseEvent *event)
{
	if (!(*activeDocument_)) return;
	
	if (panning) {
		int viewportId;
		switch (viewportMode) {
			case viewportTop:
				viewportId = 0;
				break;
			case viewportBottom:
				viewportId = 1;
				break;
			case viewportFront:
				viewportId = 2;
				break;
			case viewportBack:
				viewportId = 3;
				break;
			case viewportLeft:
				viewportId = 4;
				break;
			case viewportRight:
				viewportId = 5;
				break;
			case viewportFree:
				viewportId = 6;
				break;
			default:
				viewportId = -1;
		}
		
		KLDrawViewport &viewport = (*activeDocument_)->getViewport(viewportId);
		float dx = (event->pos().x() - lastPos.x()) / windowWidth * fabs(stretched->right - stretched->left);
		float dy = (event->pos().y() - lastPos.y()) / windowHeight * fabs(stretched->bottom - stretched->top);
		
		viewport.left   -= dx;
		viewport.right  -= dx;
		viewport.top    -= dy;
		viewport.bottom -= dy;
		
		lastPos = event->pos();
		
		updateGL();
	} else if (rotating) {
		// Arcball viewport rotation
		
		arcBall->enVec = mapToSphere(event->pos());
		
		vector3d perp = vector3d::cross_product(arcBall->stVec, arcBall->enVec);
		vector3d quat;
		
		if (perp.length() > EPSILON)
			quat = vector3d(perp.x(), perp.y(), perp.z(), vector3d::dot_product(arcBall->stVec, arcBall->enVec));
		else
			quat.w() = 0.0f;
		
		matrix &cur = (*activeDocument_)->getFreeViewport();
		matrix &last = (*activeDocument_)->getLastRotation();
		
		float n, s;
		float xs, ys, zs;
		float wx, wy, wz;
		float xx, xy, xz;
		float yy, yz, zz;
		
		n = quat.x()*quat.x() + quat.y()*quat.y() + quat.z()*quat.z() + quat.w()*quat.w();
		s = n>0.0f ? 2.0f/n : 0.0f;
		
		xs = quat.x()*s , ys = quat.y()*s , zs = quat.z()*s ;
		wx = quat.w()*xs, wy = quat.w()*ys, wz = quat.w()*zs;
		xx = quat.x()*xs, xy = quat.x()*ys, xz = quat.x()*zs;
		yy = quat.y()*ys, yz = quat.y()*zs, zz = quat.z()*zs;
		
		cur.value(0, 0) = 1.0f - (yy+zz); cur.value(0, 1) =         xy-wz ; cur.value(0, 2) =         xz+wy ;
		cur.value(1, 0) =         xy+wz ; cur.value(1, 1) = 1.0f - (xx+zz); cur.value(1, 2) =         yz-wx ;
		cur.value(2, 0) =         xz-wy ; cur.value(2, 1) =         yz+wx ; cur.value(2, 2) = 1.0f - (xx+yy);
		
		cur = cur * last;
		
		updateGL();
	} else if (dragResources_->dragging) {
		if (viewportMode == viewportFree)
			return;
		
		moveObject(event->pos());
		
		emit objectDragging();
	}
}

void KLDrawRenderWidget::mouseReleaseEvent(QMouseEvent *)
{
	if (panning) {
		panning = false;
		boundingBox = false;
		updateGL();
	} else if (rotating) {
		rotating = false;
		boundingBox = false;
		updateGL();
	} else if (dragResources_->dragging) {
		KLDrawMainWindow::self()->getEditor()->moveCurrentObject(dragResources_->dragPos);
		
		dragResources_->dragId = -1;
		
		emit stopDragging();
	}
}

void KLDrawRenderWidget::wheelEvent(QWheelEvent *e)
{
	if (!(*activeDocument_)) return;
	
	int viewportId;
	switch (viewportMode) {
		case viewportTop:
			viewportId = 0;
			break;
		case viewportBottom:
			viewportId = 1;
			break;
		case viewportFront:
			viewportId = 2;
			break;
		case viewportBack:
			viewportId = 3;
			break;
		case viewportLeft:
			viewportId = 4;
			break;
		case viewportRight:
			viewportId = 5;
			break;
		case viewportFree:
			viewportId = 6;
			break;
		default:
			viewportId = -1;
	}
	
	KLDrawViewport &viewport = (*activeDocument_)->getViewport(viewportId);
	float centerx = (viewport.left + viewport.right ) * 0.5f;
	float centery = (viewport.top  + viewport.bottom) * 0.5f;
	
	if (e->delta() < 0) {
		// Zoom in
		viewport.left   = centerx - (centerx - viewport.left) * MAGNIFICATION_FACTOR;
		viewport.right  = centerx + (viewport.right - centerx) * MAGNIFICATION_FACTOR;
		viewport.top    = centery - (centery - viewport.top) * MAGNIFICATION_FACTOR;
		viewport.bottom = centery + (viewport.bottom - centery) * MAGNIFICATION_FACTOR;
	} else if (e->delta() > 0) {
		// Zoom out
		viewport.left   = centerx - (centerx - viewport.left) / MAGNIFICATION_FACTOR;
		viewport.right  = centerx + (viewport.right - centerx) / MAGNIFICATION_FACTOR;
		viewport.top    = centery - (centery - viewport.top) / MAGNIFICATION_FACTOR;
		viewport.bottom = centery + (viewport.bottom - centery) / MAGNIFICATION_FACTOR;
	}
	
	updateGL();
}

void KLDrawRenderWidget::renderGrid(float xg, float yg, int xc, int yc, float xo, float yo, float zo)
{
	glLineWidth(1.0f);
	
	switch (viewportMode) {
		case viewportFront:
		case viewportBack:
			glColor3ub(64, 64, 64);
			glEnable(GL_DEPTH_TEST);
			glBegin(GL_LINES);
			glVertex3f(-(xg*(float)xc)+xo, zo, 0.0f);
			glVertex3f( (xg*(float)xc)+xo, zo, 0.0f);
			glEnd();
			glDisable(GL_DEPTH_TEST);
			return;
			
		case viewportLeft:
		case viewportRight:
			glColor3ub(64, 64, 64);
			glEnable(GL_DEPTH_TEST);
			glBegin(GL_LINES);
			glVertex3f(0.0f, zo, -(yg*(float)yc)+yo);
			glVertex3f(0.0f, zo,  (yg*(float)yc)+yo);
			glEnd();
			glDisable(GL_DEPTH_TEST);
			return;
			
		default:
			glBegin(GL_LINES);
			
			glColor3ub(64, 64, 64);
			
			glVertex3f(-(xg*(float)xc)+xo, zo, yo);
			glVertex3f( (xg*(float)xc)+xo, zo, yo);
			
			glVertex3f(xo, zo, -(yg*(float)yc)+yo);
			glVertex3f(xo, zo,  (yg*(float)yc)+yo);
			
			glColor4ub(128, 128, 128, 128);
			
			for (int i = 1; i <= xc; i++) {
				glVertex3f(-(xg*(float)i)+xo, zo, -(yg*(float)yc)+yo);
				glVertex3f(-(xg*(float)i)+xo, zo,  (yg*(float)yc)+yo);
				
				glVertex3f( (xg*(float)i)+xo, zo, -(yg*(float)yc)+yo);
				glVertex3f( (xg*(float)i)+xo, zo,  (yg*(float)yc)+yo);
			}
			
			for (int i = 1; i <= yc; i++) {
				glVertex3f(-(xg*(float)xc)+xo, zo, -(yg*(float)i)+yo);
				glVertex3f( (xg*(float)xc)+xo, zo, -(yg*(float)i)+yo);
				
				glVertex3f(-(xg*(float)xc)+xo, zo,  (yg*(float)i)+yo);
				glVertex3f( (xg*(float)xc)+xo, zo,  (yg*(float)i)+yo);
			}
			
			glEnd();
			
			return;
	}
}

void KLDrawRenderWidget::renderViewportName()
{
	if (!viewportText) return;
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0f, windowWidth, windowHeight, 0.0f, 10000.0f, -10000.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(0.0f, 0.0f, 0.0f);
	glDisable(GL_DEPTH_TEST);
	glRasterPos2d(10, 10+viewportText->height());
	glBitmap(viewportText->width(), viewportText->height(), 0, 0, 0, 0, viewportText->bits());
	glEnable(GL_DEPTH_TEST);
}

vector3d KLDrawRenderWidget::mapToSphere(const QPoint &point)
{
	if (!arcBall) return vector3d();
	
	vector3d out;
	float tx, ty;
	float length;
	
	tx =        ((float)point.x() * arcBall->adjustedWidth) - 1.0f;
	ty = 1.0f - ((float)point.y() * arcBall->adjustedHeight);
	
	length = tx*tx + ty*ty;
	
	if (length > 1.0f) {
		// If the point is outside of the sphere
		float norm = 1.0f / sqrt(length);
		out = vector3d(-tx * norm, ty * norm, 0.0f);
	} else
		out = vector3d(-tx, ty, sqrt(1.0f - length));
	
	return out;
}

// Set viewport
void KLDrawRenderWidget::rotate()
{
	switch (viewportMode) {
		case viewportFree:
			glLoadMatrixf((*activeDocument_)->getFreeViewport().transpose().get_pointer());
			break;
		case viewportTop:
			glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
			break;
		case viewportBottom:
			glRotatef(90.0f, -1.0f, 0.0f, 0.0f);
			break;
		case viewportBack:
			glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
			break;
		case viewportLeft:
			glRotatef(90.0f, 0.0f, -1.0f, 0.0f);
			break;
		case viewportRight:
			glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
			break;
		default:
			break;
	}
}

// When dragging a part into this view
void KLDrawRenderWidget::dragEnterEvent(QDragEnterEvent *e)
{
	if (!QTextDrag::canDecode(e) || viewportMode == viewportFree)
		return;
	
	e->accept(true);
	
	QString serialized;
	QTextDrag::decode(e, serialized);
	
	QStringList list = QStringList::split(':', serialized);
	vector3d objMin(list[1].toFloat(), list[2].toFloat(), list[3].toFloat());
	vector3d objMax(list[4].toFloat(), list[5].toFloat(), list[6].toFloat());
	emit startDragging(objMin, objMax);
}

void KLDrawRenderWidget::dragMoveEvent(QDragMoveEvent *e)
{
	if (viewportMode == viewportFree)
		return;
	
	moveObject(e->pos());
	
	emit objectDragging();
}

void KLDrawRenderWidget::dragLeaveEvent(QDragLeaveEvent */*e*/)
{
	emit stopDragging();
}

void KLDrawRenderWidget::dropEvent(QDropEvent *e)
{
	emit stopDragging();
	
	QString serialized;
	QTextDrag::decode(e, serialized);
	
	editor->insertObject(serialized.section(':', 0, 0), dragResources_->dragPos, dragResources_->lastColor);
}

void KLDrawRenderWidget::moveObject(const QPoint &p)
{
	vector3d location = viewportCoordinate(p);
	matrix matrix = dragResources_->lastMatrix;
	
	for (int i = 0; i < 3; ++i) {
		if (depthAxis() != i) {
			matrix.value(i, 3) = KLDraw::modulo(matrix.value(i, 3), editor->gridScale(i)) + ((int)(location[i] / editor->gridScale(i)) * editor->gridScale(i));
		} else {
			matrix.value(i, 3) = matrix.value(i, 3);
		}
	}
	
	dragResources_->dragPos = matrix;
}

// Context menu event
void KLDrawRenderWidget::contextMenu()
{
	
}
